#!/usr/bin/perl 

# FXserv IRC Services, Copyright (c) 2002-2004 Christian Vandaele.
# This program comes with absolutely NO WARRANTY
#
# Should you choose to use and/or modify this source code, please
# do so under the terms of the GNU General Public License under which
# this program is distributed.

require v5.6.0;

use strict;
use IO::Socket;
use POSIX ":sys_wait_h";

our (%FXserv, $ConfigFile, $PidFile, $ts);

$| = 1;

require "src/bot.pl";
require "src/channel.pl";
require "src/client.pl";
require "src/conf.pl";
require "src/dataserv.pl";
require "src/dcc.pl";
require "src/init.pl";
require "src/irc.pl";
require "src/log.pl";
require "src/routines.pl";
require "src/server.pl";
require "src/socket.pl";
require "src/timer.pl";

$SIG{'TERM'} = sub { &putinfo("y", "Received SIGTERM, shutting down"); &shutdown(0, "Received SIGTERM"); };
$SIG{'KILL'} = sub { &putinfo("y", "Received SIGKILL, restarting services"); &restart(); };
$SIG{'SEGV'} = sub { &putinfo("y", "Received SIGSEGV, shutting down"); &shutdown(0, "ACK! SIGSEGV!"); };
$SIG{'PIPE'} = sub { &putinfo("y", "Received SIGPIPE, ignoring"); };
$SIG{'ALRM'} = sub { die(); };
$SIG{'CHLD'} = '';

$FXserv{'started'} = $FXserv{'last_sync'} = $ts = time();
$FXserv{'version'} = "1.7pre8";
$FXserv{'connected'} = 0;
$FXserv{'netmerge'} = 0;
$FXserv{'pid'} = $$;

sub start {
  print STDOUT "FXserv IRC Services version $FXserv{'version'}\n";
  print STDOUT "Copyright (c) 2002-2004 Christian Vandaele\n";

  if (($^O ne "MSWin32") && ($> == 0)) {
    print STDERR "Don't run FXserv as root!\n";
    exit(1);
  }

  for (my $idx = 0; $idx < $#ARGV + 1; $idx++) {
    if (($ARGV[$idx] eq "-h") || ($ARGV[$idx] eq "--help")) {
      print STDOUT "Usage: $0 <options>\n";
      print STDOUT "  -c --config <file>   use <file> for settings configuration\n";
      print STDOUT "  -h --help            this screen\n";
      print STDOUT "  -v --version         print version number\n";
      print STDOUT "\n";
      exit(0);
    } elsif ($ARGV[$idx] eq "-v" || ($ARGV[$idx] eq "--version")) {
      print STDOUT "\n";
      exit(0);
    } elsif ($ARGV[$idx] eq "-c" || ($ARGV[$idx] eq "--config")) {
      $idx++;

      if (($ARGV[$idx] ne "") && (-e $ARGV[$idx])) {
        $ConfigFile = $ARGV[$idx];
      } else {
        print STDOUT "Usage:\n";
        print STDOUT "  --config <file>\n";
        print STDOUT "\n";
        exit(0);
      }
    }
  }

  if (!$ConfigFile) { $ConfigFile = "FXserv.conf"; }

  if (&load_conf()) {
    print STDERR "Error parsing configuration file $ConfigFile\n";
    exit(1);
  }

  my $pid = fork();

  if ($pid == -1) {
    print STDERR "Unable to fork()\n";
    exit(1);
  }

  if ($pid) {
    print STDOUT "Launched into background (pid: $pid)\n";
    exit(0);
  }

  if (!$FXserv{'debug'}) {
    close(STDIN);
    close(STDOUT);
    close(STDERR);
  }

  &init_vars();
  &init_listen();

  if (!open(PID, ">$PidFile")) {
    &putlog("y", "Unable to open pid file: %s", $PidFile);
  } else {
    print PID $$;
    close(PID);
  }

  &putlog(0, "FXserv IRC Services version %s started (pid: %d)", $FXserv{'version'}, $$);

  $FXserv{'pid'} = $$;

  &init_connect();

  for (;;) {
    &loop_timer();
    &loop_socket();
  }

  exit(1);
}

&start();

1;
