use strict;
use Config;
use Digest::MD5 qw(md5_hex);

our (%FXbnet, %FXcfg, %FXflags, %FXirc, %FXmsg, %FXserv, %FXbot, %FXsock, %FXoper, %FXstats, %MySQL, $LangPack, $bit, $ts);
my %map = ();

$FXbot{'DataServ'}{'sub'} = {
  'HELP'      => \&ds_help,
  'ID'        => \&ds_identify,
  'IDENTIFY'  => \&ds_identify,
  'MAP'       => \&ds_map,
  'MONITOR'   => \&ds_monitor,
  'CLONES'    => \&ds_clones,
  'RESTART'   => \&ds_restart,
  'DIE'       => \&ds_die,
  'OMODE'     => \&ds_omode,
  'KILL'      => \&ds_kill,
  'UNKILL'    => \&ds_unkill,
  'STATS'     => \&ds_stats,
  'STATUS'    => \&ds_status,
  'ENABLE'    => \&ds_enable,
  'DISABLE'   => \&ds_disable,
  'SAVE'      => \&ds_save,
  'GLOBAL'    => \&ds_global,
  'TRACE'     => \&ds_trace,
  'SEEN'      => \&ds_seen,
  'WHOWAS'    => \&ds_seen,
  'UNSEEN'    => \&ds_unseen,
  'SEENSTATS' => \&ds_seenstats,
  'EXEMPT'    => \&ds_exempt,
  'IGNORE'    => \&ds_ignore
};

$FXbot{'DataServ'}{'public_sub'} = {
  'MAP'    => \&ds_map,
  'SEEN'   => \&ds_seen,
  'WHOWAS' => \&ds_seen,
};

$FXbot{'DataServ'}{'channel_sub'} = {
  'OP'       => \&ds_public,
  'DEOP'     => \&ds_public,
  'HOP'      => \&ds_public,
  'DEHOP'    => \&ds_public,
  'HALFOP'   => \&ds_public,
  'DEHALFOP' => \&ds_public,
  'VOICE'    => \&ds_public,
  'DEVOICE'  => \&ds_public,
  'KICK'     => \&ds_public,
  'BAN'      => \&ds_public,
  'BANKICK'  => \&ds_public,
  'MODE'     => \&ds_public,
  'TOPIC'    => \&ds_public,
  'INVITE'   => \&ds_public
};

$FXbot{'DataServ'}{'other_sub'} = {
  'process_ds' => \&process_ds,
  'ds_dumpmap' => \&ds_dumpmap,
  'ds_dcc'     => \&ds_dcc,
  'ds_public'  => \&ds_public,
  'ds_notice'  => \&ds_notice,
  'ds_ctcp'    => \&ds_ctcp
};

sub ds_notice {
  my ($nick, $format, @arg) = (shift, shift, @_);
  my $buffer = sprintf($format, @arg);

  if ($FXsock{'list'}{$nick}) {
    &putsock($nick, $buffer);
    return(0);
  }

  if ($FXirc{'client'}{lc($nick)}) {
    &putsock($FXsock{'irc'}, ":%s %s %s :%s", $FXbot{'DataServ'}{'nick'}, $FXcfg{'general'}{'communicate'}, $nick, $buffer);
    return(0);
  }
}

sub ds_ctcp {
  my ($nick, $notice, $format, @arg) = (shift, shift, @_);
  my $buffer = sprintf($format, @arg);

  &putsock($FXsock{'irc'}, ":%s %s %s :\cA%s\cA", $FXbot{'DataServ'}{'nick'}, $notice ? "NOTICE" : "PRIVMSG", $nick, $buffer);
  return(0);
}

sub process_ds {
  my ($nick, $query) = (shift, shift);

  if (&check_flood($nick)) { return(0); }

  if ($query =~ /\cA(.*)\cA/) {
    my $ctcp = uc((split(/ /, $1, 0))[0]);

    if ($ctcp eq "VERSION") {
      &ds_ctcp($nick, 1, "VERSION %s", "FXserv IRC services $FXserv{'version'}");
      &ds_ctcp($nick, 1, "VERSION %s", "Administrator: $FXserv{'admin'}");
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $FXbot{'DataServ'}{'nick'}, $ctcp, $FXirc{'client'}{lc($nick)}{'nick'}, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    } elsif ($ctcp eq "PING") {
      &ds_ctcp($nick, 1, "PING %s", (split(/ /, $1, 0))[1]);
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $FXbot{'DataServ'}{'nick'}, $ctcp, $FXirc{'client'}{lc($nick)}{'nick'}, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    } elsif ($ctcp eq "TIME") {
      &ds_ctcp($nick, 1, "TIME %s", scalar(localtime()));
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $FXbot{'DataServ'}{'nick'}, $ctcp, $FXirc{'client'}{lc($nick)}{'nick'}, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    } elsif ($ctcp eq "DCC") {
      if ($query =~ /DCC CHAT chat (\d+) (\d+)/i) { &ds_dcc($nick, $1, $2); }
    } else {
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $FXbot{'DataServ'}{'nick'}, $1, $FXirc{'client'}{lc($nick)}{'nick'}, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    }

    return(0);
  }

  my $cmd = (split(/ /, $query, 0))[0];
  my $arg = (split(/ /, $query, 2))[1];

  if ($FXbot{'DataServ'}{'sub'}->{uc($cmd)}) {
    &{$FXbot{'DataServ'}{'sub'}->{uc($cmd)}}($nick, $arg);
  } else {
    &ds_notice($nick, $FXmsg{'unkn_cmd'}, $cmd);
  }

  return(0);
}

sub ds_help {
  my ($nick, $arg) = (shift, shift);

  if (!$arg) { $arg = "index"; }

  my $file = lc($arg);
  $file =~ s/ /\//g;

  if ($FXsock{'list'}{$nick}) { $file = "help/" . $LangPack . "/dataserv/dcc/" . $file; }
  if ($FXirc{'client'}{lc($nick)}) { $file = "help/" . $LangPack . "/dataserv/" . $file; }

  if (!open(FILE, "<" . $file)) {
    &ds_notice($nick, $FXmsg{'no_help'}, $arg);
    return(0);
  }

  my @lines = <FILE>;
  close(FILE);

  foreach my $buffer (@lines) {
    chomp($buffer);
    $buffer = &substitute($buffer);

    if ($buffer =~ /^\%\+a/) {
      if (&valid_oper($nick, "a")) {
        $buffer =~ s/^\%\+a//;
        &ds_notice($nick, $buffer);
      }
    } elsif ($buffer =~ /^\%\+o/) {
      if (&valid_oper($nick, "o")) {
        $buffer =~ s/^\%\+o//;
        &ds_notice($nick, $buffer);
      }
    } else {
      &ds_notice($nick, $buffer);
    }
  }

  &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "HELP %s", $arg);
  return(0);
}

sub ds_map {
  my ($nick, $server) = (shift, shift);
  my @prefix = ();

  if (!$FXsock{'irc'}) {
    &ds_notice($nick, $FXmsg{'no_irc'});
    return(0);
  }

  for (my $inc = 0; $inc < 64; ++$inc) { $prefix[$inc] = ""; }

  if (!$server) {
    $server = $FXserv{'hub'};
  } else {
    $server = (split(/ /, $server, 0))[0];

    if (!$FXirc{'server'}{lc($server)}) {
      &ds_notice($nick, $FXmsg{'no_server'}, $server);
      return(0);
    }
  }

  &ds_notice($nick, "-- Network map --");
  &ds_dumpmap($nick, $server, 0, @prefix);
  &ds_notice($nick, " ");
  &ds_notice($nick, "Total: %d user%s on %d server%s", $map{lc($nick)}{'users'},  $map{lc($nick)}{'users'} > 1 ? "s" : "", $map{lc($nick)}{'servers'}, $map{lc($nick)}{'servers'} > 1 ? "s" : "");
  &ds_notice($nick, "-- End of map --");

  delete($map{lc($nick)});
  &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "MAP %s", $server);
  return(0);
}

sub ds_dumpmap {
  my ($nick, $server, $prompt, @prefix) = (shift, shift, shift, @_);
  my $cnt = scalar(keys(%{$FXirc{'server'}{lc($server)}{'leaf'}}));
  my $output = "";

  ++$map{lc($nick)}{'servers'};
  $map{lc($nick)}{'users'} += $FXirc{'server'}{lc($server)}{'users'}{'current'};
  $map{lc($nick)}{'opers'} += $FXirc{'server'}{lc($server)}{'opers'}{'current'};

  $prefix[$prompt] = "\0";

  foreach (@prefix) {
    if ($_ eq "\0") { last; }
    $output .= $_;
  }

  &ds_notice($nick, "%s%s (%d user%s, %d oper%s) [lag: %s]", $output, $FXirc{'server'}{lc($server)}{'name'}, $FXirc{'server'}{lc($server)}{'users'}{'current'}, $FXirc{'server'}{lc($server)}{'users'}{'current'} > 1 ? "s" : "", $FXirc{'server'}{lc($server)}{'opers'}{'current'}, $FXirc{'server'}{lc($server)}{'opers'}{'current'} > 1 ? "s" : "", &format_time($FXirc{'server'}{lc($server)}{'lag'} + $ts));

  if ($prompt > 0) {
    $prefix[$prompt - 1] = " ";
    if ($prefix[$prompt - 2] eq "`") { $prefix[$prompt - 2] = " "; }
  }

  $prefix[$prompt] = "|";
  $prefix[$prompt + 1] = "-";

  foreach (keys(%{$FXirc{'server'}{lc($server)}{'leaf'}})) {
    if (--$cnt == 0) { $prefix[$prompt] = "`"; }
    &ds_dumpmap($nick, $FXirc{'server'}{$_}{'name'}, $prompt + 2, @prefix);
  }

  if ($prompt > 0) { $prefix[$prompt - 1] = "-"; }

  return(0);
}

sub ds_dcc {
  my ($nick, $addr, $port) = (shift, shift, shift);

  if (!&valid_oper($nick, "d")) {
    &ds_notice($nick, $FXmsg{'no_priv'});
    &putinfo("y", "Unauthorized DCC Chat attempt from %s (%s\@%s)", $nick, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    return(0);
  }

  my $oper = $FXirc{'client'}{lc($nick)}{'oper'};

  if (!&valid_port($port)) {
    &ds_notice($nick, $FXmsg{'dcc_invalid'});
    &putlog("y", "Bogus DCC Chat port (%d) specified by %s!%s\@%s", $port, $nick, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    return(0);
  }

  if ($FXstats{'partyline'} + 1 > $FXcfg{'general'}{'max_connections'}) {
    &ds_notice($nick, $FXmsg{'dcc_max'});
    return(0);
  }

  my $ip = inet_ntoa(pack("N", $addr));
  my $dcc = &open_socket("tcp", $ip, $port, 3);

  if (!$dcc) {
    &ds_notice($nick, $FXmsg{'dcc_failed'});
    return(0);
  }

  &add_socket($dcc, "DCC");

  $FXsock{'list'}{$dcc}{'nick'} = $FXirc{'client'}{lc($nick)}{'nick'};
  $FXsock{'list'}{$dcc}{'user'} = $FXirc{'client'}{lc($nick)}{'user'};
  $FXsock{'list'}{$dcc}{'oper'} = $oper;
  $FXsock{'list'}{$dcc}{'lastcmd'} = $ts;
  &send_motd($dcc, $FXcfg{'files'}{'motd'});

  ++$FXstats{'partyline'};

  my $prefix = "User";

  if (&valid_oper($nick, "a")) {
    $prefix = "Admin";
  } elsif (&valid_oper($nick, "o")) {
    $prefix = "Oper";
  }

  &putinfo(0, "%s %s (%s\@%s) has connected", $prefix, $FXsock{'list'}{$dcc}{'nick'}, $FXsock{'list'}{$dcc}{'user'}, $FXsock{'list'}{$dcc}{'host'});
  &putbnet("(%s) %s %s (%s\@%s) has connected", $FXbot{'DataServ'}{'nick'}, $prefix, $FXsock{'list'}{$dcc}{'nick'}, $FXsock{'list'}{$dcc}{'user'}, $FXsock{'list'}{$dcc}{'host'});

  return(0);
}

sub ds_identify {
  my ($nick, $arg) = (shift, shift);
  my $count = $arg =~ tr/ / /; ++$count;
  my $oper = "";
  my $password = "";

  if ($count > 1) {
    $oper = (split(/ /, $arg, 0))[0];
    $password = (split(/ /, $arg, 0))[1];
  } else {
    $oper = $nick;
    $password = (split(/ /, $arg, 0))[0];
  }

  if (!$password) {
    &ds_notice($nick, $FXmsg{'syntax'}, "IDENTIFY [login] <password>");
    return(0);
  }

  if (&valid_oper($nick, undef)) {
    &ds_notice($nick, $FXmsg{'id_already'});
    return(0);
  }

  if (!$FXoper{lc($oper)}) {
    &ds_notice($nick, $FXmsg{'denied'});
    return(0);
  }

  my $usermask = (split(/\@/, $FXoper{lc($oper)}{'mask'}, 0))[0]; my $umaskrx = &plain2regexp($usermask);
  my $hostmask = (split(/\@/, $FXoper{lc($oper)}{'mask'}, 0))[1]; my $hmaskrx = &plain2regexp($hostmask);

  if (($FXirc{'client'}{lc($nick)}{'user'} !~ /$umaskrx/i) && ($FXirc{'client'}{lc($nick)}{'host'} !~ /$hmaskrx/i)) {
    &ds_notice($nick, $FXmsg{'bad_host'});
    &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "failed IDENTIFY");
    return(0);
  }

  if (md5_hex($password) eq $FXoper{lc($oper)}{'password'}) {
    $FXirc{'client'}{lc($nick)}{'oper'} = $FXoper{lc($oper)}{'name'};
    &ds_notice($nick, $FXmsg{'id_ok'});
    &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "IDENTIFY");
    return(0);
  } else {
    &ds_notice($nick, $FXmsg{'bad_pass'});
    &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "failed IDENTIFY");
    return(0);
  }
}

sub ds_restart {
  my $nick = shift;

  if (!&valid_oper($nick, "a")) {
    &ds_notice($nick, $FXmsg{'no_priv'});
    return(0);
  }

  &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "RESTART");
  &restart();
}

sub ds_die {
  my ($nick, $reason) = (shift, shift);

  if (!&valid_oper($nick, "a")) {
    &ds_notice($nick, $FXmsg{'no_priv'});
    return(0);
  }

  &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "DIE %s", $reason);
  &shutdown(1, $reason);
}

sub ds_omode {
  my ($nick, $omode) = (shift, shift);
  my $count = $omode =~ tr/ / /; ++$count;

  if (!&valid_oper($nick, "o")) {
    &ds_notice($nick, $FXmsg{'no_priv'});
    return(0);
  }

  if (!$FXsock{'irc'}) {
    &ds_notice($nick, $FXmsg{'no_irc'});
    return(0);
  }

  if ($count < 2) {
    &ds_notice($nick, $FXmsg{'syntax'}, "OMODE <channel> <modes> [list]");
    return(0);
  }

  my $channel = (split(/ /, $omode, 0))[0];
  my $mode = (split(/ /, $omode, 0))[1];
  my $args = (split(/ /, $omode, 3))[2];
  my @list = ();

  foreach (split(/ /, $args)) { push(@list, $_); }

  if (!$FXirc{'channel'}{lc($channel)}) {
    &ds_notice($nick, $FXmsg{'no_chan'}, $nick, $channel);
    return(0);
  }

  my $idx = 0;
  my $cnt = 0;
  my $what = 1;
  my $exec = "";
  my $target = "";

  foreach (split(//, $mode)) {
    if ($_ eq "+") {
      $what = 1;
      next;
    } elsif ($_ eq "-") {
      $what = 0;
      next;
    }

    if (($_ eq "o") || ($_ eq "h") || ($_ eq "v")) {
      if ($FXirc{'channel'}{lc($channel)}{'names'}{lc($list[$idx])}) {
        if (!$what) { $exec .= "-"; }
        if ($what) { $exec .= "+"; }

        $exec .= $_;
        $target .= $list[$idx++] . " ";
        $cnt++;
      } else {
        $idx++;
      }
    } elsif (($_ eq "b") || ($_ eq "d") || ($_ eq "e") || ($_ eq "I")) {
      if (!$what) { $exec .= "-"; }
      if ($what) { $exec .= "+"; }

      $exec .= $_;
      $target .= $list[$idx++] . " ";
      $cnt++;
    } elsif (($_ =~ /[fklL]/) && !(($_ eq "l") && !$what)) {
      if (!$what) { $exec .= "-"; }
      if ($what) { $exec .= "+"; }

      $exec .= $_;
      $target .= $list[$idx++] . " ";
      $cnt++;
    } elsif ($_ =~ /[cilmnpst]/) {
      if (!$what) { $exec .= "-"; }
      if ($what) { $exec .= "+"; }

      $exec .= $_;
      $cnt++;
    }

    if (!$cnt) { next; }

    if (!($cnt % $FXcfg{'general'}{'modes'})) {
      $target =~ s/\s+$//;

      &putsock($FXsock{'irc'}, ":%s MODE %s %s %s", $FXbot{'DataServ'}{'nick'}, $FXirc{'channel'}{lc($channel)}{'channel'}, $exec, $target);
      &chanlog("m", $FXirc{'channel'}{lc($channel)}{'channel'}, $FXbot{'DataServ'}{'nick'}, $exec, $target);
      &channel_mode($FXbot{'DataServ'}{'nick'}, $FXirc{'channel'}{lc($channel)}{'channel'}, $exec, $target);
      &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "OMODE %s %s %s", $FXirc{'channel'}{lc($channel)}{'channel'}, $exec, $target);

      $exec = "";
      $target = "";
    }
  }

  if ($exec) {
    $target =~ s/\s+$//;

    &putsock($FXsock{'irc'}, ":%s MODE %s %s %s", $FXbot{'DataServ'}{'nick'}, $FXirc{'channel'}{lc($channel)}{'channel'}, $exec, $target);
    &chanlog("m", $FXirc{'channel'}{lc($channel)}{'channel'}, $FXbot{'DataServ'}{'nick'}, $exec, $target);
    &channel_mode($FXbot{'DataServ'}{'nick'}, $FXirc{'channel'}{lc($channel)}{'channel'}, $exec, $target);
    &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "OMODE %s %s %s", $FXirc{'channel'}{lc($channel)}{'channel'}, $exec, $target);
  }

  return(0);
}

sub ds_stats {
  my ($nick, $switch) = (shift, shift);

  if (!&valid_oper($nick, "o")) {
    &ds_notice($nick, $FXmsg{'no_priv'});
    return(0);
  }

  if (!$switch) {
    &ds_notice($nick, $FXmsg{'syntax'}, "STATS <switch>");
    return(0);
  }

  $switch = (split(/ /, $switch, 0))[0];
  &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "STATS %s", $switch);

  if (uc($switch) eq "C") {
    my $idx = 0;

    &ds_notice($nick, "-- Monitored channels --");
    &ds_notice($nick, "[Service     ] [Channel             ] [Log] [Pub]");

    foreach my $service (keys(%FXbot)) {
      if (!$FXbot{$service}{'enabled'}) { next; }
      if (!$FXbot{$service}{'channel'}) { next; }

      $idx += scalar(keys(%{$FXbot{$service}{'channel'}}));
      foreach my $chan (keys(%{$FXbot{$service}{'channel'}})) {
        my $channel = $FXbot{$service}{'channel'}{$chan}{'channel'};
        my $mode = "";

        if ($FXbot{$service}{'channel'}{lc($channel)}{'mode'} eq "o") {
          $mode = "\@";
        } elsif ($FXbot{$service}{'channel'}{lc($channel)}{'mode'} eq "h") {
          $mode = "\%";
        } elsif ($FXbot{$service}{'channel'}{lc($channel)}{'mode'} eq "v") {
          $mode = "+";
        }

        &ds_notice($nick, "[%-12s] [%-20s] [%-3s] [%-3s]", $FXbot{$service}{'nick'}, $mode . $FXbot{$service}{'channel'}{lc($channel)}{'channel'}, $FXbot{$service}{'channel'}{lc($channel)}{'log_file'} ? "YES" : "NO", $FXbot{$service}{'channel'}{lc($channel)}{'public'} ? "YES" : "NO");
      }
    }

    &ds_notice($nick, "-- End of list [%d channel%s] --", $idx, $idx > 1 ? "s" : "");
    return(0);

  } elsif (uc($switch) eq "E") {
    &ds_notice($nick, "-- Listing exempted --");
    &ds_notice($nick, "[Hostmask                 ] [Server mask              ]");

    foreach (keys(%{$FXirc{'exempt'}})) { &ds_notice($nick, "[%-25s] [%-25s]", $FXirc{'exempt'}{$_}{'mask'}, $FXirc{'exempt'}{$_}{'server'}); }
    my $idx = scalar(keys(%{$FXirc{'exempt'}}));
    &ds_notice($nick, "-- End of list [%d exempt%s found] --", $idx, $idx > 1 ? "s" : "");
    return(0);

  } elsif (uc($switch) eq "I") {
    &ds_notice($nick, "-- Listing ignored --");
    &ds_notice($nick, "[Hostmask                 ] [Reason                   ]");

    foreach (keys(%{$FXirc{'ignore'}})) {
      my $reason = $FXirc{'ignore'}{$_}{'reason'};
      if (length($reason) > 22) { $reason = substr($reason, 0, 22) . "..."; }
      &ds_notice($nick, "[%-25s] [%-25s]", $FXirc{'ignore'}{$_}{'mask'}, $reason);
    }

    my $idx = scalar(keys(%{$FXirc{'ignore'}}));
    &ds_notice($nick, "-- End of list [%d ignore%s found] --", $idx, $idx > 1 ? "s" : "");
    return(0);

  } elsif (uc($switch) eq "O") {
    &ds_notice($nick, "-- Authorized users --");
    &ds_notice($nick, "[Nickname       ] [Hostname                                ] [Flags] [DCC]");

    foreach my $oper (keys(%FXoper)) {
      my $uname = $FXoper{$oper}{'name'};
      my $uhost = "";
      my $udcc = "";
      my $uflag = "";

      foreach my $client (keys(%{$FXirc{'client'}})) {
        my $unick = $FXirc{'client'}{$client}{'nick'};

        if (&valid_oper($FXirc{'client'}{lc($unick)}{'nick'}, undef) && ($FXirc{'client'}{lc($unick)}{'oper'} eq $FXoper{$oper}{'name'})) {
          $uname = $FXirc{'client'}{lc($unick)}{'nick'};
          $uhost = $FXirc{'client'}{lc($unick)}{'host'};

          foreach my $socket (keys(%{$FXsock{'list'}})) {
            if (($FXsock{'list'}{$socket}{'type'} eq "DCC") && ($FXsock{'list'}{$socket}{'oper'} eq $FXoper{$oper}{'name'})) { $udcc = " X "; }
          }
        }
      }

      foreach my $flag (keys(%{$FXflags{'oper'}})) {
        if ($FXoper{$oper}{'flags'} & $FXflags{'oper'}{$flag}) { $uflag .= $flag; }
      }

      &ds_notice($nick, "[%-15s] [%-40s] [%-5s] [%-3s]", $uname, $uhost, $uflag, $udcc);
    }

    my $idx = scalar(keys(%FXoper));
    &ds_notice($nick, "-- End of list [%d user%s] --", $idx, $idx > 1 ? "s" : "");
    return(0);

  } elsif (uc($switch) eq "U") {
    my $started = gmtime($FXserv{'started'});
    &ds_notice($nick, "Uptime: %s (started: %s)", &format_time($FXserv{'started'}), $started);
    return(0);

  } elsif (uc($switch) eq "K") {
    &ds_notice($nick, "-- Listing auto-kill --");
    &ds_notice($nick, "[Mask                     ] [Reason                   ] [Expire        ] [Author         ]");

    foreach (keys(%{$FXirc{'kill'}})) {
      my $reason = $FXirc{'kill'}{$_}{'reason'};
      if (length($reason) > 22) { $reason = substr($reason, 0, 22) . "..."; }
      &ds_notice($nick, "[%-25s] [%-25s] [%-14s] [%-15s]", $FXirc{'kill'}{$_}{'mask'}, $reason, $FXirc{'kill'}{$_}{'expire'} ? &format_time($FXirc{'kill'}{$_}{'set'} + $FXirc{'kill'}{$_}{'expire'}) : "never", $FXirc{'kill'}{$_}{'author'});
    }

    my $idx = scalar(keys(%{$FXirc{'kill'}}));
    &ds_notice($nick, "-- End of list [%d auto-kill%s found] --", $idx, $idx > 1 ? "s" : "");
    return(0);

  } elsif (uc($switch) eq "B") {
    &ds_notice($nick, "-- Listing bots --");
    &ds_notice($nick, "[Bot               ] [Hostname                      ] [Port ] [Link]");

    foreach (keys(%FXbnet)) {
      &ds_notice($nick, "[%-18s] [%-30s] [%-5s] [ %-2s ]", $FXbnet{$_}{'nick'}, $FXbnet{$_}{'host'}, $FXbnet{$_}{'port'}, $FXbnet{$_}{'prefix'});
    }

    my $idx = scalar(keys(%FXbnet));
    &ds_notice($nick, "-- End of list [%d bot%s found] --", $idx, $idx > 1 ? "s" : "");
    return(0);

  } else {
    &ds_notice($nick, $FXmsg{'unkn_switch'}, $switch);
    return(0);
  }
}

sub ds_status {
  my $nick = shift;

  if (!&valid_oper($nick, "o")) {
    &ds_notice($nick, $FXmsg{'no_priv'});
    return(0);
  }

  my $system = $^O . " " . $Config{'osvers'};
  my $started = gmtime($FXserv{'started'});

  &ds_notice($nick, "General status");

  &ds_notice($nick, "          Running on: %s (Perl %vd)", $system, $^V);
  &ds_notice($nick, "         Current hub: %s (connected for: %s)", $FXcfg{'hub'}{'host'}, $FXserv{'connected'} ? &format_time($FXserv{'connected'}) : "not connected");
  &ds_notice($nick, "Hub connect interval: %s", &format_time($FXcfg{'general'}{'reconnect'} + $ts));
  &ds_notice($nick, "   Bot link interval: %s", &format_time($FXcfg{'general'}{'reconnect'} + $ts));
  &ds_notice($nick, "          Started at: %s", $started);
  &ds_notice($nick, " ");
  &ds_notice($nick, "Security settings");
  &ds_notice($nick, "       Flood trigger: %d messages in %s", $FXcfg{'general'}{'flood_count'}, &format_time($FXcfg{'general'}{'flood_time'} + $ts));
  &ds_notice($nick, "         Ignore time: %s", &format_time($FXcfg{'general'}{'flood_ignore'} + $ts));
  &ds_notice($nick, "          Max clones: %d (autokill: %s)", $FXcfg{'general'}{'clones_max'}, $FXcfg{'general'}{'clones_kill'} ? "yes" : "no");

  if ($FXbot{'ScanServ'}{'enabled'}) {
    my $wingate = "";
    my $http = "";
    my $post = "";
    my $socks4 = "";
    my $socks5 = "";

    foreach (&parse_range($FXbot{'ScanServ'}{'scan_wingate'})) { $wingate .= $_ . " "; }
    &ds_notice($nick, "    Ports monitoring: Wingate: %s", $wingate);

    foreach (&parse_range($FXbot{'ScanServ'}{'scan_http'})) { $http .= $_ . " "; }
    &ds_notice($nick, "                      HTTP: %s", $http);

    foreach (&parse_range($FXbot{'ScanServ'}{'scan_post'})) { $post .= $_ . " "; }
    &ds_notice($nick, "                      POST: %s", $post);

    foreach (&parse_range($FXbot{'ScanServ'}{'scan_socks4'})) { $socks4 .= $_ . " "; }
    &ds_notice($nick, "                      Socks4: %s", $socks4);

    foreach (&parse_range($FXbot{'ScanServ'}{'scan_socks5'})) { $socks5 .= $_ . " "; }
    &ds_notice($nick, "                      Socks5: %s", $socks5);

    &ds_notice($nick, "  Simultaneous scans: %d", $FXbot{'ScanServ'}{'scan_max'});
    &ds_notice($nick, "  CTCP version check: %s", $FXbot{'ScanServ'}{'scan_version'} ? "yes" : "no");
  }

  &ds_notice($nick, " ");
  &ds_notice($nick, "Databases");
  &ds_notice($nick, "       Save interval: %s", &format_time($FXcfg{'general'}{'sync'} + $ts));

  if ($MySQL{'enabled'}) {
    &ds_notice($nick, "      MySQL interval: %s", &format_time($MySQL{'sync'} + $ts));
    &ds_notice($nick, "        MySQL server: %s:%d", $MySQL{'host'}, $MySQL{'port'});
  }

  &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "STATUS");
  return(0);
}

sub ds_save {
  my $nick = shift;

  if (!&valid_oper($nick, "a")) {
    &ds_notice($nick, "%s", $FXmsg{'no_priv'});
    return(0);
  }

  &save_conf();
  if ($MySQL{'enabled'}) { &save_mysql(); }

  &ds_notice($nick, $FXmsg{'db_saved'});

  &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "SAVE");
  return(0);
}

sub ds_global {
  my ($nick, $arg) = (shift, shift);

  if (!&valid_oper($nick, "a")) {
    &ds_notice($nick, $FXmsg{'no_priv'});
    return(0);
  }

  if (!$FXsock{'irc'}) {
    &ds_notice($nick, $FXmsg{'no_irc'});
    return(0);
  }

  my $target = (split(/ /, $arg, 0))[0]; $target = lc($target);
  my $global = (split(/ /, $arg, 2))[1];
  my $idx = 0;

  if (!$target || !$global) {
    &ds_notice($nick, $FXmsg{'syntax'}, "GLOBAL <target> <message>");
    return(0);
  }

  if ($target eq "-all") {
    foreach (keys(%{$FXirc{'client'}})) {
      if (lc($FXirc{'client'}{$_}{'server'}) eq lc($FXserv{'name'})) { next; }
      &ds_notice($FXirc{'client'}{$_}{'nick'}, $global);
      ++$idx;
    }
  } elsif ($target eq "-oper") {
    foreach (keys(%{$FXirc{'client'}})) {
      if (lc($FXirc{'client'}{$_}{'server'}) eq lc($FXserv{'name'})) { next; }
      if ($FXirc{'client'}{$_}{'mode'} & $FXflags{'client'}{'o'}) {
        &ds_notice($FXirc{'client'}{$_}{'nick'}, $global);
        ++$idx;
      }
    }
  } elsif ($target =~ /^-(chan|channel):(\S+)$/) {
    if (!$2) {
      &ds_notice($nick, $FXmsg{'syntax'}, "GLOBAL <target> <message>");
      return(0);
    } elsif (!$FXirc{'channel'}{lc($2)}) {
      &ds_notice($nick, $FXmsg{'no_chan'}, $2);
      return(0);
    }

    foreach (keys(%{$FXirc{'channel'}{lc($2)}{'names'}})) {
      my $user = $FXirc{'channel'}{lc($2)}{'names'}{$_}{'nick'};
      if (lc($FXirc{'client'}{lc($user)}{'server'}) eq lc($FXserv{'name'})) { next; }
      &ds_notice($user, $global);
      ++$idx;
    }

  } elsif ($target =~ /^-(serv|server):(\S+)$/) {
    if (!$2) {
      &ds_notice($nick, $FXmsg{'syntax'}, "GLOBAL <target> <message>");
      return(0);
    } elsif (!$FXirc{'server'}{lc($2)}) {
      &ds_notice($nick, $FXmsg{'no_server'}, $2);
      return(0);
    }

    foreach (keys(%{$FXirc{'client'}})) {
      if ((lc($FXirc{'client'}{$_}{'server'}) eq lc($2)) && (lc($FXserv{'name'}) ne lc($2))) {
        &ds_notice($FXirc{'client'}{$_}{'nick'}, $global);
        ++$idx;
      }
    }

  } else {
    &ds_notice($nick, $FXmsg{'syntax'}, "GLOBAL <target> <message>");
    return(0);
  }

  &ds_notice($nick, $FXmsg{'msg_sent'}, $idx);
  &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "GLOBAL %s %s", $target, $global);
  return(0);
}

sub ds_kill {
  my ($nick, $arg) = (shift, shift);
  my $count = $arg =~ tr/ / /; ++$count;

  if (!&valid_oper($nick, "k")) {
    &ds_notice($nick, $FXmsg{'no_priv'});
    return(0);
  }

  if (!$FXsock{'irc'}) {
    &ds_notice($nick, $FXmsg{'no_irc'});
    return(0);
  }

  if (!$arg) {
    &ds_notice($nick, $FXmsg{'syntax'}, "[time] <nick|hostmask> [reason]");
    return(0);
  }

  my $target = "";
  my $reason = "";
  my $expire = &format_timestr((split(/ /, $arg, 0))[0]);
  my $author = "";

  if (($count > 1) && ($expire =~ /\d+/) && ($expire > 0)) {
    $target = (split(/ /, $arg, 0))[1];
    $reason = (split(/ /, $arg, 3))[2];
  } else {
    $target = (split(/ /, $arg, 0))[0];
    $reason = (split(/ /, $arg, 2))[1];
    $expire = 0;
  }

  if ($FXsock{'list'}{$nick}) {
    $author = $FXsock{'list'}{$nick}{'nick'};
  } else {
    $author = $FXirc{'client'}{lc($nick)}{'nick'};
  }

  if ($FXirc{'client'}{lc($target)}) {
    if (lc($FXirc{'client'}{lc($target)}{'server'}) eq lc($FXserv{'name'})) {
      &ds_notice($nick, "Cannot kill Network Service Daemon [%s]", $FXirc{'client'}{lc($target)}{'nick'});
      return(0);
    }

    if (&valid_oper($target, "f")) {
      &ds_notice($nick, $FXmsg{'kill_friend'}, $FXirc{'client'}{lc($target)}{'nick'}, $FXirc{'client'}{lc($target)}{'user'}, $FXirc{'client'}{lc($target)}{'host'});
      &putlog("y", "%s attempted to KILL protected user %s!%s\@%s %s", $author, $FXirc{'client'}{lc($target)}{'nick'}, $FXirc{'client'}{lc($target)}{'user'}, $FXirc{'client'}{lc($target)}{'host'}, $reason ? "[" . $reason . "]" : "");
      return(0);
    }

    if ($expire) {
      &add_akill($FXirc{'client'}{lc($target)}{'user'}, $FXirc{'client'}{lc($target)}{'host'}, $author, $expire, $reason);

      &ds_notice($nick, $FXmsg{'autokill'}, $FXirc{'client'}{lc($target)}{'nick'}, &format_time($ts + $expire));
      &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "KILL [%s] %s [%s]", $expire, $target, $reason);
    } else {
      &ds_notice($nick, $FXmsg{'kill_nick'}, $FXirc{'client'}{lc($target)}{'nick'}, $FXirc{'client'}{lc($target)}{'user'}, $FXirc{'client'}{lc($target)}{'host'}, $reason);
      &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "KILL %s [%s]", $target, $reason);
    }

    &kill_client($target, $reason, $author, "DataServ");
    &add_seen($target, "QUIT", $reason);
    &chanlog("q", undef, $target, $reason, undef);
    &del_client($target);
    ++$FXstats{'kills'}{'oper'};

    return(0);
  }

  if ($target !~ "\@") {
    $target = "*\@" . $target;
  } else {
    if ((split(/\@/, $target, 0))[0] eq "") { $target = "*" . $target; }
  }

  my $idx = 0;

  foreach (split(//, $target)) {
    if (($_ eq "\@") || ($_ eq "*") || ($_ eq "?") || ($_ eq ".")) { next; }
    ++$idx;
  }

  if ($idx < $FXcfg{'general'}{'non_wildcard'}) {
    &ds_notice($nick, $FXmsg{'kill_bad'}, $target);
    return(0);
  }

  if ($FXirc{'kill'}{lc($target)}) {
    &ds_notice($nick, $FXmsg{'kill_in'}, $target);
    return(0);
  }

  my $usermask = (split(/\@/, $target, 0))[0]; my $umaskrx = &plain2regexp($usermask);
  my $hostmask = (split(/\@/, $target, 0))[1]; my $hmaskrx = &plain2regexp($hostmask);
  my $total = 0;

  foreach (keys(%{$FXirc{'client'}})) {
    if (($FXirc{'client'}{$_}{'user'} =~ /$umaskrx/i) && ($FXirc{'client'}{$_}{'host'} =~ /$hmaskrx/i)) {
      if (lc($FXirc{'client'}{$_}{'server'}) eq lc($FXserv{'name'})) { next; }
      if (&valid_oper($FXirc{'client'}{$_}{'nick'}, "f")) { next; }

      &add_seen($FXirc{'client'}{$_}{'nick'}, "QUIT", $reason);
      &kill_client($FXirc{'client'}{$_}{'nick'}, $reason, $author, "DataServ");
      &del_client($FXirc{'client'}{$_}{'nick'});
      ++$FXstats{'kills'}{'oper'};
      ++$total;
    }
  }

  if ($expire) {
    $FXirc{'kill'}{lc($target)}{'mask'} = $target;
    $FXirc{'kill'}{lc($target)}{'reason'} = $reason ? $reason : $FXcfg{'general'}{'kill_reason'};
    $FXirc{'kill'}{lc($target)}{'author'} = $author;
    $FXirc{'kill'}{lc($target)}{'set'} = $ts;
    $FXirc{'kill'}{lc($target)}{'expire'} = $expire;

    &add_akill($usermask, $hostmask, $author, $expire, $reason);

    &ds_notice($nick, $FXmsg{'autokill'}, &format_time($ts + $expire), $target, $total);
    &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "KILL [%s] %s [%s] (%d killed)", $expire, $target, $reason, $total);
  } else {
    &ds_notice($nick, $FXmsg{'kill_host'}, $total, $target);
    &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "KILL %s [%s] (%d killed)", $target, $reason, $total);
  }

  return(0);
}

sub ds_unkill {
  my ($nick, $mask) = (shift, shift);

  if (!&valid_oper($nick, "k")) {
    &ds_notice($nick, $FXmsg{'no_priv'});
    return(0);
  }

  if (!$mask) {
    &ds_notice($nick, $FXmsg{'syntax'}, "UNKILL <hostmask>");
    return(0);
  }

  $mask = (split(/ /, $mask, 0))[0];

  if (!$FXirc{'kill'}{lc($mask)}) {
    &ds_notice($nick, $FXmsg{'no_mask'}, $mask);
    return(0);
  }

  if ($FXirc{'kill'}{lc($mask)}{'expire'}) {
    my ($user, $host) = split(/\@/, $FXirc{'kill'}{lc($mask)}{'mask'}, 2);
    &rem_akill($user, $host);
  }

  delete($FXirc{'kill'}{lc($mask)});

  &ds_notice($nick, $FXmsg{'kill_del'}, $mask);
  &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "UNKILL %s", $mask);
  return(0);
}

sub ds_clones {
  my $nick = shift;

  if (!&valid_oper($nick, "o")) {
    &ds_notice($nick, $FXmsg{'no_priv'});
    return(0);
  }

  if (!$FXsock{'irc'}) {
    &ds_notice($nick, $FXmsg{'no_irc'});
    return(0);
  }

  my $idx = 0;
  my $clones = 0;
  my %list = ();

  &ds_notice($nick, "-- Listing Clones --");

  foreach (keys(%{$FXirc{'client'}})) {
    if ($FXirc{'client'}{$_}{'server'} ne $FXserv{'name'}) { push(@{$list{lc($FXirc{'client'}{$_}{'host'})}}, $FXirc{'client'}{$_}{'nick'}); }
  }

  foreach my $key (keys(%list)) {
    if (scalar(@{$list{$key}}) > 1) {
      ++$clones;

      &ds_notice($nick, "*\@%s", $key);
      foreach my $clone (@{$list{$key}}) {
        ++$idx;

        my $exempted = 0;

        foreach my $exempt (keys(%{$FXirc{'exempt'}})) {
          my $usermask = (split(/\@/, $FXirc{'exempt'}{$exempt}{'mask'}, 0))[0]; my $umaskrx = &plain2regexp($usermask);
          my $hostmask = (split(/\@/, $FXirc{'exempt'}{$exempt}{'mask'}, 0))[1]; my $hmaskrx = &plain2regexp($hostmask);
          my $servmask = &plain2regexp($FXirc{'exempt'}{$exempt}{'server'});

          if (($FXirc{'client'}{lc($clone)}{'user'} =~ /^$umaskrx$/i) && ($FXirc{'client'}{lc($clone)}{'host'} =~ /^$hmaskrx$/i) && ($FXirc{'client'}{lc($clone)}{'server'} =~ /^$servmask$/i)) {
            $exempted = 1;
            last;
          }
        }

        &ds_notice($nick, "  %02d) %-15s [%s] %s", $idx, $FXirc{'client'}{lc($clone)}{'nick'}, $FXirc{'client'}{lc($clone)}{'server'}, $exempted ? "(exempted)" : "");
      }

      $idx = 0;
    }
  }

  &ds_notice($nick, "-- End of list [%d clone%s found] --", $clones, $clones > 1 ? "s" : "");
  &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "CLONES");
  return(0);
}

sub ds_trace {
  my ($nick, $arg) = (shift, shift);

  if (!&valid_oper($nick, "o")) {
    &ds_notice($nick, $FXmsg{'no_priv'});
    return(0);
  }

  if (!$FXsock{'irc'}) {
    &ds_notice($nick, $FXmsg{'no_irc'});
    return(0);
  }

  my ($oper, $nonoper, $long, $kill) = (0, 0, 0, 0);
  my ($i, $j, $s) = (0, 0, 0);
  my $server = "";
  my $search = my $trace = (split(/ /, $arg, 0))[0];
  my $option = (split(/ /, $arg, 2))[1];

  if (!$search) {
    &ds_notice($nick, $FXmsg{'syntax'}, "TRACE <pattern> [options]");
    return(0);
  }

  if ($option) {
    foreach (split(/ /, $option)) {
      ++$i;

      if (lc($_) eq "-oper") {
        $oper = 1;
      } elsif (lc($_) eq "-nonoper") {
        $nonoper = 1;
      } elsif (lc($_) eq "-kill") {
        $kill = 1;

        my $idx = 0;

        foreach my $special (split(//, $search)) {
          if (($special eq "\@") || ($special eq "*") || ($special eq "?") || ($special eq ".")) { next; }
          ++$idx;
        }

        if ($idx < $FXcfg{'general'}{'non_wildcard'}) {
          &ds_notice($nick, $FXmsg{'kill_bad'}, $search);
          $kill = 0;
        }

      } elsif (lc($_) eq "-long") {
        $long = 1;
      } elsif (lc($_) eq "-server") {
        $s = $i + 1;
      } elsif ($s == $i) {
        $server = $_;
      }
    }

    if ($s && !$server) {
      &ds_notice($nick, $FXmsg{'syntax'}, "TRACE <pattern> [options]");
      return(0);
    } elsif ($server && !$FXirc{'server'}{lc($server)}) {
      &ds_notice($nick, $FXmsg{'no_server'}, $server);
      return(0);
    }
  }

  if ($trace !~ "\@") {
    $trace = "*\@" . $trace;
  } else {
    if ((split(/\@/, $trace, 0))[0] eq "") { $trace = "*" . $trace; }
  }

  my $usermask = (split(/\@/, $trace, 0))[0]; my $umaskrx = &plain2regexp($usermask);
  my $hostmask = (split(/\@/, $trace, 0))[1]; my $hmaskrx = &plain2regexp($hostmask);
  my $searchrx = &plain2regexp($search);

  &ds_notice($nick, "-- Listing results --");

  foreach my $key (keys(%{$FXirc{'client'}})) {
    my $client = $FXirc{'client'}{lc($key)}{'nick'};

    if (($client =~ /$searchrx/i) || (($FXirc{'client'}{lc($client)}{'user'} =~ /$umaskrx/i) && ($FXirc{'client'}{lc($client)}{'host'} =~ /$hmaskrx/i))) {
      if ($nonoper && ($FXirc{'client'}{lc($client)}{'mode'} & $FXflags{'client'}{'o'})) { next; }
      if ($oper && !($FXirc{'client'}{lc($client)}{'mode'} & $FXflags{'client'}{'o'})) { next; }
      if ($server && (lc($FXirc{'client'}{lc($client)}{'server'}) ne lc($server))) { next; }
      ++$j;

      if ($long) {
        my $modes = "";
        my $list = "";

        foreach my $flag (keys(%{$FXflags{'client'}})) {
          if ($FXirc{'client'}{lc($client)}{'mode'} & $FXflags{'client'}{$flag}) { $modes .= $flag; }
        }

        foreach my $chan (keys(%{$FXirc{'channel'}})) {
          my $channel = $FXirc{'channel'}{$chan}{'channel'};

          if ($FXirc{'channel'}{lc($channel)}{'names'}{lc($client)}) {
            if ($FXirc{'channel'}{lc($channel)}{'names'}{lc($client)}{'mode'} & $FXflags{'channel'}{'o'}) {
              $list .= "\@" . $channel . " ";
            } elsif ($FXirc{'channel'}{lc($channel)}{'names'}{lc($client)}{'mode'} & $FXflags{'channel'}{'h'}) {
              $list .= "\%" . $channel . " ";
            } elsif ($FXirc{'channel'}{lc($channel)}{'names'}{lc($client)}{'mode'} & $FXflags{'channel'}{'v'}) {
              $list .= "+" . $channel . " ";
            } else {
              $list .= $channel . " ";
            }
          }
        }

        $list =~ tr/ //s; $list =~ s/\s+$//;

        if ($j > 1) { &ds_notice($nick, " "); }
        &ds_notice($nick, "User:     %s (%s\@%s)", $client, $FXirc{'client'}{lc($client)}{'user'}, $FXirc{'client'}{lc($client)}{'host'});
        &ds_notice($nick, "Realname: %s", $FXirc{'client'}{lc($client)}{'info'});
        &ds_notice($nick, "Server:   %s", $FXirc{'client'}{lc($client)}{'server'});
        &ds_notice($nick, "Usermode: +%s", $modes);
        &ds_notice($nick, "Signon:   %s", &format_time($FXirc{'client'}{lc($client)}{'time'}));
        if ($list) { &ds_notice($nick, "Channel:  %s", $list); }
      } else {
        &ds_notice($nick, "%s!%s\@%s (%s) [%s]", $client, $FXirc{'client'}{lc($client)}{'user'}, $FXirc{'client'}{lc($client)}{'host'}, $FXirc{'client'}{lc($client)}{'info'}, $FXirc{'client'}{lc($client)}{'server'});
      }

      if ($kill) {
        my $author = "";

        if ($FXsock{'list'}{$nick}) { $author = $FXsock{'list'}{$nick}{'nick'}; }
        if ($FXirc{'client'}{lc($nick)}) { $author = $FXirc{'client'}{lc($nick)}{'nick'}; }

        &kill_client($client, $FXcfg{'general'}{'kill_reason'}, $author, $FXbot{'DataServ'}{'service'});
        &add_seen($client, "QUIT", $FXcfg{'general'}{'kill_reason'});
        &chanlog("q", undef, $client, $FXcfg{'general'}{'kill_reason'}, undef);
        &del_client($client);
        ++$FXstats{'kills'}{'oper'};
      }

    }
  }

  &ds_notice($nick, "-- End of list [%d match%s found] --", $j, $j > 1 ? "es" : "");
  &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "TRACE %s %s", $search, $option ? "[" . $option . "]" : "");

  return(0);
}

sub ds_exempt {
  my ($nick, $arg) = (shift, shift);

  if (!&valid_oper($nick, "o")) {
    &ds_notice($nick, $FXmsg{'no_priv'});
    return(0);
  }

  if (!$arg) {
    &ds_notice($nick, $FXmsg{'syntax'}, "EXEMPT <ADD|DEL|LIST> [mask] [server]");
    return(0);
  }

  my $what = (split(/ /, $arg, 0))[0];
  my $mask = (split(/ /, $arg, 0))[1];
  my $server = (split(/ /, $arg, 0))[2];
  if (!$server) { $server = "*"; }

  if (uc($what) eq "ADD") {
    if (!$mask) {
      &ds_notice($nick, $FXmsg{'syntax'}, "EXEMPT ADD <mask> [server]");
      return(0);
    }

    if ($FXirc{'exempt'}{lc($mask)}) {
      &ds_notice($nick, $FXmsg{'exempt_in'}, $mask, $server);
      return(0);
    }

    $FXirc{'exempt'}{lc($mask)}{'mask'} = $mask;
    $FXirc{'exempt'}{lc($mask)}{'server'} = $server;
    &ds_notice($nick, $FXmsg{'exempt_add'}, $mask, $server);
    &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "EXEMPT ADD %s [%s]", $mask, $server);
    return(0);

  } elsif (uc($what) eq "DEL") {
    if (!$mask) {
      &ds_notice($nick, $FXmsg{'syntax'}, "EXEMPT DEL <mask>");
      return(0);
    }

    if (!$FXirc{'exempt'}{lc($mask)}) {
      &ds_notice($nick, $FXmsg{'no_mask'}, $mask);
      return(0);
    }

    delete($FXirc{'exempt'}{lc($mask)});
    &ds_notice($nick, $FXmsg{'exempt_del'}, $mask);
    &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "EXEMPT DEL %s", $mask);
    return(0);

  } elsif (uc($what) eq "LIST") {
    &ds_notice($nick, "-- Listing Exempted --");
    &ds_notice($nick, "[Hostmask                 ] [Server mask              ]");

    foreach (keys(%{$FXirc{'exempt'}})) { &ds_notice($nick, "[%-25s] [%-25s]", $FXirc{'exempt'}{$_}{'mask'}, $FXirc{'exempt'}{$_}{'server'}); }
    my $idx = scalar(keys(%{$FXirc{'exempt'}}));
    &ds_notice($nick, "-- End of list [%d exempt%s found] --", $idx, $idx > 1 ? "s" : "");
    &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "EXEMPT LIST");
    return(0);

  } else {
    &ds_notice($nick, $FXmsg{'syntax'}, "EXEMPT <ADD|DEL|LIST> [mask] [server]");
    return(0);
  }
}

sub ds_ignore {
  my ($nick, $arg) = (shift, shift);

  if (!&valid_oper($nick, "o")) {
    &ds_notice($nick, $FXmsg{'no_priv'});
    return(0);
  }

  if (!$arg) {
    &ds_notice($nick, $FXmsg{'syntax'}, "IGNORE <ADD|DEL|LIST> [mask] [reason]");
    return(0);
  }

  my $what = (split(/ /, $arg, 0))[0];
  my $mask = (split(/ /, $arg, 0))[1];
  my $reason = (split(/ /, $arg, 0))[2];

  if (uc($what) eq "ADD") {
    if (!$mask) {
      &ds_notice($nick, $FXmsg{'syntax'}, "IGNORE ADD <mask> [reason]");
      return(0);
    }

    if ($FXirc{'ignore'}{lc($mask)}) {
      &ds_notice($nick, $FXmsg{'ignore_in'}, $mask, $reason);
      return(0);
    }

    $FXirc{'ignore'}{lc($mask)}{'mask'} = $mask;
    $FXirc{'ignore'}{lc($mask)}{'reason'} = $reason;
    &ds_notice($nick, $FXmsg{'ignore_add'}, $mask, $reason);
    &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "IGNORE ADD %s [%s]", $mask, $reason);
    return(0);

  } elsif (uc($what) eq "DEL") {
    if (!$mask) {
      &ds_notice($nick, $FXmsg{'syntax'}, "IGNORE DEL <mask>");
      return(0);
    }

    if (!$FXirc{'ignore'}{lc($mask)}) {
      &ds_notice($nick, $FXmsg{'no_mask'}, $mask);
      return(0);
    }

    delete($FXirc{'ignore'}{lc($mask)});
    &ds_notice($nick, $FXmsg{'ignore_del'}, $mask);
    &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "IGNORE DEL %s", $mask);
    return(0);

  } elsif (uc($what) eq "LIST") {
    &ds_notice($nick, "-- Listing Ignored --");
    &ds_notice($nick, "[Hostmask                 ] [Reason                   ]");

    foreach (keys(%{$FXirc{'ignore'}})) {
      my $reason = $FXirc{'ignore'}{$_}{'reason'};
      if (length($reason) > 22) { $reason = substr($reason, 0, 22) . "..."; }
      &ds_notice($nick, "[%-25s] [%-25s]", $FXirc{'ignore'}{$_}{'mask'}, $reason);
    }

    my $idx = scalar(keys(%{$FXirc{'ignore'}}));
    &ds_notice($nick, "-- End of list [%d ignore%s found] --", $idx, $idx > 1 ? "s" : "");
    &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "IGNORE LIST");
    return(0);

  } else {
    &ds_notice($nick, $FXmsg{'syntax'}, "IGNORE <ADD|DEL|LIST> [mask] [reason]");
    return(0);
  }
}

sub ds_seen {
  my ($nick, $search) = (shift, shift);
  my $idx = 0;
  my @seen = ();
  my @list = ();
  my $format = "";

  if (!$search) {
    &ds_notice($nick, $FXmsg{'syntax'}, "SEEN <nick|hostmask>");
    return(0);
  }

  my $trace = $search = (split(/ /, $search, 0))[0];

  if ($trace !~ "\@") {
    $trace = "*\@" . $trace;
  } else {
    if ((split(/\@/, $trace, 0))[0] eq "") { $trace = "*" . $trace; }
  }

  my $usermask = (split(/\@/, $trace, 0))[0]; my $umaskrx = &plain2regexp($usermask);
  my $hostmask = (split(/\@/, $trace, 0))[1]; my $hmaskrx = &plain2regexp($hostmask);
  my $searchrx = &plain2regexp($search);

  if (($search =~ /\*/) || ($search =~ /\?/)) {
    foreach (keys(%{$FXirc{'seen'}})) {
      if (($FXirc{'seen'}{$_}{'nick'} =~ /$searchrx/i) || (($FXirc{'seen'}{$_}{'user'} =~ /$umaskrx/i) && ($FXirc{'seen'}{$_}{'host'} =~ /$hmaskrx/i)) || (($FXirc{'seen'}{$_}{'action'} eq "NICK") && ($FXirc{'seen'}{$_}{'xtra'} =~ /$searchrx/i))) {
        ++$idx;
        push(@seen, $FXirc{'seen'}{$_}{'nick'});
      }
    }

    if (!$idx) {
      &ds_notice($nick, $FXmsg{'seen_norec'}, $search);
      return(0);
    }

    if ($idx > $FXbot{'DataServ'}{'seen_maxwild'}) {
      &ds_notice($nick, $FXmsg{'seen_refine'}, $FXbot{'DataServ'}{'seen_maxwild'}, $search);
      return(0);
    }

    if ($idx > $FXbot{'DataServ'}{'seen_maxprint'}) {
      &ds_notice($nick, $FXmsg{'seen_recent'}, $idx, $search);
    } else {
      &ds_notice($nick, $FXmsg{'seen_print'}, $idx, $search);
    }

    $idx = 0;
    @list = sort { $FXirc{'seen'}{$b}{'time'} <=> $FXirc{'seen'}{$a}{'time'} } @seen;

    foreach my $user (@list) {
      if ($FXirc{'seen'}{lc($user)}{'action'} eq "NICK") {
        $format = $FXmsg{'seen_nick'};
        if ($FXbot{'DataServ'}{'seen_verbose'}) { $format = $FXmsg{'seen_vnick'}; }
      } elsif ($FXirc{'seen'}{lc($user)}{'action'} eq "JOIN") {
        $format = $FXmsg{'seen_join'};
        if ($FXbot{'DataServ'}{'seen_verbose'}) { $format = $FXmsg{'seen_vjoin'}; }
      } elsif ($FXirc{'seen'}{lc($user)}{'action'} eq "PART") {
        $format = $FXmsg{'seen_part'};
        if ($FXbot{'DataServ'}{'seen_verbose'}) { $format = $FXmsg{'seen_vpart'}; }
      } elsif ($FXirc{'seen'}{lc($user)}{'action'} eq "KICK") {
        $format = $FXmsg{'seen_kick'};
        if ($FXbot{'DataServ'}{'seen_verbose'}) { $format = $FXmsg{'seen_vkick'}; }
      } elsif ($FXirc{'seen'}{lc($user)}{'action'} eq "QUIT") {
        $format = $FXmsg{'seen_quit'};
        if ($FXbot{'DataServ'}{'seen_verbose'}) { $format = $FXmsg{'seen_vquit'}; }
      }

      if ($FXbot{'DataServ'}{'seen_verbose'}) {
        &ds_notice($nick, "  %d) $format", ++$idx, $FXirc{'seen'}{lc($user)}{'nick'}, $FXirc{'seen'}{lc($user)}{'user'}, $FXirc{'seen'}{lc($user)}{'host'}, $FXirc{'seen'}{lc($user)}{'xtra'}, &format_time($FXirc{'seen'}{lc($user)}{'time'}));
      } else {
        &ds_notice($nick, "  %d) $format", ++$idx, $FXirc{'seen'}{lc($user)}{'nick'}, $FXirc{'seen'}{lc($user)}{'xtra'}, &format_time($FXirc{'seen'}{lc($user)}{'time'}));
      }

      if ($idx >= $FXbot{'DataServ'}{'seen_maxprint'}) { last; }
    }

  } else {
    if ($FXirc{'client'}{lc($search)}) {
      &ds_notice($nick, $FXmsg{'seen_online'}, $FXirc{'client'}{lc($search)}{'nick'});
    } elsif ($FXirc{'seen'}{lc($search)}) {
      if ($FXirc{'seen'}{lc($search)}{'action'} eq "NICK") {
        $format = $FXmsg{'seen_nick'};
        if ($FXbot{'DataServ'}{'seen_verbose'}) { $format = $FXmsg{'seen_vnick'}; }
      } elsif ($FXirc{'seen'}{lc($search)}{'action'} eq "JOIN") {
        $format = $FXmsg{'seen_join'};
        if ($FXbot{'DataServ'}{'seen_verbose'}) { $format = $FXmsg{'seen_vjoin'}; }
      } elsif ($FXirc{'seen'}{lc($search)}{'action'} eq "PART") {
        $format = $FXmsg{'seen_part'};
        if ($FXbot{'DataServ'}{'seen_verbose'}) { $format = $FXmsg{'seen_vpart'}; }
      } elsif ($FXirc{'seen'}{lc($search)}{'action'} eq "KICK") {
        $format = $FXmsg{'seen_kick'};
        if ($FXbot{'DataServ'}{'seen_verbose'}) { $format = $FXmsg{'seen_vkick'}; }
      } elsif ($FXirc{'seen'}{lc($search)}{'action'} eq "QUIT") {
        $format = $FXmsg{'seen_quit'};
        if ($FXbot{'DataServ'}{'seen_verbose'}) { $format = $FXmsg{'seen_vquit'}; }
      }

      if ($FXbot{'DataServ'}{'seen_verbose'}) {
        &ds_notice($nick, $format, $FXirc{'seen'}{lc($search)}{'nick'}, $FXirc{'seen'}{lc($search)}{'user'}, $FXirc{'seen'}{lc($search)}{'host'}, $FXirc{'seen'}{lc($search)}{'xtra'}, &format_time($FXirc{'seen'}{lc($search)}{'time'}));
      } else {
        &ds_notice($nick, $format, $FXirc{'seen'}{lc($search)}{'nick'}, $FXirc{'seen'}{lc($search)}{'xtra'}, &format_time($FXirc{'seen'}{lc($search)}{'time'}));
      }

    } else {
      &ds_notice($nick, $FXmsg{'seen_norec'}, $search);
    }
  }

  &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "SEEN %s", $search);
  return(0);
}

sub ds_unseen {
  my ($nick, $search) = (shift, shift);
  my $idx = 0;

  if (!&valid_oper($nick, "a")) {
    &ds_notice($nick, $FXmsg{'denied'});
    return(0);
  }

  if (!$search) {
    &ds_notice($nick, $FXmsg{'syntax'}, "UNSEEN <nick|hostmask>");
    return(0);
  }

  my $trace = $search = (split(/ /, $search, 0))[0];

  if ($trace !~ "\@") {
    $trace = "*\@" . $trace;
  } else {
    if ((split(/\@/, $trace, 0))[0] eq "") { $trace = "*" . $trace; }
  }

  my $usermask = (split(/\@/, $trace, 0))[0]; my $umaskrx = &plain2regexp($usermask);
  my $hostmask = (split(/\@/, $trace, 0))[1]; my $hmaskrx = &plain2regexp($hostmask);
  my $searchrx = &plain2regexp($search);

  if (($search =~ /\*/) || ($search =~ /\?/)) {
    foreach (keys(%{$FXirc{'seen'}})) {
      if (($FXirc{'seen'}{$_}{'nick'} =~ /$searchrx/i) || (($FXirc{'seen'}{$_}{'user'} =~ /$umaskrx/i) && ($FXirc{'seen'}{$_}{'host'} =~ /$hmaskrx/i)) || (($FXirc{'seen'}{$_}{'action'} eq "NICK") && ($FXirc{'seen'}{$_}{'xtra'} =~ /$searchrx/i))) {
        delete($FXirc{'seen'}{$_});
        ++$idx;
      }
    }

    if (!$idx) {
      &ds_notice($nick, $FXmsg{'no_match'}, $search);
      return(0);
    }
  } else {
    if ($FXirc{'seen'}{lc($search)}) {
      delete($FXirc{'seen'}{lc($search)});
      ++$idx;
    } else {
      &ds_notice($nick, $FXmsg{'no_match'}, $search);
      return(0);
    }
  }

  &ds_notice($nick, $FXmsg{'seen_del'}, $idx, $search);
  &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "UNSEEN %s", $search);
  return(0);
}

sub ds_seenstats {
  my $nick = shift;

  if (!&valid_oper($nick, "a")) {
    &ds_notice($nick, $FXmsg{'denied'});
    return(0);
  }

  my ($nicks, $joins, $parts, $kicks, $quits) = (0, 0, 0, 0, 0);

  foreach (keys(%{$FXirc{'seen'}})) {
    if ($FXirc{'seen'}{$_}{'action'} eq "NICK") {
      ++$nicks;
    } elsif ($FXirc{'seen'}{$_}{'action'} eq "JOIN") {
      ++$joins;
    } elsif ($FXirc{'seen'}{$_}{'action'} eq "PART") {
      ++$parts;
    } elsif ($FXirc{'seen'}{$_}{'action'} eq "KICK") {
      ++$kicks;
    } elsif ($FXirc{'seen'}{$_}{'action'} eq "QUIT") {
      ++$quits;
    }
  }

  &ds_notice($nick, "Total number of seen entries: %d", scalar(keys(%{$FXirc{'seen'}})));
  &ds_notice($nick, "  NICK: %d", $nicks);
  &ds_notice($nick, "  JOIN: %d", $joins);
  &ds_notice($nick, "  PART: %d", $parts);
  &ds_notice($nick, "  KICK: %d", $kicks);
  &ds_notice($nick, "  QUIT: %d", $quits);

  &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "SEENSTATS");
  return(0);
}

sub ds_disable {
  my ($nick, $service) = (shift, shift);

  if (!&valid_oper($nick, "a")) {
    &ds_notice($nick, $FXmsg{'denied'});
    return(0);
  }

  if (!$service) {
    &ds_notice($nick, $FXmsg{'syntax'}, "DISABLE <service>");
    return(0);
  }

  foreach (keys(%FXbot)) {
    if (lc($service) eq lc($FXbot{$_}{'nick'})) {
      if ($FXbot{$_}{'service'} eq "DataServ") {
        &ds_notice($nick, "This service must remain enabled");
        return(0);
      }

      if (!$FXbot{$_}{'enabled'}) {
        &ds_notice($nick, "This service is already disabled");
        return(0);
      }

      &disable_service($FXbot{$_}{'service'});
      &ds_notice($nick, "Service disabled");
      &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "DISABLE %s", $FXbot{$_}{'nick'});
      return(0);
    }
  }

  &ds_notice($nick, $FXmsg{'no_service'}, $service);
  return(0);
}

sub ds_enable {
  my ($nick, $service) = (shift, shift);

  if (!&valid_oper($nick, "a")) {
    &ds_notice($nick, $FXmsg{'denied'});
    return(0);
  }

  if (!$service) {
    &ds_notice($nick, $FXmsg{'syntax'}, "ENABLE <service>");
    return(0);
  }

  foreach (keys(%FXbot)) {
    if (lc($service) eq lc($FXbot{$_}{'nick'})) {
      if ($FXbot{$_}{'enabled'}) {
        &ds_notice($nick, "This service is already enabled");
        return(0);
      }

      &enable_service($FXbot{$_}{'service'}, 0);
      &ds_notice($nick, "Service enabled");
      &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "ENABLE %s", $FXbot{$_}{'nick'});
      return(0);
    }
  }

  &ds_notice($nick, $FXmsg{'no_service'}, $service);
  return(0);
}

sub ds_monitor {
  my ($nick, $arg) = (shift, shift);

  if (!&valid_oper($nick, "a")) {
    &ds_notice($nick, $FXmsg{'no_priv'});
    return(0);
  }

  if (!$arg) {
    &ds_notice($nick, $FXmsg{'syntax'}, "MONITOR <ADD|DEL|SET|LIST> [service] [channel] [option] [parameter]");
    return(0);
  }

  my $what = (split(/ /, $arg, 0))[0];
  my $service = (split(/ /, $arg, 0))[1];
  my $channel = (split(/ /, $arg, 0))[2];
  my $bot = "";

  if (uc($what) eq "ADD") {
    if (!$service || !$channel || ($channel !~ /^#/)) {
      &ds_notice($nick, $FXmsg{'syntax'}, "MONITOR ADD <service> <channel>");
      return(0);
    }

    foreach (keys(%FXbot)) {
      if (lc($FXbot{$_}{'nick'}) eq lc($service)) {
        $bot = $FXbot{$_}{'service'};

        if (!$FXbot{$_}{'enabled'}) {
          &ds_notice($nick, $FXmsg{'service_disabled'}, $FXbot{$_}{'nick'});
          return(0);
        }

        if ($FXbot{$_}{'channel'} && $FXbot{$_}{'channel'}{lc($channel)}) {
          &ds_notice($nick, $FXmsg{'monitor_in'}, $channel, $FXbot{$_}{'nick'});
          return(0);
        }

      }
    }

    if (!$bot) {
      &ds_notice($nick, $FXmsg{'unkn_service'}, $service);
      return(0);
    }

    $FXbot{$bot}{'channel'}{lc($channel)}{'channel'} = $channel;
    $FXbot{$bot}{'channel'}{lc($channel)}{'service'} = $bot;
    $FXbot{$bot}{'channel'}{lc($channel)}{'mode'} = "";
    $FXbot{$bot}{'channel'}{lc($channel)}{'public'} = 0;
    $FXbot{$bot}{'channel'}{lc($channel)}{'prefix'} = "";
    $FXbot{$bot}{'channel'}{lc($channel)}{'log_event'} = '';
    $FXbot{$bot}{'channel'}{lc($channel)}{'log_file'} = '';
    $FXbot{$bot}{'channel'}{lc($channel)}{'joinmsg'} = "";
    $FXbot{$bot}{'channel'}{lc($channel)}{'partmsg'} = "";

    &ds_notice($nick, $FXmsg{'monitor_add'}, $channel, $FXbot{$bot}{'nick'});

    if ($FXsock{'irc'}) {
      &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "MONITOR ADD %s %s", $FXbot{$bot}{'nick'}, $channel);
      &service_join($bot, $channel);
    }

    return(0);

  } elsif (uc($what) eq "DEL") {
    if (!$service || !$channel || ($channel !~ /^#/)) {
      &ds_notice($nick, $FXmsg{'syntax'}, "MONITOR DEL <service> <channel>");
      return(0);
    }

    my $deleted = 0;

    foreach (keys(%FXbot)) {
      if (lc($FXbot{$_}{'nick'}) eq lc($service)) {
        $bot = $FXbot{$_}{'service'};

        if (!$FXbot{$_}{'enabled'}) {
          &ds_notice($nick, $FXmsg{'service_disabled'}, $FXbot{$_}{'nick'});
          return(0);
        }

        if ($FXbot{$_}{'channel'} && $FXbot{$_}{'channel'}{lc($channel)}) {
          delete($FXbot{$_}{'channel'}{lc($channel)});

          if ($FXsock{'irc'}) { &service_part($bot, $channel); }

          &ds_notice($nick, $FXmsg{'monitor_del'}, $channel, $FXbot{$bot}{'nick'});
          &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "MONITOR DEL %s %s", $FXbot{$bot}{'nick'}, $channel);
          $deleted = 1;
          return(0);
        }

      }
    }

    if (!$bot) {
      &ds_notice($nick, $FXmsg{'unkn_service'}, $service);
      return(0);
    }

    if (!$deleted) {
      &ds_notice($nick, $FXmsg{'no_monitor'}, $channel, $FXbot{$bot}{'nick'});
      return(0);
    }

  } elsif (uc($what) eq "SET") {
    my $option = (split(/ /, $arg, 0))[3];

    if (!$service || !$channel || ($channel !~ /^#/) || !$option) {
      &ds_notice($nick, $FXmsg{'syntax'}, "MONITOR SET <service> <channel> <option> <parameter>");
      return(0);
    }

    foreach (keys(%FXbot)) {
      if (lc($FXbot{$_}{'nick'}) eq lc($service)) {
        $bot = $FXbot{$_}{'service'};

        if (!$FXbot{$_}{'enabled'}) {
          &ds_notice($nick, $FXmsg{'service_disabled'}, $FXbot{$_}{'nick'});
          return(0);
        }

        if (!$FXbot{$_}{'channel'} || ($FXbot{$_}{'channel'} && !$FXbot{$_}{'channel'}{lc($channel)})) {
          &ds_notice($nick, $FXmsg{'no_monitor'}, $channel, $FXbot{$_}{'nick'});
          return(0);
        }

        my $param = (split(/ /, $arg, 5))[4];

        if (uc($option) eq "PUBLIC") {
          my $public = 0;
          if ($param) { $public = 1; }
          $FXbot{$_}{'channel'}{lc($channel)}{'public'} = $public;
          if (!$FXbot{$_}{'channel'}{lc($channel)}{'prefix'}) { $FXbot{$_}{'channel'}{lc($channel)}{'prefix'} = "!"; }

          &ds_notice($nick, $FXmsg{'monitor_set'}, "PUBLIC", $channel, $FXbot{$bot}{'nick'}, $public);
          &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "MONITOR SET %s %s PUBLIC %s", $FXbot{$bot}{'nick'}, $channel, $public);
          return(0);
        }

        if (uc($option) eq "PREFIX") {
          $FXbot{$_}{'channel'}{lc($channel)}{'prefix'} = $param;
          &ds_notice($nick, $FXmsg{'monitor_set'}, "PREFIX", $channel, $FXbot{$bot}{'nick'}, $param);
          &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "MONITOR SET %s %s PREFIX %s", $FXbot{$bot}{'nick'}, $channel, $param);
          return(0);
        }

        if (uc($option) eq "JOINMSG") {
          $FXbot{$_}{'channel'}{lc($channel)}{'joinmsg'} = $param;
          &ds_notice($nick, $FXmsg{'monitor_set'}, "JOINMSG", $channel, $FXbot{$bot}{'nick'}, $param);
          &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "MONITOR SET %s %s JOINMSG %s", $FXbot{$bot}{'nick'}, $channel, $param);
          return(0);
        }

        if (uc($option) eq "PARTMSG") {
          $FXbot{$_}{'channel'}{lc($channel)}{'partmsg'} = $param;
          &ds_notice($nick, $FXmsg{'monitor_set'}, "PARTMSG", $channel, $FXbot{$bot}{'nick'}, $param);
          &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "MONITOR SET %s %s PARTMSG %s", $FXbot{$bot}{'nick'}, $channel, $param);
          return(0);
        }

        if (uc($option) eq "LOG_FILE") {
          $FXbot{$_}{'channel'}{lc($channel)}{'log_file'} = $param;
          &ds_notice($nick, $FXmsg{'monitor_set'}, "LOG_FILE", $channel, $FXbot{$bot}{'nick'}, $param);
          &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "MONITOR SET %s %s LOG_FILE %s", $FXbot{$bot}{'nick'}, $channel, $param);
          return(0);
        }

        if (uc($option) eq "LOG_EVENT") {
          my $event = "";
          my $what = 1;

          foreach my $flag (split(//, $param)) {
            if ($flag eq "+") {
              $what = 1;
            } elsif ($flag eq "-") {
              $what = 0;
            } elsif ($FXflags{'monitor'}{$flag}) {
              if ($what) {
                $FXbot{$_}{'channel'}{lc($channel)}{'log_event'} |= $FXflags{'monitor'}{$flag};
              } else {
                $FXbot{$_}{'channel'}{lc($channel)}{'log_event'} &= ~$FXflags{'monitor'}{$flag};
              }
            }
          }

          foreach my $flag (keys(%{$FXflags{'monitor'}})) {
            if ($FXbot{$_}{'channel'}{lc($channel)}{'log_event'} & $FXflags{'monitor'}{$flag}) { $event .= $flag; }
          }

          &ds_notice($nick, $FXmsg{'monitor_set'}, "LOG_EVENT", $channel, $FXbot{$bot}{'nick'}, $event);
          &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "MONITOR SET %s %s LOG_EVENT %s", $FXbot{$bot}{'nick'}, $channel, $event);
          return(0);
        }

        if (uc($option) eq "MODE") {
          if (($param ne "o") && ($param ne "h") && ($param ne "v"))  {
            &ds_notice($nick, $FXmsg{'syntax'}, "MONITOR SET <service> <channel> MODE <o/h/v>");
            return(0);
          }

          $FXbot{$_}{'channel'}{lc($channel)}{'mode'} = $param;
          &ds_notice($nick, $FXmsg{'monitor_set'}, "MODE", $channel, $FXbot{$bot}{'nick'}, $param);
          &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "MONITOR SET %s %s MODE %s", $FXbot{$bot}{'nick'}, $channel, $param);

          if ((($FXbot{$bot}{'channel'}{lc($channel)}{'mode'} eq "o") && !($FXirc{'channel'}{lc($channel)}{'names'}{lc($nick)}{'mode'} & $FXflags{'channel'}{'o'})) ||
              (($FXbot{$bot}{'channel'}{lc($channel)}{'mode'} eq "h") && !($FXirc{'channel'}{lc($channel)}{'names'}{lc($nick)}{'mode'} & $FXflags{'channel'}{'h'})) ||
              (($FXbot{$bot}{'channel'}{lc($channel)}{'mode'} eq "v") && !($FXirc{'channel'}{lc($channel)}{'names'}{lc($nick)}{'mode'} & $FXflags{'channel'}{'v'}))) {
            &service_part($bot, $FXirc{'channel'}{lc($channel)}{'channel'});
            &service_join($bot, $FXirc{'channel'}{lc($channel)}{'channel'});
          }

          return(0);
        }

        &ds_notice($nick, $FXmsg{'unkn_option'}, $option);
        return(0);
      }
    }

    if (!$bot) {
      &ds_notice($nick, $FXmsg{'unkn_service'}, $service);
      return(0);
    }

  } elsif (uc($what) eq "LIST") {
    my $idx = 0;

    &ds_notice($nick, "-- Monitored Channels --");
    &ds_notice($nick, "[Service     ] [Channel             ] [Log] [Pub]");

    foreach my $service (keys(%FXbot)) {
      if (!$FXbot{$service}{'enabled'}) { next; }
      if (!$FXbot{$service}{'channel'}) { next; }

      $idx += scalar(keys(%{$FXbot{$service}{'channel'}}));
      foreach my $chan (keys(%{$FXbot{$service}{'channel'}})) {
        my $channel = $FXbot{$service}{'channel'}{$chan}{'channel'};
        my $mode = "";

        if ($FXbot{$service}{'channel'}{lc($channel)}{'mode'} eq "o") {
          $mode = "\@";
        } elsif ($FXbot{$service}{'channel'}{lc($channel)}{'mode'} eq "h") {
          $mode = "\%";
        } elsif ($FXbot{$service}{'channel'}{lc($channel)}{'mode'} eq "v") {
          $mode = "+";
        }

        &ds_notice($nick, "[%-12s] [%-20s] [%-3s] [%-3s]", $FXbot{$service}{'nick'}, $mode . $FXbot{$service}{'channel'}{lc($channel)}{'channel'}, $FXbot{$service}{'channel'}{lc($channel)}{'log_file'} ? "YES" : "NO", $FXbot{$service}{'channel'}{lc($channel)}{'public'} ? "YES" : "NO");
      }
    }

    &ds_notice($nick, "-- End of list [%d channel%s] --", $idx, $idx > 1 ? "s" : "");
    &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "MONITOR LIST");
    return(0);

  } else {
    &ds_notice($nick, $FXmsg{'syntax'}, "MONITOR <ADD|DEL|LIST|SET> [service] [channel] [option] [parameter]");
    return(0);
  }
}

sub ds_public {
  my ($nick, $channel, $cmd, $query) = (shift, shift, shift, shift);

  if (!&valid_oper($nick, "o")) {
    &ds_notice($nick, $FXmsg{'denied'});
    return(0);
  }

  $cmd = uc($cmd);
  my $arg = (split(/ /, $query, 2))[1];

  my $what = "";
  my $mode = "";

  if ($cmd =~ /^(OP|HALFOP|HOP|VOICE|DEOP|DEHALFOP|DEHOP|DEVOICE)$/) {
    if ($cmd =~ /^(OP|HALFOP|HOP|VOICE)$/) { $what = "+"; }
    if ($cmd =~ /^(DEOP|DEHALFOP|DEHOP|DEVOICE)$/) { $what = "-"; }

    if ($cmd =~ /^(OP|DEOP)$/) { $mode = "o"; }
    if ($cmd =~ /^(HALFOP|HOP|DEHALFOP|DEHOP)$/) { $mode = "h"; }
    if ($cmd =~ /^(VOICE|DEVOICE)$/) { $mode = "v"; }

    if (!$arg) {
      &putsock($FXsock{'irc'}, ":%s MODE %s %s%s %s", $FXbot{'DataServ'}{'nick'}, $channel, $what, $mode, $nick);
      &channel_mode($FXbot{'DataServ'}{'nick'}, $channel, $what . $mode, $nick);
      &chanlog("m", $channel, $FXbot{'DataServ'}{'nick'}, $what . $mode, $nick);
      &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "MODE on %s: %s%s %s", $channel, $what, $mode, $nick);
      return(0);
    }

    my $inc = 0;
    my $cnt = 0;
    my $exec = $what;
    my $target = "";
    my $who = "";
    my @user = ();

    foreach (split(/ /, $arg)) {
      if ($FXirc{'channel'}{lc($channel)} && $FXirc{'channel'}{lc($channel)}{'names'}{lc($_)}) { push(@user, $_); }
    }

    foreach (@user) {
      ++$inc;

      $who .= $_;
      $cnt = $who =~ tr/ / /; ++$cnt;
      $exec .= $mode;
      $target .= $_ . " ";

      if (($#user + 1) == $inc) {
        $who =~ s/\s+$//;
        &putsock($FXsock{'irc'}, ":%s MODE %s %s%s %s", $FXbot{'DataServ'}{'nick'}, $channel, $what, $mode x $cnt, $who);
      } elsif (!($cnt % $FXcfg{'general'}{'modes'})) {
        &putsock($FXsock{'irc'}, ":%s MODE %s %s%s %s", $FXbot{'DataServ'}{'nick'}, $channel, $what, $mode x $FXcfg{'general'}{'modes'}, $who);
        $who = "";
      } else {
        $who .= " ";
      }
    }

    if ($target) {
      $target =~ s/\s+$//;

      &chanlog("m", $channel, $FXbot{'DataServ'}{'nick'}, $exec, $target);
      &channel_mode($FXbot{'DataServ'}{'nick'}, $channel, $exec, $target);
      &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "MODE on %s: %s %s", $channel, $exec, $target);
    }

    return(0);

  } elsif ($cmd eq "KICK") {
    my $target = (split(/ /, $arg, 0))[0];
    my $reason = (split(/ /, $arg, 2))[1];

    if (!$FXirc{'client'}{lc($target)} || ($FXirc{'client'}{lc($target)} && !$FXirc{'channel'}{lc($channel)}{'names'}{lc($target)})) {
      &ds_notice($nick, $FXmsg{'no_user'}, $target);
      return(0);
    }

    if (lc($FXirc{'client'}{lc($target)}{'server'}) eq lc($FXserv{'name'})) {
      &ds_notice($nick, "Cannot kick Network Service Daemon [%s]", $FXirc{'client'}{lc($target)}{'nick'});
      return(0);
    }

    if (&valid_oper($target, "f")) {
      &ds_notice($nick, $FXmsg{'kill_friend'}, $FXirc{'client'}{lc($target)}{'nick'}, $FXirc{'client'}{lc($target)}{'user'}, $FXirc{'client'}{lc($target)}{'host'});
      &putlog("y", "%s attempted to KICK protected user %s!%s\@%s on %s", $nick, $FXirc{'client'}{lc($target)}{'nick'}, $FXirc{'client'}{lc($target)}{'user'}, $FXirc{'client'}{lc($target)}{'host'}, $channel);
      return(0);
    }

    &putsock($FXsock{'irc'}, ":%s KICK %s %s :%s", $FXbot{'DataServ'}{'nick'}, $channel, $FXirc{'client'}{lc($target)}{'nick'}, $reason);
    &add_seen($FXirc{'client'}{lc($target)}{'nick'}, "KICK", $channel);
    &putinfo("k", "Channel kick: %s [%s] by %s!%s\@%s (%s)", $channel, $FXirc{'client'}{lc($target)}{'nick'}, $FXbot{'DataServ'}{'nick'}, $FXbot{'DataServ'}{'user'}, $FXbot{'DataServ'}{'host'}, $reason);
    &chanlog("k", $channel, $FXbot{'DataServ'}{'nick'}, $FXirc{'client'}{lc($target)}{'nick'}, $reason);
    &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "KICK on %s: %s (%s)", $channel, $FXirc{'client'}{lc($nick)}{'nick'}, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'}, $FXirc{'client'}{lc($target)}{'nick'}, $reason);
    &del_channel($channel, $target);
    return(0);

  } elsif ($cmd eq "MODE") {
    my $mode = (split(/ /, $arg, 0))[0];
    my $args = (split(/ /, $arg, 2))[1];
    my @list = ();

    foreach (split(/ /, $args)) { push(@list, $_); }

    my $idx = 0;
    my $cnt = 0;
    my $what = 1;
    my $exec = "";
    my $target = "";

    foreach (split(//, $mode)) {
      if ($_ eq "+") {
        $what = 1;
        next;
      } elsif ($_ eq "-") {
        $what = 0;
        next;
      }

      if (($_ eq "o") || ($_ eq "h") || ($_ eq "v")) {
        if ($FXirc{'channel'}{lc($channel)}{'names'}{lc($list[$idx])}) {
          if (!$what) { $exec .= "-"; }
          if ($what) { $exec .= "+"; }

          $exec .= $_;
          $target .= $list[$idx++] . " ";
          $cnt++;
        } else {
          $idx++;
        }
      } elsif (($_ eq "b") || ($_ eq "d") || ($_ eq "e") || ($_ eq "I")) {
        if (!$what) { $exec .= "-"; }
        if ($what) { $exec .= "+"; }

        $exec .= $_;
        $target .= $list[$idx++] . " ";
        $cnt++;
      } elsif (($_ =~ /[fklL]/) && !(($_ eq "l") && !$what)) {
        if (!$what) { $exec .= "-"; }
        if ($what) { $exec .= "+"; }

        $exec .= $_;
        $target .= $list[$idx++] . " ";
        $cnt++;
      } elsif ($_ =~ /[cilmnpst]/) {
        if (!$what) { $exec .= "-"; }
        if ($what) { $exec .= "+"; }

        $exec .= $_;
        $cnt++;
      }

      if (!$cnt) { next; }

      if (!($cnt % $FXcfg{'general'}{'modes'})) {
        $target =~ s/\s+$//;

        &putsock($FXsock{'irc'}, ":%s MODE %s %s %s", $FXbot{'DataServ'}{'nick'}, $FXirc{'channel'}{lc($channel)}{'channel'}, $exec, $target);
        &chanlog("m", $FXirc{'channel'}{lc($channel)}{'channel'}, $FXbot{'DataServ'}{'nick'}, $exec, $target);
        &channel_mode($FXbot{'DataServ'}{'nick'}, $FXirc{'channel'}{lc($channel)}{'channel'}, $exec, $target);
        &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "MODE on %s: %s %s", $FXirc{'channel'}{lc($channel)}{'channel'}, $exec, $target);

        $exec = "";
        $target = "";
      }
    }

    if ($exec) {
      $target =~ s/\s+$//;

      &putsock($FXsock{'irc'}, ":%s MODE %s %s %s", $FXbot{'DataServ'}{'nick'}, $FXirc{'channel'}{lc($channel)}{'channel'}, $exec, $target);
      &chanlog("m", $FXirc{'channel'}{lc($channel)}{'channel'}, $FXbot{'DataServ'}{'nick'}, $exec, $target);
      &channel_mode($FXbot{'DataServ'}{'nick'}, $FXirc{'channel'}{lc($channel)}{'channel'}, $exec, $target);
      &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "MODE on %s: %s %s", $FXirc{'channel'}{lc($channel)}{'channel'}, $exec, $target);
    }

    return(0);

  } elsif ($cmd eq "TOPIC") {
    &putsock($FXsock{'irc'}, ":%s TOPIC %s :%s", $FXbot{'DataServ'}{'nick'}, $channel, $arg);

    &chanlog("t", $channel, $FXbot{'DataServ'}{'nick'}, $arg, undef);

    $FXirc{'channel'}{lc($channel)}{'topic'}{'string'} = $arg;
    $FXirc{'channel'}{lc($channel)}{'topic'}{'author'} = $FXbot{'DataServ'}{'nick'};
    $FXirc{'channel'}{lc($channel)}{'topic'}{'set'} = $ts;
    &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "TOPIC on %s: %s", $channel, $arg);
    return(0);

  } elsif ($cmd eq "INVITE") {
    if (!$FXirc{'client'}{lc($arg)}) {
      &ds_notice($nick, $FXmsg{'no_user'}, $arg);
      return(0);
    }

    &putsock($FXsock{'irc'}, ":%s INVITE %s :%s", $FXbot{'DataServ'}{'nick'}, $channel, $FXirc{'client'}{lc($arg)}{'nick'});
    &servicelog($nick, $FXbot{'DataServ'}{'nick'}, "INVITE on %s: %s", $channel, $FXirc{'client'}{lc($arg)}{'nick'});
    return(0);
  }

  return(0);
}

1;