use strict;
use Cwd;
use File::Spec;

our (%FXcfg, %FXflags, %FXbot, %FXserv, %FXfserv, %FXoper, %FXirc, %FXmsg, %FXsock, %FXlisten, $LangPack, $ts, $bit);

my %status = ();
my %search = ();

$FXbot{'FileServ'}{'sub'} = {
  'HELP'   => \&xs_help,
  'WHO'    => \&xs_who,
  'BOOT'   => \&xs_boot,
  'START'  => \&xs_start,
  'STATUS' => \&xs_status
};

$FXbot{'FileServ'}{'public_sub'} = {
  'START'  => \&xs_start,
  'STATUS' => \&xs_status
};

$FXbot{'FileServ'}{'fserv_sub'} = {
  'HELP'  => \&fserv_help,
  'MOTD'  => \&fserv_motd,
  'DIR'   => \&fserv_dir,
  'LS'    => \&fserv_dir,
  'CD'    => \&fserv_cd,
  'GET'   => \&fserv_get,
  'QUEUE' => \&fserv_queue,
  'FIND'  => \&fserv_find,
  'QUIT'  => \&fserv_quit,
  'EXIT'  => \&fserv_quit,
  'BYE'   => \&fserv_quit
};

$FXbot{'FileServ'}{'other_sub'} = {
  'process_xs'    => \&process_xs,
  'xs_dumpdir'    => \&xs_dumpdir,
  'xs_send'       => \&xs_send,
  'xs_get'        => \&xs_get,
  'xs_queued'     => \&xs_queued,
  'fserv_parse'   => \&fserv_parse,
  'fserv_flood'   => \&fserv_flood,
  'fserv_finddir' => \&fserv_finddir,
  'fserv_send'    => \&fserv_send,
  'xs_notice'     => \&xs_notice,
  'xs_ctcp'       => \&xs_ctcp
};

sub fserv_flood {
  my $socket = shift;

  if (($ts - $FXsock{'list'}{$socket}{'lastcmd'}) < $FXbot{'FileServ'}{'flood_time'}) {
    if (++$FXsock{'list'}{$socket}{'floodcnt'} < $FXbot{'FileServ'}{'flood_count'}) {
      $FXsock{'list'}{$socket}{'lastcmd'} = $ts;
      $FXsock{'list'}{$socket}{'warning'} = 0;
      return(0);
    } else {
      &putsock($socket, $FXmsg{'flood_fserv'}, $FXbot{'FileServ'}{'flood_time'}, $FXbot{'FileServ'}{'flood_count'});
      &putinfo("y", "%s: detected flood from %s\@%s (closing socket)", $FXbot{'FileServ'}{'nick'}, $FXsock{'list'}{$socket}{'nick'}, $FXsock{'list'}{$socket}{'fserv'});
      &del_socket($socket);
      return(1);
    }
  } else {
    $FXsock{'list'}{$socket}{'lastcmd'}  = $ts;
    $FXsock{'list'}{$socket}{'floodcnt'} = 0;
    $FXsock{'list'}{$socket}{'warning'}  = 0;
    return(0);
  }
}

sub xs_notice {
  my ($nick, $format, @arg) = (shift, shift, @_);
  my $buffer = sprintf($format, @arg);

  &putsock($FXsock{'irc'}, ":%s %s %s :%s", $FXbot{'FileServ'}{'nick'}, $FXcfg{'general'}{'communicate'}, $nick, $buffer);
  return(0);
}

sub xs_ctcp {
  my ($nick, $notice, $format, @arg) = (shift, shift, @_);
  my $buffer = sprintf($format, @arg);

  &putsock($FXsock{'irc'}, ":%s %s %s :\cA%s\cA", $FXbot{'FileServ'}{'nick'}, $notice ? "NOTICE" : "PRIVMSG", $nick, $buffer);
  return(0);
}

sub process_xs {
  my ($nick, $query) = (shift, shift);

  if (&check_flood($nick)) { return(0); }

  if ($query =~ /\cA(.*)\cA/) {
    my $ctcp = uc((split(/ /, $1, 0))[0]);

    if ($ctcp eq "VERSION") {
      &xs_ctcp($nick, 1, "VERSION %s", "FXserv IRC Services $FXserv{'version'}");
      &xs_ctcp($nick, 1, "VERSION %s", "Administrator: $FXserv{'admin'}");
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $FXbot{'FileServ'}{'nick'}, $ctcp, $FXirc{'client'}{lc($nick)}{'nick'}, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    } elsif ($ctcp eq "PING") {
      &xs_ctcp($nick, 1, "PING %s", (split(/ /, $1, 0))[1]);
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $FXbot{'FileServ'}{'nick'}, $ctcp, $FXirc{'client'}{lc($nick)}{'nick'}, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    } elsif ($ctcp eq "TIME") {
      &xs_ctcp($nick, 1, "TIME %s", scalar localtime());
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $FXbot{'FileServ'}{'nick'}, $ctcp, $FXirc{'client'}{lc($nick)}{'nick'}, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    } elsif ($ctcp eq "DCC") {
      if ($query =~ /DCC CHAT chat (\d+) (\d+)/i) {
        &xs_notice($nick, "Please use: /msg %s START", $FXbot{'FileServ'}{'nick'});
      } elsif ($query =~ /DCC SEND (\S+) (\d+) (\d+) (\d+)/i) {
        if ($FXbot{'FileServ'}{'fserv_upload'}) {
          &xs_get($nick, $1, $2, $3, $4);
        } else {
          &xs_notice($nick, $FXmsg{'dcc_reject'});
        }
      } elsif ($query =~ /DCC RESUME (\S+) (\d+) (\d+)/i) {
        foreach (keys(%{$FXsock{'list'}})) {
          if (($FXsock{'list'}{$_}{'type'} eq "FSERV-B") && ($2 == $FXsock{'list'}{$_}{'port'})) {
            $FXsock{'list'}{$_}{'seek'} = $3;
            &xs_ctcp($nick, 0, "DCC ACCEPT %s %d %d", $1, $2, $3);
            last;
          }
        }
      }
    } else {
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $FXbot{'FileServ'}{'nick'}, $1, $FXirc{'client'}{lc($nick)}{'nick'}, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    }

    return(0);
  }

  my $cmd = (split(/ /, $query, 0))[0];
  my $arg = (split(/ /, $query, 2))[1];

  if ($FXbot{'FileServ'}{'sub'}->{uc($cmd)}) {
    &{$FXbot{'FileServ'}{'sub'}->{uc($cmd)}}($nick, $arg);
  } else {
    &xs_notice($nick, $FXmsg{'unkn_cmd'}, $cmd);
  }

  return(0);
}

sub xs_help {
  my ($nick, $arg) = (shift, shift);

  if (!$arg) { $arg = "index"; }

  my $file = lc($arg);
  $file =~ s/ /\//g;
  $file = "help/" . $LangPack . "/fileserv/" . $file;

  if (!open(FILE, "<" . $file)) {
    &xs_notice($nick, $FXmsg{'no_help'}, $arg);
    return(0);
  }

  my @lines = <FILE>;
  close(FILE);

  foreach my $buffer (@lines) {
    chomp($buffer);
    $buffer = &substitute($buffer);

    if ($buffer =~ /^\%\+a/) {
      if (&valid_oper($nick, "a")) {
        $buffer =~ s/^\%\+a//;
        &xs_notice($nick, $buffer);
      }
    } elsif ($buffer =~ /^\%\+o/) {
      if (&valid_oper($nick, "o")) {
        $buffer =~ s/^\%\+o//;
        &xs_notice($nick, $buffer);
      }
    } else {
      &xs_notice($nick, $buffer);
    }
  }

  &servicelog($nick, $FXbot{'FileServ'}{'nick'}, "HELP %s", $arg);
  return(0);
}

sub xs_boot {
  my ($nick, $target) = (shift, shift);

  if (!&valid_oper($nick, "o")) {
    &xs_notice($nick, $FXmsg{'no_priv'});
    return(0);
  }

  if (!$target) {
    &xs_notice($nick, $FXmsg{'syntax'}, "BOOT <nick>");
    return(0);
  }

  foreach (keys %{$FXsock{'list'}}) {
    if (($FXsock{'list'}{$_}{'type'} eq "FSERV") && (lc($FXsock{'list'}{$_}{'nick'}) eq lc($target))) {
      &xs_notice($nick, $FXmsg{'booted'}, $target);
      &servicelog($nick, $FXbot{'FileServ'}{'nick'}, "BOOT %s", $FXsock{'list'}{$_}{'nick'});
      &del_socket($FXsock{'list'}{$_}{'sock'});
      return(0);
    }
  }

  &xs_notice($nick, $FXmsg{'no_user'}, $target);
  return(0);
}

sub xs_start {
  my $nick = shift;
  my $idx = 0;

  foreach (keys(%{$FXfserv{'list'}})) {
    if ($FXfserv{'list'}{$_}{'sock'}) {
      if (lc($FXfserv{'list'}{$_}{'nick'}) eq lc($nick)) {
        &xs_notice($nick, $FXmsg{'one_dcc'});
        return(0);
      }

      if (++$idx > $FXbot{'FileServ'}{'fserv_global'}) {
        &xs_notice($nick, $FXmsg{'dcc_max'});
        return(0);
      }

    }
  }

  foreach (keys(%{$FXfserv{'query'}})) {
    if ((lc($FXfserv{'query'}{$_}{'nick'}) eq lc($nick)) && ($ts - $FXfserv{'query'}{$_}{'time'} < $FXbot{'FileServ'}{'fserv_timeout'})) {
      &xs_notice($nick, $FXmsg{'fserv_wait'}, abs($ts - $FXfserv{'query'}{$_}{'time'} - $FXbot{'FileServ'}{'fserv_timeout'}));
      return(0);
    }
  }

  my @range = &parse_range($FXbot{'FileServ'}{'fserv_dcc'});
  my $socket = "";
  my $port = 0;

  foreach (@range) {
    $port = $_;
    if (!&valid_port($port)) { next; }

    $socket = &init_listen($port, "tcp", "FSERV-A");
    if ($socket) {
      $FXsock{'list'}{$socket}{'nick'} = $nick;
      last;
    }
  }

  if (!$socket) {
    &putlog("y", "Unable to listen port %d", $port);
    &xs_notice($nick, $FXmsg{'dcc_failed'});
    return(0);
  }

  $FXfserv{'query'}{lc($nick)}{'nick'} = $nick;
  $FXfserv{'query'}{lc($nick)}{'time'} = $ts;

  my $longip = unpack("N", inet_aton($FXserv{'ip'}));

  &xs_notice($nick, "DCC Chat (%s)", $FXserv{'ip'});
  &xs_ctcp($nick, 0, "DCC CHAT chat %d %d", $longip, $port);

  &servicelog($nick, $FXbot{'FileServ'}{'nick'}, "START");
  return(0);
}

sub xs_who {
  my $nick = shift;

  if (!&valid_oper($nick, "o")) {
    &xs_notice($nick, $FXmsg{'no_priv'});
    return(0);
  }

  my $idx = 0;

  &xs_notice($nick, "-- Listing users --");
  &xs_notice($nick, "[Nickname       ] [IP             ] [D/L] [Traffic   ] [DCC]");

  foreach (keys(%{$FXfserv{'list'}})) {
    ++$idx;
    my $dcc = "";

    if ($FXfserv{'list'}{$_}{'sock'}) { $dcc = " X "; }
    &xs_notice($nick, "[%-15s] [%-15s] [%-3s] [%-10s] [%-3s]", $FXfserv{'list'}{$_}{'nick'}, $FXfserv{'list'}{$_}{'ip'}, $FXfserv{'list'}{$_}{'total'}, &format_size($FXfserv{'list'}{$_}{'traffic'}), $dcc);
  }

  &xs_notice($nick, "-- End of list [%d user%s found] --", $idx, $idx > 1 ? "s" : "");
  &servicelog($nick, $FXbot{'FileServ'}{'nick'}, "WHO");
  return(0);
}

sub xs_status {
  my $nick = shift;
  my $ondcc = 0;
  my $uploads = 0;
  my $downloads = 0;
  my $cps = 0;

  $status{lc($nick)}{'files'} = 0;
  $status{lc($nick)}{'size'} = 0;
  &xs_dumpdir($nick, $FXbot{'FileServ'}{'fserv_root'});

  foreach (keys(%{$FXfserv{'list'}})) {
    if ($FXfserv{'list'}{$_}{'sock'}) { ++$ondcc; }
    $downloads += $FXfserv{'list'}{$_}{'xfer'};
  }

  foreach (keys(%{$FXsock{'list'}})) {
    if (($FXsock{'list'}{$_}{'type'} eq "FSERV-C") || ($FXsock{'list'}{$_}{'type'} eq "FSERV-X")) {
      if ($ts - $FXsock{'list'}{$_}{'start'} > 0) {
        $cps += $FXsock{'list'}{$_}{'seek'} / ($ts - $FXsock{'list'}{$_}{'start'});
      } else {
        $cps += $FXsock{'list'}{$_}{'seek'};
      }
    }

    if ($FXsock{'list'}{$_}{'type'} eq "FSERV-G") { ++$uploads; }
  }

  &xs_notice($nick, "-- Status monitor --");
  &xs_notice($nick, "Currently on DCC Chat: %d user%s", $ondcc, $ondcc > 1 ? "s" : "");
  &xs_notice($nick, "Current downloads:     %d", $downloads);
  if ($FXbot{'FileServ'}{'fserv_upload'}) { &xs_notice($nick, "Current uploads:       %d", $uploads); }
  &xs_notice($nick, "Current speed:         %s/sec", &format_size($cps));
  &xs_notice($nick, "Free slots:            %d/%d", ($FXbot{'FileServ'}{'fserv_global'} - $ondcc), $FXbot{'FileServ'}{'fserv_global'});
  &xs_notice($nick, "Files served:          %d (traffic: %s)", $FXbot{'FileServ'}{'total'}, &format_size($FXbot{'FileServ'}{'traffic'}));
  &xs_notice($nick, "Total files:           %d (%s)", $status{lc($nick)}{'files'}, &format_size($status{lc($nick)}{'size'}));

  delete($status{lc($nick)});
  &servicelog($nick, $FXbot{'FileServ'}{'nick'}, "STATUS");
  return(0);
}

sub xs_dumpdir {
  my ($nick, $folder) = (shift, shift);

  if (!opendir(DIR, $folder)) {
    &putlog("y", "Unable to read folder: %s", $folder);
    return(0);
  }

  my @list = readdir(DIR);
  closedir(DIR);

  foreach (@list) {
    if ($_ =~ /^\./) { next; }

    my $absolute = $folder . "/" . $_;
    if (-d $absolute) {
      &xs_dumpdir($nick, $absolute);
    } else {
      ++$status{lc($nick)}{'files'};
      $status{lc($nick)}{'size'} += -s $absolute;
      push(@{$status{lc($nick)}{'list'}}, $absolute);
    }
  }

  return(0);
}

sub fserv_parse {
  my ($socket, $buffer) = (shift, shift);

  if (&fserv_flood($socket)) { return(0); }

  $FXsock{'list'}{$socket}{'lastcmd'} = $ts;
  $FXsock{'list'}{$socket}{'warning'} = 0;

  my $cmd = (split(/ /, $buffer, 0))[0];
  my $arg = (split(/ /, $buffer, 2))[1];

  if ($FXbot{'FileServ'}{'fserv_sub'}->{uc($cmd)}) {
    &{$FXbot{'FileServ'}{'fserv_sub'}->{uc($cmd)}}($socket, $arg);
  } else {
    &putsock($socket, $FXmsg{'unkn_cmd'}, $cmd);
  }

  return(0);
}

sub fserv_help {
  my $socket = shift;
  my $ip = $FXsock{'list'}{$socket}{'fserv'};
  my $nick = $FXfserv{'list'}{$ip}{'nick'};

  &send_motd($socket, $FXbot{'FileServ'}{'fserv_help'});

  &servicelog($nick, $FXbot{'FileServ'}{'nick'}, "#HELP#");
  return(0);
}

sub fserv_motd {
  my $socket = shift;
  my $ip = $FXsock{'list'}{$socket}{'fserv'};
  my $nick = $FXfserv{'list'}{$ip}{'nick'};

  &send_motd($socket, $FXbot{'FileServ'}{'fserv_welcome'});

  &servicelog($nick, $FXbot{'FileServ'}{'nick'}, "#MOTD#");
  return(0);
}

sub fserv_queue {
  my $socket = shift;
  my $ip = $FXsock{'list'}{$socket}{'fserv'};
  my $idx = 0;

  if ($FXfserv{'list'}{$ip}{'queue'} && scalar(@{$FXfserv{'list'}{$ip}{'queue'}})) {
    &putsock($socket, "-- Listing files --");

    foreach (@{$FXfserv{'list'}{$ip}{'queue'}}) {
      my ($vol, $dir, $file) = File::Spec->splitpath($_);
      my $size = -s $_;
      &putsock($socket, "%02d) %-32s %s", ++$idx, $file, &format_size($size));
    }

    &putsock($socket, "-- End of list [%d file%s queued] --", $idx, $idx > 1 ? "s" : "");
  } else {
    &putsock($socket, $FXmsg{'queue_empty'});
  }

  &servicelog($FXfserv{'list'}{$ip}{'nick'}, $FXbot{'FileServ'}{'nick'}, "#QUEUE#");
  return(0);
}

sub fserv_find {
  my ($socket, $pattern) = (shift, shift);
  my $ip = $FXsock{'list'}{$socket}{'fserv'};

  if (!$pattern) {
    &putsock($socket, $FXmsg{'syntax'}, "FIND <pattern>");
    return(0);
  }

  $search{$socket}{'files'} = 0;
  $search{$socket}{'size'} = 0;
  $search{$socket}{'print'} = 0;

  &putsock($socket, "-- Listing files matching [%s] --", $pattern);
  &fserv_finddir($socket, $pattern, $FXsock{'list'}{$socket}{'folder'});
  if ($search{$socket}{'print'} > $FXbot{'FileServ'}{'fserv_maxwild'}) { &putsock($socket, "(%d other matches not displayed)", $search{$socket}{'print'} - $FXbot{'FileServ'}{'fserv_maxwild'}); }
  &putsock($socket, "-- End of list [%d file%s, %s] --", $search{$socket}{'files'}, $search{$socket}{'files'} > 1 ? "s" : "", &format_size($search{$socket}{'size'}));
  delete($search{$socket});

  &servicelog($FXfserv{'list'}{$ip}{'nick'}, $FXbot{'FileServ'}{'nick'}, "#FIND %s#", $pattern);
  return(0);
}

sub fserv_finddir {
  my ($socket, $pattern, $folder) = (shift, shift, shift);
  my $patternrx = &plain2regexp($pattern);
  my $dir = "";

  if ($folder eq "/") {
    $dir = $FXbot{'FileServ'}{'fserv_root'};
  } else {
    $dir = $FXbot{'FileServ'}{'fserv_root'} . $folder;
  }

  if (!opendir(DIR, $dir)) {
    &putlog("y", "Unable to read folder: %s", $dir);
    return(0);
  }

  my @list = readdir(DIR);
  closedir(DIR);

  foreach (@list) {
    if ($_ =~ /^\./) { next; }

    my $absolute = $dir . "/" . $_;
    if (-d $absolute) {
      &fserv_finddir($socket, $pattern, $folder eq "/" ? "/" . $_ : $folder . "/" . $_);
    } elsif ((-e $absolute) && ($_ =~ /$patternrx/i)) {
      if (++$search{$socket}{'print'} < $FXbot{'FileServ'}{'fserv_maxwild'}) { &putsock($socket, "%-32s %s", $folder eq "/" ? $_ : $folder . "/" . $_, &format_size(-s $absolute)); }
      ++$search{$socket}{'files'};
      $search{$socket}{'size'} += -s $absolute;
    }
  }

  return(0);
}

sub fserv_cd {
  my ($socket, $folder) = (shift, shift);
  my $ip = $FXsock{'list'}{$socket}{'fserv'};

  if (!$folder || ($folder eq ".")) {
    &putsock($socket, "Current folder: [%s]", $FXsock{'list'}{$socket}{'folder'});
    return(0);
  }

  my $absolute = "";

  if ($folder =~ /^\//) {
    if ($folder eq "/") {
      $absolute = $FXbot{'FileServ'}{'fserv_root'};
    } else {
      $absolute = $FXbot{'FileServ'}{'fserv_root'} . $folder;
    }
  } else {
    $absolute = $FXbot{'FileServ'}{'fserv_root'} . $FXsock{'list'}{$socket}{'folder'} . "/". $folder;
  }

  $absolute = File::Spec->canonpath($absolute);
  $absolute = Cwd::abs_path($absolute);

  if ($absolute eq "/") { $absolute = $FXbot{'FileServ'}{'fserv_root'}; }

  if ($absolute !~ /^$FXbot{'FileServ'}{'fserv_root'}/) {
    &putsock($socket, "Invalid directory [%s]", $folder);
    return(0);
  } elsif ((-d $absolute) && ($absolute !~ /\/\./)) {
    $folder = substr($absolute, length($FXbot{'FileServ'}{'fserv_root'}));
    if (!$folder) { $folder = "/"; }
    &putsock($socket, "Entering directory: [%s]", $folder);
    $FXsock{'list'}{$socket}{'folder'} = $folder;

    if (-e $absolute . "/" . $FXbot{'FileServ'}{'fserv_dirinfo'}) {
      &putsock($socket, "-------------------------------------");
      &send_motd($socket, $absolute . "/" . $FXbot{'FileServ'}{'fserv_dirinfo'});
      &putsock($socket, "-------------------------------------");
    }

    &servicelog($FXfserv{'list'}{$ip}{'nick'}, $FXbot{'FileServ'}{'nick'}, "#CD %s#", $folder);
    return(0);
  } else {
    &putsock($socket, "Invalid directory [%s]", $folder);
    return(0);
  }
}

sub fserv_dir {
  my $socket = shift;
  my $ip = $FXsock{'list'}{$socket}{'fserv'};

  if (!opendir(DIR, $FXbot{'FileServ'}{'fserv_root'} . $FXsock{'list'}{$socket}{'folder'})) {
    &putlog("y", "Unable to read folder: %s", $FXbot{'FileServ'}{'fserv_root'} . $FXsock{'list'}{$socket}{'folder'});
    return(0);
  }

  my @list = readdir(DIR);
  closedir(DIR);

  my @files = ();
  my @folders = ();

  foreach (@list) {
    if ($_ =~ /^\./) { next; }

    my $absolute = $FXbot{'FileServ'}{'fserv_root'} . $FXsock{'list'}{$socket}{'folder'} . $_;
    if (-d $absolute) {
      push(@folders, $_);
    } else {
      push(@files, $_);
    }
  }

  @folders = sort(@folders);
  @files = sort(@files);

  if (!scalar(@folders) && !scalar(@files)) {
    &putsock($socket, "Directory is empty");
    return(0);
  }

  my $tfiles = 0;
  my $tsize = 0;

  &putsock($socket, "-- Listing [%s] --", $FXsock{'list'}{$socket}{'folder'});
  foreach (@folders) { &putsock($socket, "<%s>", $_); }
  foreach (@files) {
    my $absolute = $FXbot{'FileServ'}{'fserv_root'} . $FXsock{'list'}{$socket}{'folder'} . "/" . $_; $absolute = File::Spec->canonpath($absolute);
    my $size = -s $absolute;
    ++$tfiles;
    $tsize += $size;
    &putsock($socket, "%-50s [%-5s]", $_, &format_size($size));
  }

  &putsock($socket, "-- End of list [%d file%s, %s] --", $tfiles, $tfiles > 1 ? "s" : "", &format_size($tsize));
  &servicelog($FXfserv{'list'}{$ip}{'nick'}, $FXbot{'FileServ'}{'nick'}, "#DIR %s#", $FXsock{'list'}{$socket}{'folder'});
  return(0);
}

sub fserv_get {
  my ($socket, $filename) = (shift, shift);
  my $ip = $FXsock{'list'}{$socket}{'fserv'};

  if (!$filename) {
    &putsock($socket, $FXmsg{'syntax'}, "GET <file>");
    return(0);
  }

  my $absolute = "";

  if ($filename =~ /^\//) {
    $absolute = $FXbot{'FileServ'}{'fserv_root'} . $filename;
  } else {
    $absolute = $FXbot{'FileServ'}{'fserv_root'} . $FXsock{'list'}{$socket}{'folder'} . "/" . $filename;
  }

  my ($vol, $dir, $file) = File::Spec->splitpath($absolute);
  $absolute = File::Spec->canonpath($dir);
  $absolute = Cwd::abs_path($absolute);
  $absolute .= "/" . $file;

  if ($absolute !~ /^$FXbot{'FileServ'}{'fserv_root'}/) {
    &putsock($socket, $FXmsg{'no_file'}, $filename);
    return(0);
  } elsif ((!-d $absolute) && (-e $absolute) && ($absolute !~ /\/\./)) {
    &servicelog($FXfserv{'list'}{$ip}{'nick'}, $FXbot{'FileServ'}{'nick'}, "#GET %s#", $file);
    my $size = -s $absolute;

    if ($FXfserv{'list'}{$ip}{'xfer'} >= $FXbot{'FileServ'}{'fserv_maxsend'}) {
      if (scalar(@{$FXfserv{'list'}{$ip}{'queue'}}) >= $FXbot{'FileServ'}{'fserv_queue'}) {
        &putsock($socket, $FXmsg{'queue_full'});
        return(0);
      } else {
        foreach (@{$FXfserv{'list'}{$ip}{'queue'}}) {
          if ($_ eq $absolute) {
            &putsock($socket, $FXmsg{'queue_in'});
            return(0);
          }
        }

        push(@{$FXfserv{'list'}{$ip}{'queue'}}, $absolute);
        &putsock($socket, $FXmsg{'queue_added'}, $file, &format_size($size));
        return(0);
      }
    } else {
      &xs_send($ip, $absolute);
      return(0);
    }

  } else {
    &putsock($socket, $FXmsg{'no_file'}, $filename);
    return(0);
  }
}

sub fserv_quit {
  my $socket = shift;
  my $ip = $FXsock{'list'}{$socket}{'fserv'};

  &putsock($socket, "OK, bye!");
  &servicelog($FXfserv{'list'}{$ip}{'nick'}, $FXbot{'FileServ'}{'nick'}, "#QUIT#");
  &del_socket($socket);
  return(0);
}

sub xs_send {
  my ($ip, $absolute) = (shift, shift);

  my $size = -s $absolute;
  my ($vol, $dir, $file) = File::Spec->splitpath($absolute);

  my @range = &parse_range($FXbot{'FileServ'}{'fserv_dcc'});
  my $socket = "";
  my $port = 0;

  foreach (@range) {
    $port = $_;
    if (!&valid_port($port)) { next; }
  
    $socket = &init_listen($port, "tcp", "FSERV-B");
    if ($socket) {
      $FXsock{'list'}{$socket}{'fserv'} = $ip;
      $FXsock{'list'}{$socket}{'absolute'} = $absolute;
      $FXsock{'list'}{$socket}{'seek'} = 0;
      last;
    }
  }

  if (!$socket) {
    &putlog("y", "Unable to listen port %d", $port);
    &xs_notice($FXfserv{'list'}{$ip}{'nick'}, $FXmsg{'dcc_failed'});
    &xs_queued($ip);
    return(0);
  }

  if ($FXfserv{'list'}{$ip}{'sock'}) {
    &putsock($FXfserv{'list'}{$ip}{'sock'}, "Now sending you: %s (%s)", $file, &format_size(-s $absolute));
  } else {
    &xs_notice($FXfserv{'list'}{$ip}{'nick'}, "Now sending you: %s (%s)", $file, &format_size(-s $absolute));
  }

  ++$FXfserv{'list'}{$ip}{'xfer'};

  my $longip = unpack("N", inet_aton($FXserv{'ip'}));

  &xs_notice($FXfserv{'list'}{$ip}{'nick'}, "DCC Send %s (%s)", $file, $FXserv{'ip'});
  $file =~ s/ /_/g;
  &xs_ctcp($FXfserv{'list'}{$ip}{'nick'}, 0, "DCC SEND %s %d %d %d", $file, $longip, $port, $size);

  return(0);
}

sub xs_queued {
  my $ip = shift;

  if ($FXfserv{'list'}{$ip}{'queue'} && scalar(@{$FXfserv{'list'}{$ip}{'queue'}})) {
    my $absolute = shift(@{$FXfserv{'list'}{$ip}{'queue'}});
    &xs_send($ip, $absolute);
    return(0);
  } else {
    if (!$FXfserv{'list'}{$ip}{'sock'} && !$FXfserv{'list'}{$ip}{'xfer'}) { delete($FXfserv{'list'}{$ip}); }
    return(0);
  }
}

sub fserv_send {
  my ($socket, $seek) = (shift, shift);
  my $ip = $FXsock{'list'}{$socket}{'fserv'};

  if (!open(FILE, "<" . $FXsock{'list'}{$socket}{'absolute'})) {
    &putlog("y", "Unable to read file: %s", $FXsock{'list'}{$socket}{'absolute'});
    &del_socket($socket);
    return(0);
  }

  my $buffer = "";
  my $size = -s $FXsock{'list'}{$socket}{'absolute'};

  $FXsock{'list'}{$socket}{'time'} = $ts;

  binmode FILE;
  seek(FILE, $seek, 0);

  if ($size - $seek < $FXbot{'FileServ'}{'fserv_buffer'}) {
    read(FILE, $buffer, $size - $seek);
  } else {
    read(FILE, $buffer, $FXbot{'FileServ'}{'fserv_buffer'});
  }

  if (!$buffer) {
    close(FILE);

    my ($vol, $dir, $file) = File::Spec->splitpath($FXsock{'list'}{$socket}{'absolute'});
    my $cps = $size;

    if ($ts - $FXsock{'list'}{$socket}{'start'} > 0) { $cps = $size / ($ts - $FXsock{'list'}{$socket}{'start'}); }
    $FXsock{'list'}{$socket}{'type'} = "FSERV-D";

    --$FXfserv{'list'}{$ip}{'xfer'};
    ++$FXfserv{'list'}{$ip}{'total'};
    ++$FXbot{'FileServ'}{'total'};

    if ($FXfserv{'list'}{$ip}{'sock'}) {
      &putsock($FXfserv{'list'}{$ip}{'sock'}, "DCC Send of %s (%s) complete (%s, %s/sec)", $file, &format_size($size), &format_time($FXsock{'list'}{$socket}{'start'}), &format_size($cps));
    } else {
      &xs_notice($FXfserv{'list'}{$ip}{'nick'}, "DCC Send of %s (%s) complete (%s, %s/sec)", $file, &format_size($size), &format_time($FXsock{'list'}{$socket}{'start'}), &format_size($cps));
    }

    &del_socket($socket);
    &xs_queued($ip);
    return(0);
  }

  my $length = length($buffer);

  $FXfserv{'list'}{$ip}{'traffic'} += $length;
  $FXbot{'FileServ'}{'traffic'} += $length;
  $FXsock{'list'}{$socket}{'seek'} += $length;
  print $socket $buffer;
  close(FILE);

  if ($FXbot{'FileServ'}{'fserv_cps'}) { $FXfserv{'list'}{$ip}{'limitcps'} += $length; }
  return(0);
}

sub xs_get {
  my ($nick, $filename, $addr, $port, $size) = (shift, shift, shift, shift, shift);
  my $idx = 0;

  foreach (keys %{$FXsock{'list'}}) {
    if (($FXsock{'list'}{$_}{'type'} eq "FSERV-G") && (++$idx > $FXbot{'FileServ'}{'fserv_maxget'})) {
      &xs_notice($nick, "%s", $FXmsg{'dcc_max'});
      return(0);
    }
  }

  if ($FXbot{'FileServ'}{'fserv_maxsize'} && ($size > $FXbot{'FileServ'}{'fserv_maxsize'})) {
    &xs_notice($nick, $FXmsg{'fserv_toobig'}, $filename, &format_size($size));
    return(0);
  }

  if (!opendir(DIR, $FXbot{'FileServ'}{'fserv_upload'})) {
    &putlog("y", "Unable to open folder: %s", $FXbot{'FileServ'}{'fserv_upload'});
    &xs_notice($nick, "%s", $FXmsg{'dcc_failed'});
    return(0);
  }

  my @list = readdir(DIR);
  closedir(DIR);

  foreach (@list) {
    if ($_ eq $filename) {
      &xs_notice($nick, "$FXmsg{'fserv_exists'}", $filename);
      return(0);
    }
  }

  my $FILE;

  if (!open($FILE, ">" . $FXbot{'FileServ'}{'fserv_upload'} . "/" . $filename)) {
    &xs_notice($nick, "%s", $FXmsg{'dcc_failed'});
    return(0);
  }

  binmode $FILE;

  my $ip = inet_ntoa(pack("N", $addr));
  my $socket = &open_socket("tcp", $ip, $port, 3);

  if (!$socket) {
    &servicelog($nick, $FXbot{'FileServ'}{'nick'}, "failed DCC Get %s (%s:%d)", $filename, $ip, $port);
    &xs_notice($nick, "%s", $FXmsg{'dcc_failed'});
    return(0);
  }

  &add_socket($socket, "FSERV-G");

  $FXsock{'list'}{$socket}{'fserv'} = $ip;
  $FXsock{'list'}{$socket}{'start'} = $ts;
  $FXsock{'list'}{$socket}{'file'} = $filename;
  $FXsock{'list'}{$socket}{'size'} = $size;
  $FXsock{'list'}{$socket}{'read'} = 0;
  $FXsock{'list'}{$socket}{'filehd'} = $FILE;
  $FXsock{'list'}{$socket}{'nick'} = $nick;

  $size = &format_size($size);

  &servicelog($nick, $FXbot{'FileServ'}{'nick'}, "DCC Get %s (%s) [starting...]", $filename, $size);
  return(0);
}

1;