use strict;
use Time::HiRes qw(tv_interval gettimeofday);

our (%FXbot, %FXserv, %FXrcon, %FXoper, %FXirc, %FXcfg, %FXmsg, %FXsock, %FXgames, $LangPack, $ts);

$FXbot{'GameServ'}{'sub'} = {
  'HELP'    => \&gs_help,
  'LIST'    => \&gs_list,
  'INFO'    => \&gs_info,
  'RCON'    => \&gs_rcon,
  'WHO'     => \&gs_who,
  'BOOT'    => \&gs_boot,
  'SERVER'  => \&gs_server
};

$FXbot{'GameServ'}{'public_sub'} = {
  'LIST' => \&gs_list,
  'INFO' => \&gs_info
};

$FXbot{'GameServ'}{'other_sub'} = {
  'process_gs'      => \&process_gs,
  'valve_info'      => \&valve_info,
  'valve_players'   => \&valve_players,
  'quake3_info'     => \&quake3_info,
  'quake3_players'  => \&quake3_players,
  'gamespy_info'    => \&gamespy_info,
  'gamespy_players' => \&gamespy_players,
  'udp_socket'      => \&udp_socket,
  'gs_notice'       => \&gs_notice,
  'gs_ctcp'         => \&gs_ctcp,
  'rcon_parse'      => \&rcon_parse,
  'rcon_flood'      => \&rcon_flood
};

my %game_type = (
  'arena'        => 'Quake III Arena',
  'threewave'    => 'Quake III Threewave CTF',
  'osp'          => 'Quake III OSP',
  'baseq3'       => 'Quake III Arena',
  'desertcombat' => 'Desert Combat',
  'ut2004'       => 'Unreal Tournament 2004',
  'ut2003'       => 'Unreal Tournament 2003',
  'ut2d'         => 'Unreal Tournament 2003',
  'ut2'          => 'Unreal Tournament 2003',
  'bfield1942'   => 'Battlefield 1942',
  'czero'        => 'Counter-Strike: Condition Zero',
  'dod'          => 'Day of Defeat',
  'cstrike'      => 'Counter-Strike',
  'ns'           => 'Natural Selection',
  'nsp'          => 'Natural Selection',
  'tfc'          => 'TeamFortress Classic',
  'ts'           => 'The Specialists',
  'call of duty' => 'Call of Duty',
  'etpro'        => 'Wolfenstein: Enemy Territories (ETPro)',
  'osp'          => 'Wolfenstein: Enemy Territories (OSP)',
  'shrub'        => 'Wolfenstein: Enemy Territories (shrub)',
  'mohaa'        => 'Medal of Honor Allied Assault',
  'mohaab'       => 'Medal of Honor: Breakthrough',
  'mohaas'       => 'Medal of Honor: Spearhead',
  'armygame'     => 'America\'s Army'
);

sub valve_info {
  my ($host, $port) = (shift, shift);
  my $ping = 0;
  my $query = "\xFF" x 4 . "info" . "\x00";

  my $t0 = [gettimeofday()];
  my $buffer = &udp_socket($host, $port, $query);
  $ping = tv_interval($t0);
  $ping = sprintf("%.0d", 1000 * $ping);

  if (!$buffer) { return(0); }

  substr($buffer, 0, 5) = "";
  my ($addr, $name, $map, $game, $descr, $playing) = split(/\0/, $buffer);
  my $players = ord(substr($playing, 0, 1));
  my $slots = ord(substr($playing, 1, 1));

  my $info = {
    'name'    => $name,
    'map'     => $map,
    'game'    => lc($game),
    'players' => $players,
    'slots'   => $slots,
    'ping'    => $ping
  };

  return($info);
}

sub gamespy_info {
  my ($host, $port) = (shift, shift);
  my $ping = 0;
  my $query = "\\status\\" . "\x00";

  my $t0 = [gettimeofday()];
  my $buffer = &udp_socket($host, $port, $query);
  $ping = tv_interval($t0);
  $ping = sprintf("%.0d", 1000 * $ping);

  if (!$buffer) { return(0); }

  my %reply = ();
  my $option = 0;

  foreach (split(/\\/, $buffer)) {
    if (!$option) {
      $option = $_;
      $reply{$option} = "";
    } else {
      $reply{$option} = $_;
      $option = 0;
    }
  }

  my $info = {
    'name'    => $reply{"hostname"},
    'map'     => $reply{"mapname"},
    'game'    => lc($reply{"gamename"}),
    'players' => $reply{"numplayers"},
    'slots'   => $reply{"maxplayers"},
    'ping'    => $ping
  };

  return($info);
}

sub quake3_info {
  my ($host, $port) = (shift, shift);
  my $ping = 0;
  my $query = "\xFF" x 4 . "\x02" . "getstatus" . "\x0a\x00";

  my $t0 = [gettimeofday()];
  my $buffer = &udp_socket($host, $port, $query);
  $ping = tv_interval($t0);
  $ping = sprintf("%.0d", 1000 * $ping);

  if (!$buffer) { return(0); }

  my %reply = ();
  my $option = 0;

  my @lines = split(/\n/, $buffer);

  foreach (split(/\\/, $lines[1])) {
    if (!$option) {
      $option = $_;
      $reply{$option} = "";
    } else {
      $reply{$option} = $_;
      $option = 0;
    }
  }

  if (!$reply{"gamename"}) {
    if ($reply{"version"} =~ /Medal of Honor Allied Assault/) { $reply{"gamename"} = "mohaa"; }
    if ($reply{"version"} =~ /Medal of Honor: Breakthrough/) { $reply{"gamename"} = "mohaab"; }
    if ($reply{"version"} =~ /Medal of Honor: Spearhead/) { $reply{"gamename"} = "mohaas"; }
  }

  my $info = {
    'name'    => $reply{"sv_hostname"},
    'map'     => $reply{"mapname"},
    'game'    => lc($reply{"gamename"}),
    'players' => scalar(@lines) - 2,
    'slots'   => $reply{"sv_maxclients"} - $reply{"sv_privateClients"},
    'ping'    => $ping
  };

  return($info);
}

sub valve_players {
  my ($host, $port) = (shift, shift);

  my $query = "\xFF" x 4 . "players" . "\x00";
  my $buffer = &udp_socket($host, $port, $query);
  if (!$buffer) { return(0); }

  substr($buffer, 0, 6) = "";

  my %players = ();

  while (length($buffer) > 0) {
    my $idx = ord(substr($buffer, 0, 1, ""));

    $buffer =~ s/^(.*?)\x00//;

    my $name = $1;
    my $frags = unpack("V", substr($buffer, 0, 4, ""));
    my $time = unpack("f", substr($buffer, 0, 4, ""));

    $players{$idx}{'name'} = $name;
    $players{$idx}{'frags'} = $frags;
    $players{$idx}{'time'} = int($time);
  }

  return(%players);
}

sub gamespy_players {
  my ($host, $port, $type) = (shift, shift, shift);

  my $query = "\\players\\" . "\x00";
  my $buffer = &udp_socket($host, $port, $query);
  if (!$buffer) { return(0); }

  my %reply = ();
  my $option = 0;

  foreach (split(/\\/, $buffer)) {
    if (!$option) {
      $option = $_;
      if ($option =~ /^playername_(\d+)$/) { $option = "player_" . $1; }
      $reply{$option} = "";
    } else {
      $reply{$option} = $_;
      $option = 0;
    }
  }

  my %players = ();

  for (my $idx = 0; defined($reply{"player_" . $idx}); $idx++) {
    $players{$idx}{'name'} = $reply{"player_" . $idx};

    if ($type eq "unreal") {
      $players{$idx}{'frags'} = $reply{"frags_" . $idx};
      $players{$idx}{'deaths'} = $reply{"deaths_" . $idx};
      $players{$idx}{'team'} = $reply{"team_" . $idx};

    } elsif ($type eq "bf1942") {
      $players{$idx}{'frags'} = $reply{"score_" . $idx};
      $players{$idx}{'kills'} = $reply{"kills_" . $idx};
      $players{$idx}{'deaths'} = $reply{"deaths_" . $idx};
      $players{$idx}{'team'} = $reply{"team_" . $idx};

    } elsif ($type eq "army") {
      $players{$idx}{'frags'} = $reply{"enemy_" . $idx};
      $players{$idx}{'team'} = $reply{"team_" . $idx};
      $players{$idx}{'goal'} = $reply{"goal_" . $idx};
      $players{$idx}{'honor'} = $reply{"honor_" . $idx};
      $players{$idx}{'kia'} = $reply{"kia_" . $idx};
      $players{$idx}{'roe'} = $reply{"roe_" . $idx};
      $players{$idx}{'leader'} = $reply{"leader_" . $idx};
    }
  }

  return(%players);
}

sub quake3_players {
  my ($host, $port) = (shift, shift);

  my $query = "\xFF" x 4 . "\x02" . "getstatus" . "\x0a\x00";
  my $buffer = &udp_socket($host, $port, $query);
  if (!$buffer) { return(0); }

  my @lines = split(/\n/, $buffer);

  my %players = ();
  my $idx = 0;

  for (my $cnt = 2; $cnt < scalar(@lines); $cnt++) {
    my ($frags, $lag, $name) = split(/ /, $lines[$cnt], 3);

    $name =~ s/\^[^\"]//g; $name = substr(substr($name, 1), 0, -1);

    $players{$idx}{'name'} = $name;
    $players{$idx}{'frags'} = $frags;
    $players{$idx}{'ping'} = $lag;
    $idx++;
  }

  return(%players);
}

sub udp_socket {
  my ($host, $port, $query) = (shift, shift, shift);
  my $recv = "";

  my $socket = &open_socket("udp", $host, $port, $FXbot{'GameServ'}{'games_timeout'});

  if (!$socket) { return(0); }

  $socket->autoflush(1);

  eval {
    local $SIG{'ALRM'} = sub { die(); };
    alarm($FXbot{'GameServ'}{'games_timeout'});

    print $socket $query;
    recv($socket, $recv, 4096, 0);

    alarm(0);
  };

  close($socket);
  return($recv);
}

sub gs_notice {
  my ($nick, $format, @arg) = (shift, shift, @_);
  my $buffer = sprintf($format, @arg);

  &putsock($FXsock{'irc'}, ":%s %s %s :%s", $FXbot{'GameServ'}{'nick'}, $FXcfg{'general'}{'communicate'}, $nick, $buffer);
  return(0);
}

sub gs_ctcp {
  my ($nick, $notice, $format, @arg) = (shift, shift, @_);
  my $buffer = sprintf($format, @arg);

  &putsock($FXsock{'irc'}, ":%s %s %s :\cA%s\cA", $FXbot{'GameServ'}{'nick'}, $notice ? "NOTICE" : "PRIVMSG", $nick, $buffer);
  return(0);
}

sub process_gs {
  my ($nick, $query) = (shift, shift);

  if (&check_flood($nick)) { return(0); }

  if ($query =~ /\cA(.*)\cA/) {
    my $ctcp = uc((split(/ /, $1, 0))[0]);

    if ($ctcp eq "VERSION") {
      &gs_ctcp($nick, 1, "VERSION %s", "FXserv IRC Services $FXserv{'version'}");
      &gs_ctcp($nick, 1, "VERSION %s", "Administrator: $FXserv{'admin'}");
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $FXbot{'GameServ'}{'nick'}, $ctcp, $FXirc{'client'}{lc($nick)}{'nick'}, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    } elsif ($ctcp eq "PING") {
      &gs_ctcp($nick, 1, "PING %s", (split(/ /, $1, 0))[1]);
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $FXbot{'GameServ'}{'nick'}, $ctcp, $FXirc{'client'}{lc($nick)}{'nick'}, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    } elsif ($ctcp eq "TIME") {
      &gs_ctcp($nick, 1, "TIME %s", scalar localtime());
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $FXbot{'GameServ'}{'nick'}, $ctcp, $FXirc{'client'}{lc($nick)}{'nick'}, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    } else {
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $FXbot{'GameServ'}{'nick'}, $1, $FXirc{'client'}{lc($nick)}{'nick'}, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    }

    return(0);
  }

  my $cmd = (split(/ /, $query, 0))[0];
  my $arg = (split(/ /, $query, 2))[1];

  if ($FXbot{'GameServ'}{'sub'}->{uc($cmd)}) {
    &{$FXbot{'GameServ'}{'sub'}->{uc($cmd)}}($nick, $arg);
  } else {
    &gs_notice($nick, $FXmsg{'unkn_cmd'}, $cmd);
  }

  return(0);
}

sub gs_help {
  my ($nick, $arg) = (shift, shift);

  if (!$arg) { $arg = "index"; }

  my $file = lc($arg);
  $file =~ s/ /\//g;
  $file = "help/" . $LangPack . "/gameserv/" . $file;

  if (!open(FILE, "<" . $file)) {
    &gs_notice($nick, $FXmsg{'no_help'}, $arg);
    return(0);
  }

  my @lines = <FILE>;
  close(FILE);

  foreach my $buffer (@lines) {
    chomp($buffer);
    $buffer = &substitute($buffer);

    if ($buffer =~ /^\%\+a/) {
      if (&valid_oper($nick, "a")) {
        $buffer =~ s/^\%\+a//;
        &gs_notice($nick, $buffer);
      }
    } elsif ($buffer =~ /^\%\+o/) {
      if (&valid_oper($nick, "o")) {
        $buffer =~ s/^\%\+o//;
        &gs_notice($nick, $buffer);
      }
    } else {
      &gs_notice($nick, $buffer);
    }
  }

  &servicelog($nick, $FXbot{'GameServ'}{'nick'}, "HELP %s", $arg);
  return(0);
}

sub gs_list {
  my ($nick, $arg) = (shift, shift);
  my $idx = 0;

  my $pid = undef;

  if ($pid = fork()) {
  } elsif (defined($pid)) {
    &gs_notice($nick, "-- Listing servers --");
    &gs_notice($nick, "[Server               ] [Game        ] [Players] [Map          ] [Ping]");

    foreach (keys(%FXgames)) {
      if ($arg && ($arg ne $FXgames{$_}{'type'})) { next; }

      my $info = ();

      if ($FXgames{$_}{'type'} eq "valve") {
        $info = &valve_info($FXgames{$_}{'host'}, $FXgames{$_}{'port'});
      } elsif ($FXgames{$_}{'type'} =~ /^(unreal|bf1942|army)$/) {
        my $add = 0;

        if ($FXgames{$_}{'type'} eq "unreal") { $add = 10; }
        if ($FXgames{$_}{'type'} eq "bf1942") { $add = 8433; }
        if ($FXgames{$_}{'type'} eq "army") { $add = 1; }

        $info = &gamespy_info($FXgames{$_}{'host'}, $FXgames{$_}{'port'} + $add);
      } elsif ($FXgames{$_}{'type'} =~ /^(quake3|duty|rtcw|moh)$/) {
        $info = &quake3_info($FXgames{$_}{'host'}, $FXgames{$_}{'port'});
      }

      if (!$info) { next; }
 
      ++$idx;
      &gs_notice($nick, "[%-21s] [%-12s] [ %02d/%02d ] [%-13s] [%-4d]", $FXgames{$_}{'server'}, $info->{'game'}, $info->{'players'}, $info->{'slots'}, $info->{'map'}, $info->{'ping'} ? $info->{'ping'} : "DOWN");
    }

    &gs_notice($nick, "-- End of list [%d/%d repl%s] --", $idx, scalar(keys(%FXgames)), $idx > 1 ? "ies" : "y");

    if (!$arg) { $arg = "*"; }

    &servicelog($nick, $FXbot{'GameServ'}{'nick'}, "LIST %s", $arg);
    exit(0);
  }
}

sub gs_info {
  my ($nick, $arg) = (shift, shift);

  my $server = (split(/ /, $arg, 0))[0];
  my $type = (split(/ /, $arg, 0))[1];
  my $host = (split(/:/, $server, 0))[0];
  my $port = (split(/:/, $server, 0))[1];

  if (!$type) { $type = "valve"; }

  if (!$server || !&valid_port($port)) {
    &gs_notice($nick, $FXmsg{'syntax'}, "INFO <server:port> [type]");
    return(0);
  }

  my $pid = undef;

  if ($pid = fork()) {
  } elsif (defined($pid)) {
    my $info = ();
    my %players = ();

    if (lc($type) eq "valve") {
      $info = &valve_info($host, $port);
      %players = &valve_players($host, $port);

    } elsif (lc($type) =~ /^(unreal|bf1942|army)$/) {
      my $add = 0;

      if (lc($type) eq "unreal") { $add = 10; }
      if (lc($type) eq "bf1942") { $add = 8433; }
      if (lc($type) eq "army") { $add = 1; }

      $info = &gamespy_info($host, $port + $add);
      %players = &gamespy_players($host, $port + $add, lc($type));

    } elsif (lc($type) =~ /^(quake3|rtcw|moh|duty)$/) {
      $info = &quake3_info($host, $port);
      %players = &quake3_players($host, $port);
    }

    if (!$info) {
      &gs_notice($nick, $FXmsg{'serv_down'});
      exit(0);
    }

    &gs_notice($nick, "   Name: %s", $info->{'name'});
    &gs_notice($nick, " Server: %s:%d", $host, $port);
    &gs_notice($nick, "   Game: %s", defined($game_type{$info->{'game'}}) ? $game_type{$info->{'game'}} : $info->{'game'});
    &gs_notice($nick, "    Map: %s", $info->{'map'});
    &gs_notice($nick, "Players: %d/%d", $info->{'players'}, $info->{'slots'});
    &gs_notice($nick, "   Ping: %d", $info->{'ping'});

    if (scalar(keys(%players))) {
      &gs_notice($nick, " ");
      &gs_notice($nick, "-- Listing players --");

      my @player = sort { $players{$b}{'frags'} <=> $players{$a}{'frags'} } keys(%players);

      if (lc($type) eq "valve") {
        &gs_notice($nick, "[Name                ] [Frags] [Time      ]");
        foreach (@player) { &gs_notice($nick, "[%-20s] [%-5d] [%-10s]", $players{$_}{'name'}, $players{$_}{'frags'}, &format_time($ts - $players{$_}{'time'})); }
      } elsif (lc($type) =~ /^(quake3|duty|rtcw|moh)$/) {
        &gs_notice($nick, "[Name                ] [Frags] [Ping]");
        foreach (@player) { &gs_notice($nick, "[%-20s] [%-5d] [%-4s]", $players{$_}{'name'}, $players{$_}{'frags'}, $players{$_}{'ping'}); }
      } elsif (lc($type) eq "unreal") {
        &gs_notice($nick, "[Name                ] [Frags] [Deaths] [Team]");
        foreach (@player) { &gs_notice($nick, "[%-20s] [%-5d] [%-6d] [%-4s]", $players{$_}{'name'}, $players{$_}{'frags'}, $players{$_}{'deaths'}, $players{$_}{'team'} ? "Red" : "Blue"); }
      } elsif (lc($type) eq "bf1942") {
        &gs_notice($nick, "[Name                ] [Score] [Kills] [Deaths] [Team  ]");
        foreach (@player) { &gs_notice($nick, "[%-20s] [%-5d] [%-5d] [%-6d] [%-6s]", $players{$_}{'name'}, $players{$_}{'frags'}, $players{$_}{'kills'}, $players{$_}{'deaths'}, $players{$_}{'team'} == 1 ? "Axis" : "Allied"); }
      } elsif (lc($type) eq "army") {
        &gs_notice($nick, "[Name                ] [Honor] [Frags] [KIA ] [Leader] [Goal] [ROE ]");
        foreach (@player) { &gs_notice($nick, "[%-20s] [%-5d] [%-5d] [%-4d] [%-6s] [%-4s] [%-4s]", $players{$_}{'name'}, $players{$_}{'honor'}, $players{$_}{'frags'}, $players{$_}{'kia'}, $players{$_}{'leader'}, $players{$_}{'goal'}, $players{$_}{'roe'}); }
      }

      &gs_notice($nick, "-- End of list --");
    }

    &servicelog($nick, $FXbot{'GameServ'}{'nick'}, "INFO %s %s", $server, $type);
    exit(0);
  }
}

sub gs_server {
  my ($nick, $arg) = (shift, shift);

  if (!&valid_oper($nick, "o")) {
    &gs_notice($nick, $FXmsg{'no_priv'});
    return(0);
  }

  if (!$arg) {
    &gs_notice($nick, $FXmsg{'syntax'}, "SERVER <ADD|DEL|LIST> [server:port] [type]");
    return(0);
  }

  my $what = (split(/ /, $arg, 0))[0];
  my $server = (split(/ /, $arg, 0))[1];
  my $type = (split(/ /, $arg, 0))[2];
  my $port = (split(/:/, $server, 0))[1];

  if (uc($what) eq "ADD") {
    if (!$server || !$type || !&valid_port($port)) {
      &gs_notice($nick, $FXmsg{'syntax'}, "SERVER ADD <server:port> <type>");
      return(0);
    }

    if ($FXgames{lc($server)}) {
      &gs_notice($nick, $FXmsg{'game_in'}, $server, $type);
      return(0);
    }

    if (lc($type) !~ (/^(valve|unreal|quake3|bf1942|duty|rtcw|moh|army)$/)) {
      &gs_notice($nick, "Invalid game type %s", $type);
      return(0);
    }

    $FXgames{lc($server)}{'server'} = $server;
    $FXgames{lc($server)}{'host'} = (split(/:/, $server, 0))[0];
    $FXgames{lc($server)}{'port'} = (split(/:/, $server, 0))[1];
    $FXgames{lc($server)}{'type'} = $type;
    &gs_notice($nick, $FXmsg{'game_add'}, $server, $type);
    &servicelog($nick, $FXbot{'GameServ'}{'nick'}, "SERVER ADD %s %s", $server, $type);
    return(0);

  } elsif (uc($what) eq "DEL") {
    if (!$server || !&valid_port($port)) {
      &gs_notice($nick, $FXmsg{'syntax'}, "SERVER DEL <server:port>");
      return(0);
    }

    if (!$FXgames{lc($server)}) {
      &gs_notice($nick, $FXmsg{'no_game'}, $server);
      return(0);
    }

    delete($FXgames{lc($server)});
    &gs_notice($nick, $FXmsg{'game_del'}, $server);
    &servicelog($nick, $FXbot{'GameServ'}{'nick'}, "SERVER DEL %s", $server);
    return(0);

  } elsif (uc($what) eq "LIST") {
    &gs_notice($nick, "-- Listing servers --");
    &gs_notice($nick, "[Server               ] [Type      ]");

    foreach (keys(%FXgames)) { &gs_notice($nick, "[%-21s] [%-10s]", $FXgames{$_}{'server'}, $FXgames{$_}{'type'}); }
    my $idx = scalar(keys(%FXgames));
    &gs_notice($nick, "-- End of list [%d server%s found] --", $idx, $idx > 1 ? "s" : "");
    &servicelog($nick, $FXbot{'GameServ'}{'nick'}, "SERVER LIST");
    return(0);

  } else {
    &gs_notice($nick, $FXmsg{'syntax'}, "SERVER <ADD|DEL|LIST> [server:port] [type]");
    return(0);
  }
}

sub gs_rcon {
  my ($nick, $arg) = (shift, shift);

  my $server = (split(/ /, $arg, 0))[0];
  my $password = (split(/ /, $arg, 0))[1];
  my $rcon_host = (split(/:/, $server, 0))[0];
  my $rcon_port = (split(/:/, $server, 0))[1];

  if (!$server || !&valid_port($rcon_port) || !$password) {
    &gs_notice($nick, $FXmsg{'syntax'}, "RCON <server:port> <password>");
    return(0);
  }

  my $idx = 0;

  foreach (keys(%{$FXrcon{'list'}})) {
    if ($FXrcon{'list'}{$_}{'sock'}) {
      if (lc($FXrcon{'list'}{$_}{'nick'}) eq lc($nick)) {
        &gs_notice($nick, $FXmsg{'one_dcc'});
        return(0);
      }

      if (++$idx > $FXbot{'GameServ'}{'rcon_global'}) {
        &gs_notice($nick, $FXmsg{'dcc_max'});
        return(0);
      }

    }
  }

  foreach (keys(%{$FXrcon{'query'}})) {
    if ((lc($FXrcon{'query'}{$_}{'nick'}) eq lc($nick)) && ($ts - $FXrcon{'query'}{$_}{'time'} < $FXbot{'GameServ'}{'rcon_timeout'})) {
      &gs_notice($nick, $FXmsg{'rcon_wait'}, abs($ts - $FXrcon{'query'}{$_}{'time'} - $FXbot{'GameServ'}{'rcon_timeout'}));
      return(0);
    }
  }

  my @range = &parse_range($FXbot{'GameServ'}{'rcon_dcc'});
  my $socket = "";
  my $port = 0;

  foreach (@range) {
    $port = $_;
    if (!&valid_port($port)) { next; }

    $socket = &init_listen($port, "tcp", "RCON-A");
    if ($socket) {
      $FXsock{'list'}{$socket}{'nick'} = $nick;
      $FXsock{'list'}{$socket}{'server'} = $server;
      $FXsock{'list'}{$socket}{'password'} = $password;
      last;
    }
  }

  if (!$socket) {
    &putlog("y", "Unable to listen port %d", $port);
    &gs_notice($nick, $FXmsg{'dcc_failed'});
    return(0);
  }

  $FXrcon{'query'}{lc($nick)}{'nick'} = $nick;
  $FXrcon{'query'}{lc($nick)}{'time'} = $ts;

  my $longip = unpack("N", inet_aton($FXserv{'ip'}));

  &gs_notice($nick, "DCC Chat (%s)", $FXserv{'ip'});
  &gs_ctcp($nick, 0, "DCC CHAT chat %d %d", $longip, $port);

  &servicelog($nick, $FXbot{'GameServ'}{'nick'}, "RCON %s:%d", $rcon_host, $rcon_port);
  return(0);
}

sub rcon_parse {
  my ($socket, $buffer) = (shift, shift);

  if (&rcon_flood($socket)) { return(0); }

  $FXsock{'list'}{$socket}{'lastcmd'} = $ts;
  $FXsock{'list'}{$socket}{'warning'} = 0;

  my $ip = $FXsock{'list'}{$socket}{'ip'};
  my $host = (split(/:/, $FXrcon{'list'}{$ip}{'server'}, 0))[0];
  my $port = (split(/:/, $FXrcon{'list'}{$ip}{'server'}, 0))[1];
  my $password = $FXrcon{'list'}{$ip}{'password'};

  my $pid = undef;

  if ($pid = fork()) {
  } elsif (defined($pid)) {
    my $socket = &open_socket("udp", $host, $port, 3);
    my $recv = "";

    if (!$socket) {
      &putsock($FXrcon{'list'}{$ip}{'sock'}, "Error: unable to connect to %s:%d (udp)", $host, $port);
      return(0);
    }

    $socket->autoflush(1);

    print $socket "\xFF" x 4 . "challenge rcon\n\x00";

    recv($socket, $recv, 8192, 0);

    if ($recv =~ /^\xFF\xFF\xFF\xFFchallenge rcon (\d+)/) {
      print $socket "\xFF" x 4 . "rcon " . $1 . " " . $password . " " . $buffer . "\x00";
    } else {
      &putsock($FXrcon{'list'}{$ip}{'sock'}, "Error: server returned wrong message");
      exit(0);
    }

    recv($socket, $recv, 8192, 0);

    $recv =~ s/\x00+$//;
    $recv =~ s/^\xFF\xFF\xFF\xFFl//;
    $recv =~ s/^\xFF\xFF\xFF\xFFn//;
    $recv =~ s/^\xFF\xFF\xFF\xFF//;

    foreach (split(/\n/, $recv)) { &putsock($FXrcon{'list'}{$ip}{'sock'}, $_); }

    close($socket);
    exit(0);
  }

  return(0);
}

sub rcon_flood {
  my $socket = shift;

  if (($ts - $FXsock{'list'}{$socket}{'lastcmd'}) < $FXbot{'GameServ'}{'flood_time'}) {
    if (++$FXsock{'list'}{$socket}{'floodcnt'} < $FXbot{'GameServ'}{'flood_count'}) {
      $FXsock{'list'}{$socket}{'lastcmd'} = $ts;
      $FXsock{'list'}{$socket}{'warning'} = 0;
      return(0);
    } else {
      &putsock($socket, $FXmsg{'flood_rcon'}, $FXbot{'GameServ'}{'flood_time'}, $FXbot{'GameServ'}{'flood_count'});
      &putinfo("y", "%s: detected flood from %s\@%s (closing socket)", $FXbot{'GameServ'}{'nick'}, $FXsock{'list'}{$socket}{'nick'}, $FXsock{'list'}{$socket}{'ip'});
      &del_socket($socket);
      return(1);
    }
  } else {
    $FXsock{'list'}{$socket}{'lastcmd'}  = $ts;
    $FXsock{'list'}{$socket}{'floodcnt'} = 0;
    $FXsock{'list'}{$socket}{'warning'}  = 0;
    return(0);
  }
}

sub gs_who {
  my $nick = shift;

  if (!&valid_oper($nick, "o")) {
    &gs_notice($nick, $FXmsg{'no_priv'});
    return(0);
  }

  my $idx = 0;

  &gs_notice($nick, "-- Listing users --");
  &gs_notice($nick, "[Nickname       ] [IP             ] [Server               ] [Time      ]");

  foreach (keys(%{$FXrcon{'list'}})) {
    ++$idx;
    &gs_notice($nick, "[%-15s] [%-15s] [%-21s] [%-10s]", $FXrcon{'list'}{$_}{'nick'}, $FXrcon{'list'}{$_}{'ip'}, $FXrcon{'list'}{$_}{'server'}, &format_time($FXrcon{'list'}{$_}{'time'}));
  }

  &gs_notice($nick, "-- End of list [%d user%s found] --", $idx, $idx > 1 ? "s" : "");
  &servicelog($nick, $FXbot{'GameServ'}{'nick'}, "WHO");
  return(0);
}

sub gs_boot {
  my ($nick, $target) = (shift, shift);

  if (!&valid_oper($nick, "o")) {
    &gs_notice($nick, $FXmsg{'no_priv'});
    return(0);
  }

  if (!$target) {
    &gs_notice($nick, $FXmsg{'syntax'}, "BOOT <nick>");
    return(0);
  }

  foreach (keys %{$FXsock{'list'}}) {
    if (($FXsock{'list'}{$_}{'type'} eq "RCON") && (lc($FXsock{'list'}{$_}{'nick'}) eq lc($target))) {
      &gs_notice($nick, $FXmsg{'booted'}, $target);
      &servicelog($nick, $FXbot{'GameServ'}{'nick'}, "BOOT %s", $FXsock{'list'}{$_}{'nick'});
      &del_socket($FXsock{'list'}{$_}{'sock'});
      return(0);
    }
  }

  &gs_notice($nick, $FXmsg{'no_user'}, $target);
  return(0);
}

1;