use strict;

our (%FXserv, %FXgames, %FXrcon, %FXim, %FXscan, %FXfserv, %FXflags, %FXbot, %FXirc, %FXstats, %FXsock, %FXboard, %FXlisten, %FXbnet, $bit);

$FXflags{'monitor'} = { 'j' => 0x0001, 'k' => 0x0002, 'm' => 0x0004, 'n' => 0x0008, 'p' => 0x0010, 'q' => 0x0020, 't' => 0x0040, 'x' => 0x0080, 'y' => 0x0100 };
$FXflags{'event'} = { 'b' => 0x0001, 'C' => 0x0002, 'E' => 0x0004, 'j' => 0x0008, 'k' => 0x0010, 'l' => 0x0020, 'm' => 0x0040, 'n' => 0x0080, 'o' => 0x0100, 'O' => 0x0200, 'p' => 0x0400, 's' => 0x0800, 'S' => 0x1000, 't' => 0x2000, 'x' => 0x4000, 'y' => 0x8000 };
$FXflags{'channel'} = { 'c' => 0x0001, 'h' => 0x0002, 'i' => 0x0004, 'k' => 0x0008, 'l' => 0x0010, 'm' => 0x0020, 'n' => 0x0040, 'o' => 0x0080, 'p' => 0x0100, 'r' => 0x0200, 's' => 0x0400, 't' => 0x0800, 'v' => 0x1000 };
$FXflags{'oper'} = { 'a' => 0x0001, 'd' => 0x0002, 'f' => 0x0004, 'k' => 0x0008, 'o' => 0x0010 };
$FXflags{'client'} = { 'a' => 0x0001, 'g' => 0x0002, 'h' => 0x0004, 'i' => 0x0008, 'o' => 0x0010, 'p' => 0x0020, 'q' => 0x0040, 'r' => 0x0080, 's' => 0x0100, 'S' => 0x0200, 'w' => 0x0400, 'x' => 0x0800 };

sub cleanup_vars {
  $FXstats{'clients'}{'current'} = 0;
  $FXstats{'channels'}{'current'} = 0;
  $FXstats{'servers'}{'current'} = 0;
  $FXstats{'opers'}{'current'} = 0;

  if ($FXirc{'client'}) { delete($FXirc{'client'}); }
  if ($FXirc{'channel'}) { delete($FXirc{'channel'}); }
  if ($FXirc{'server'}) { delete($FXirc{'server'}); }

  return(0);
}

sub init_vars {
  &cleanup_vars();

  $FXstats{'clients'}{'max'} = 0;
  $FXstats{'clients'}{'max_ts'} = 0;
  $FXstats{'channels'}{'max'} = 0;
  $FXstats{'channels'}{'max_ts'} = 0;
  $FXstats{'servers'}{'max'} = 0;
  $FXstats{'server'}{'max_ts'} = 0;
  $FXstats{'opers'}{'max'} = 0;
  $FXstats{'oper'}{'max_ts'} = 0;
  $FXstats{'kills'}{'oper'} = 0;
  $FXstats{'kills'}{'server'} = 0;
  $FXstats{'bots'} = 0;
  $FXstats{'partyline'} = 0;

  $FXstats{'domain'} = ();
  $FXstats{'tld'} = ();
  $FXstats{'version'} = ();

  $FXirc{'client'} = ();
  $FXirc{'channel'} = ();
  $FXirc{'server'} = ();

  return(0);
}

sub init_connect {
  &connect_irc();

  if ($FXbot{'PageServ'}{'enabled'}) { &connect_im(); }

  foreach (keys(%FXbnet)) { &connect_bot($FXbnet{$_}{'nick'}); }

  return(0);
}

sub init_listen {
  foreach (keys(%FXlisten)) {
    if ($FXlisten{$_}{'listen'}) { next; }

    my $port = $FXlisten{$_}{'port'};
    
    my $listen = IO::Socket::INET->new(
      Proto => "tcp",
      LocalPort => $port,
      LocalAddr => $FXlisten{$_}{'host'},
      Listen => 1,
      Reuse => 1
    );

    if (!$listen) {
      &putlog("y", "Unable to listen port %d", $port);
      next;
    }

    my $type = "";

    if ($FXlisten{$_}{'type'} eq "bot") {
      $type = "BOT-A";
    } elsif ($FXlisten{$_}{'type'} eq "user") {
      $type = "DCC-A";
    }

    &add_socket($listen, $type);

    $FXlisten{$port}{'listen'} = 1;
  }

  return(0);
}

sub enable_service {
  my ($service, $init) = (shift, shift);

  if ($service eq "FileServ") { require "src/fileserv.pl"; }
  if ($service eq "FindServ") { require "src/findserv.pl"; }
  if ($service eq "NewsServ") { require "src/newsserv.pl"; }
  if ($service eq "MobiServ") { require "src/mobiserv.pl"; }
  if (($service eq "FindServ") || ($service eq "MobiServ") || ($service eq "NewsServ")) { require "src/http.pl"; }

  if ($service eq "ScanServ") {
    require "src/scanserv.pl";

    %FXscan = ();
    $FXbot{'ScanServ'}{'id'} = 0;

    if ($FXbot{'ScanServ'}{'scan_version'}) { &load_version(); }
  }

  if ($service eq "GameServ") {
    require "src/gameserv.pl";
    &load_games();
  }

  if ($service eq "PageServ") {
    require "src/pageserv.pl";

    &load_buddies();

    foreach my $im ("msn", "yahoo", "icq", "aim", "jabber") {
      if ($FXbot{'PageServ'}{$im . '_id'}) {
        require "src/" . $im . ".pl";
        $FXim{$im}{'attempt'} = 0;
        $FXim{$im}{'board'} = ();
      }
    }
  }

  $FXbot{$service}{'enabled'} = 1;

  &load_monitor($service);

  if (!$init && $FXserv{'connected'}) {
    &introduce_service($service);
    if ($FXbot{$service}{'channel'}) { &check_channel($service); }
  }

  return(0);
}

sub disable_service {
  my $service = shift;

  &save_conf();

  if ($FXserv{'connected'}) {
    &putsock($FXsock{'irc'}, ":%s QUIT :%s", $FXbot{$service}{'nick'}, "Service disabled");
    &add_seen($FXbot{$service}{'nick'}, "QUIT", "Service disabled");
    &chanlog("q", undef, $FXbot{$service}{'nick'}, "Service disabled", undef);
    &del_client($FXbot{$service}{'nick'});
  }

  $FXbot{$service}{'enabled'} = 0;

  if ($FXbot{$service}{'channel'}) { delete($FXbot{$service}{'channel'}); }

  foreach (%{$FXbot{$service}{'sub'}}) { undef($FXbot{$service}{'sub'}{$_}); }
  foreach (%{$FXbot{$service}{'other_sub'}}) { undef($FXbot{$service}{'other_sub'}{$_}); }

  if (exists($INC{"src/" . lc($service) . ".pl"})) { delete($INC{"src/" . lc($service) . ".pl"}); }

  if ($service eq "FileServ") {
    foreach (%{$FXbot{$service}{'fserv_sub'}}) { undef($FXbot{$service}{'fserv_sub'}{$_}); }
    %FXfserv = ();
  }

  if ($service eq "GameServ") {
    %FXgames = ();
    %FXrcon = ();
  }

  if ($service eq "PageServ") {
    my $im_sub = {
      'msn'    => 'msn_sub',
      'yahoo'  => 'yahoo_sub',
      'icq'    => 'icq_sub',
      'aim'    => 'aim_sub',
      'jabber' => 'jabber_sub'
    };

    foreach my $im ("msn", "yahoo", "icq", "aim", "jabber") {
      if (!$FXbot{'PageServ'}{$im . "_id"}) { next; }
      if ($FXsock{$im}) { &del_socket($FXsock{$im}); }
      foreach (%{$FXbot{$service}{$im_sub->{$im}}}) { undef($FXbot{$service}{$im_sub->{$im}}{$_}); }
      if (exists($INC{"src/" . $im . ".pl"})) { delete($INC{"src/" . $im . ".pl"}); }
    }

    %FXim = ();
    %FXboard = ();
  }

  return(0);
}

1;