use strict;

our (%FXbot, %FXserv, %FXirc, %FXmsg, %FXcfg, %FXsock, $LangPack);

$FXbot{'MobiServ'}{'sub'} = {
  'HELP' => \&ms_help,
  'SMS'  => \&ms_sms,
  'TXT'  => \&ms_sms,
};

$FXbot{'MobiServ'}{'public_sub'} = {
  'SMS' => \&ms_sms,
  'TXT' => \&ms_sms
};

$FXbot{'MobiServ'}{'other_sub'} = {
  'process_ms' => \&process_ms,
  'ms_notice'  => \&ms_notice,
  'ms_ctcp'    => \&ms_ctcp,
  'sms_us_att'      => \&sms_us_att,
  'sms_us_cingular' => \&sms_us_cingular,
  'sms_us_verizon'  => \&sms_us_verizon,
  'sms_us_sprint'   => \&sms_us_sprint,
  'sms_us_nextel'   => \&sms_us_nextel,
  'sms_us_tmobile'  => \&sms_us_tmobile
};

my $area = {
  'AT' => \&sms_us_att,
  'CI' => \&sms_us_cingular,
  'VZ' => \&sms_us_verizon,
  'SP' => \&sms_us_sprint,
  'NX' => \&sms_us_nextel,
  'TM' => \&sms_us_tmobile
};

sub ms_notice {
  my ($nick, $format, @arg) = (shift, shift, @_);
  my $buffer = sprintf($format, @arg);

  &putsock($FXsock{'irc'}, ":%s %s %s :%s", $FXbot{'MobiServ'}{'nick'}, $FXcfg{'general'}{'communicate'}, $nick, $buffer);
  return(0);
}

sub ms_ctcp {
  my ($nick, $notice, $format, @arg) = (shift, shift, @_);
  my $buffer = sprintf($format, @arg);

  &putsock($FXsock{'irc'}, ":%s %s %s :\cA%s\cA", $FXbot{'MobiServ'}{'nick'}, $notice ? "NOTICE" : "PRIVMSG", $nick, $buffer);
  return(0);
}

sub process_ms {
  my ($nick, $query) = (shift, shift);

  if (&check_flood($nick)) { return(0); }

  if ($query =~ /\cA(.*)\cA/) {
    my $ctcp = uc((split(/ /, $1, 0))[0]);

    if ($ctcp eq "VERSION") {
      &ms_ctcp($nick, 1, "VERSION %s", "FXserv IRC Services $FXserv{'version'}");
      &ms_ctcp($nick, 1, "VERSION %s", "Administrator: $FXserv{'admin'}");
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $FXbot{'MobiServ'}{'nick'}, $ctcp, $FXirc{'client'}{lc($nick)}{'nick'}, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    } elsif ($ctcp eq "PING") {
      &ms_ctcp($nick, 1, "PING %s", (split(/ /, $1, 0))[1]);
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $FXbot{'MobiServ'}{'nick'}, $ctcp, $FXirc{'client'}{lc($nick)}{'nick'}, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    } elsif ($ctcp eq "TIME") {
      &ms_ctcp($nick, 1, "TIME %s", scalar localtime());
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $FXbot{'MobiServ'}{'nick'}, $ctcp, $FXirc{'client'}{lc($nick)}{'nick'}, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    } else {
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $FXbot{'MobiServ'}{'nick'}, $1, $FXirc{'client'}{lc($nick)}{'nick'}, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    }

    return(0);
  }

  my $cmd = (split(/ /, $query, 0))[0];
  my $arg = (split(/ /, $query, 2))[1];


  if ($FXbot{'MobiServ'}{'sub'}->{uc($cmd)}) {
    &{$FXbot{'MobiServ'}{'sub'}->{uc($cmd)}}($nick, $arg);
  } else {
    &ms_notice($nick, $FXmsg{'unkn_cmd'}, $cmd);
  }

  return(0);
}

sub ms_help {
  my ($nick, $arg) = (shift, shift);

  if (!$arg) { $arg = "index"; }

  my $file = lc($arg);
  $file =~ s/ /\//g;
  $file = "help/" . $LangPack . "/mobiserv/" . $file;

  if (!open(FILE, "<" . $file)) {
    &ms_notice($nick, $FXmsg{'no_help'}, $arg);
    return(0);
  }

  my @lines = <FILE>;
  close(FILE);

  foreach my $buffer (@lines) {
    chomp($buffer);
    $buffer = &substitute($buffer);

    if ($buffer =~ /^\%\+a/) {
      if (&valid_oper($nick, "a")) {
        $buffer =~ s/^\%\+a//;
        &ms_notice($nick, $buffer);
      }
    } elsif ($buffer =~ /^\%\+o/) {
      if (&valid_oper($nick, "o")) {
        $buffer =~ s/^\%\+o//;
        &ms_notice($nick, $buffer);
      }
    } else {
      &ms_notice($nick, $buffer);
    }
  }

  &servicelog($nick, $FXbot{'MobiServ'}{'nick'}, "HELP %s", $arg);
  return(0);
}

sub ms_sms {
  my ($nick, $arg) = (shift, shift);

  if (!$arg) {
    &ms_notice($nick, $FXmsg{'syntax'}, "SMS <area> <gsm number> <message>");
    return(0);
  }

  my $prefix = (split(/ /, $arg, 0))[0];
  my $gsm = (split(/ /, $arg, 0))[1];
  my $sms = (split(/ /, $arg, 3))[2];

  if (!$prefix || !$gsm || ($gsm !~ /^\d+$/) || !$sms) {
    &ms_notice($nick, $FXmsg{'syntax'}, "SMS <area> <gsm number> <message>");
    return(0);
  }

  if ($area->{uc($prefix)}) {
    &{$area->{uc($prefix)}}($nick, $gsm, $sms);
  } else {
    &ms_notice($nick, $FXmsg{'unkn_area'}, $prefix);
  }

  return(0);
}

sub sms_us_att {
  my ($nick, $gsm, $sms) = (shift, shift, shift);

  if (length($gsm) != 10) {
    &ms_notice($nick, "$FXmsg{'bad_phone'}", $gsm);
    return(0);
  }

  if (length($sms) > 110) {
    &ms_notice($nick, $FXmsg{'sms_long'});
    $sms = substr($sms, 0, 110);
  }

  my $pid = undef;

  if ($pid = fork()) {
  } elsif (defined($pid)) {
    my $socket = &open_socket("tcp", "www.mymmode.com", 80, $FXbot{'MobiServ'}{'search_timeout'});
    my $result = 0;

    if (!$socket) {
      &ms_notice($nick, $FXmsg{'serv_down'});
      exit(0);
    }

    $socket->autoflush(1);

    eval {
      local $SIG{'ALRM'} = sub { die(); };
      alarm($FXbot{'MobiServ'}{'sms_timeout'});

      my $post = "";
      my $txt = $sms; $txt =~ tr/ /+/;

      $post .= "pin=" . substr($gsm, 0, 3) . "-" . substr($gsm, 3, 3) . "-" . substr($gsm, 6, 4);
      $post .= "\&from=" . $nick;
      $post .= "\&subject=";
      $post .= "\&message=" . $txt;

      &http_post($socket, "/messagecenter/pagersend.cgi", "www.mymmode.com", "http://www.mymmode.com/messagecenter/", $post, undef, undef);

      while (my $buffer = <$socket>) {
        $buffer =~ s/\s+$//; $buffer =~ s/^\s+//; $buffer =~ tr/ //s;

        if (($buffer =~ /^There has been an error sending your message\. Please check your message and try again\.<br><br>$/) || ($buffer =~ /is not a valid PCS phone number or PocketNet alias\./)) {
          &ms_notice($nick, "%s", $FXmsg{'sms_error'});
          $result = -1;
          last;
        } elsif ($buffer =~ /^Thank you for using AT\&amp;T. Your message has been submitted<br><br>$/) {
          &ms_notice($nick, "$FXmsg{'sms_sent'}", "+1", $gsm);
          $result = 1;
          last;
        }
      }

      alarm(0);
    };

    close($socket);

    if (!$result) { &ms_notice($nick, "%s", $FXmsg{'serv_down'}); }
    $gsm = substr($gsm, 0, length($gsm) - 2) . "XX";
    &servicelog($nick, $FXbot{'MobiServ'}{'nick'}, "SMS +1 %s (%d chars, %s)", $gsm, length($sms), $result == 1 ? "success" : "failed");
    exit(0);
  }
}

sub sms_us_cingular {
  my ($nick, $gsm, $sms) = (shift, shift, shift);
  my $result = 0;

  if (length($gsm) != 10) {
    &ms_notice($nick, "$FXmsg{'bad_phone'}", $gsm);
    return(0);
  }

  if ((length($sms) + length($nick)) > 160) {
    &ms_notice($nick, "%s", $FXmsg{'sms_long'});
    $sms = substr($sms, 0, 160);
  }

  my $pid = undef;

  if ($pid = fork()) {
  } elsif (defined($pid)) {
    my $socket = &open_socket("tcp", "208.62.68.135", 80, $FXbot{'MobiServ'}{'search_timeout'});

    if (!$socket) {
      &ms_notice($nick, "%s", $FXmsg{'serv_down'});
      exit(0);
    }

    $socket->autoflush(1);

    eval {
      local $SIG{'ALRM'} = sub { die(); };
      alarm($FXbot{'MobiServ'}{'sms_timeout'});

      my $request = "";
      my $txt = $sms; $txt =~ tr/ /+/;

      $request .= "TO=" . $gsm;
      $request .= "\&FROM=" . $nick;
      $request .= "\&MESSAGE=" . $txt;
      $request .= "\&SUBMIT=Send+";
      $request .= "\&COUNT=" . length($sms) + length($nick);
      $request .= "\&SUBJECT=";
      $request .= "\&CALLBACK=";
      $request .= "\&msg=FRM:" . $nick . "+MSG:" . $txt;
      $request .= "\&min=" . $gsm;

      &http_get($socket, "/msgresult.shtml?" . $request, "208.62.68.135", "http://www.cingular.com/sendamessage", undef);

      while (my $buffer = <$socket>) {
        if ($buffer =~ /^Your message was not sent \(8\)/) {
          &ms_notice($nick, "%s", $FXmsg{'sms_error'});
          $result = -1;
          last;
        } elsif ($buffer =~ /^Your message will be sent./) {
          &ms_notice($nick, "$FXmsg{'sms_sent'}", "+1", $gsm);
          $result = 1;
          last;
        }
      }

      alarm(0);
    };

    close($socket);

    if (!$result) { &ms_notice($nick, "%s", $FXmsg{'serv_down'}); }
    $gsm = substr($gsm, 0, length($gsm) - 2) . "XX";
    &servicelog($nick, $FXbot{'MobiServ'}{'nick'}, "SMS +1 %s (%d chars, %s)", $gsm, length($sms), $result == 1 ? "success" : "failed");
    exit(0);
  }
}

sub sms_us_verizon {
  my ($nick, $gsm, $sms) = (shift, shift, shift);

  if (length($gsm) != 10) {
    &ms_notice($nick, "$FXmsg{'bad_phone'}", $gsm);
    return(0);
  }

  if ((length($sms) + length($nick)) > 160) {
    &ms_notice($nick, "%s", $FXmsg{'sms_long'});
    $sms = substr($sms, 0, 160);
  }

  my $pid = undef;

  if ($pid = fork()) {
  } elsif (defined($pid)) {
    my $socket = &open_socket("tcp", "www.vtext.com", 80, $FXbot{'MobiServ'}{'search_timeout'});
    my $result = 0;
    my $cookie = "";

    if (!$socket) {
      &ms_notice($nick, "%s", $FXmsg{'serv_down'});
      exit(0);
    }

    $socket->autoflush(1);

    eval {
      local $SIG{'ALRM'} = sub { die(); };
      alarm($FXbot{'MobiServ'}{'sms_timeout'});

      &http_get($socket, "/customer_site/jsp/messaging_lo.jsp", "www.vtext.com", undef, undef);

      while (my $buffer = <$socket>) {
        if ($buffer =~ /^Set-cookie: JSESSIONID=(\S+);path=\//) {
          $cookie = $1;
          last;
        }
      }

      alarm(0);
    };

    close($socket);

    if (!$cookie) {
      &ms_notice($nick, "%s", $FXmsg{'sms_error'});
      exit(0);
    }

    $socket = &open_socket("tcp", "www.vtext.com", 80, $FXbot{'MobiServ'}{'search_timeout'});

    if (!$socket) {
      &ms_notice($nick, "%s", $FXmsg{'serv_down'});
      exit(0);
    }

    $socket->autoflush(1);
    my $location = "";

    eval {
      local $SIG{'ALRM'} = sub { die(); };
      alarm($FXbot{'MobiServ'}{'sms_timeout'});

      my $post = "";
      my $txt = $sms; $txt =~ tr/ /+/;

      $post .= "DOMAIN_NAME=\%40vtext.com";
      $post .= "\&min=" . $gsm;
      $post .= "\&subject=" . $nick;
      $post .= "\&text=" . $txt;
      $post .= "\&type=0";
      $post .= "\&callback=";
      $post .= "\&trackResponses=No";
      $post .= "\&sender=";
      $post .= "\&disclaimer_submit.x=0";
      $post .= "\&disclaimer_submit.y=0";

      &http_post($socket, "/customer_site/jsp/disclaimer.jsp", "www.vtext.com", "http://www.vtext.com/customer_site/jsp/disclaimer.jsp", $post, undef, "timeZoneOffset=-60; JSESSIONID=" . $cookie . "; sender=" . $nick);

      while (my $buffer = <$socket>) {
        if ($buffer =~ /^Location: (\S+)/) {
          $location = $1;
          last;
        }
      }

      alarm(0);
    };

    close($socket);

    if (!$location) {
      &ms_notice($nick, "%s", $FXmsg{'sms_error'});
      exit(0);
    }

    my $msgid = "";

    foreach (split(/\&/, $location)) { ($msgid) = $_ =~ /STATUS_MSGID=(\S+)/; }

    if ($msgid) {
      &ms_notice($nick, "$FXmsg{'sms_sent'}", "+1", $gsm);
      $result = 1;
    } else {
      &ms_notice($nick, "%s", $FXmsg{'sms_error'});
      $result = -1;
    }

    if (!$result) { &ms_notice($nick, "%s", $FXmsg{'serv_down'}); }
    $gsm = substr($gsm, 0, length($gsm) - 2) . "XX";
    &servicelog($nick, $FXbot{'MobiServ'}{'nick'}, "SMS +1 %s (%d chars, %s)", $gsm, length($sms), $result == 1 ? "success" : "failed");
    exit(0);
  }
}

sub sms_us_nextel {
  my ($nick, $gsm, $sms) = (shift, shift, shift);
  my $result = 0;

  if (length($gsm) != 10) {
    &ms_notice($nick, "$FXmsg{'bad_phone'}", $gsm);
    return(0);
  }

  if (length($sms) > 280 + length($nick) + 2) {
    &ms_notice($nick, "%s", $FXmsg{'sms_long'});
    $sms = substr($sms, 0, 280);
  }

  my $pid = undef;

  if ($pid = fork()) {
  } elsif (defined($pid)) {
    my $socket = &open_socket("tcp", "messaging.nextel.com", 80, $FXbot{'MobiServ'}{'search_timeout'});

    if (!$socket) {
      &ms_notice($nick, "%s", $FXmsg{'serv_down'});
      exit(0);
    }

    $socket->autoflush(1);

    eval {
      local $SIG{'ALRM'} = sub { die(); };
      alarm($FXbot{'MobiServ'}{'sms_timeout'});

      my $post = "";
      my $txt = $sms; $txt =~ tr/ /+/;

      $post .= "cmd=sendPage";
      $post .= "\&oneWayPTNs=" . $gsm;
      $post .= "\&twoWayPTNs=";
      $post .= "\&from=" . $nick;
      $post .= "\&subject=";
      $post .= "\&message=" . $txt;
      $post .= "\&count=" . (length($sms) + length($nick) + 2);
      $post .= "\&rep1=&rep2=&rep3=&rep4=&rep5=&rep6=&replynumber=&replyemail=&confaddr=";
      $post .= "\&x=0&y=0";
      $post .= "\&schedule=";

      &http_post($socket, "/cgi/iPageExt.dll", "messaging.nextel.com", "http://messaging.nextel.com/cgi/iPageExt.dll", $post, undef, undef);

      while (my $buffer = <$socket>) {
        if ($buffer =~ /<title>Nextel.com - An Error Has Occurred<\/title>/) {
          &ms_notice($nick, "%s", $FXmsg{'sms_error'});
          $result = -1;
          last;
        } elsif ($buffer =~ /<title>Nextel.com - Your Message Has Been Received<\/title>/) {
          &ms_notice($nick, "$FXmsg{'sms_sent'}", "+1", $gsm);
          $result = 1;
          last;
        }
      }

      alarm(0);
    };

    close($socket);

    if (!$result) { &ms_notice($nick, "%s", $FXmsg{'serv_down'}); }
    $gsm = substr($gsm, 0, length($gsm) - 2) . "XX";
    &servicelog($nick, $FXbot{'MobiServ'}{'nick'}, "SMS +1 %s (%d chars, %s)", $gsm, length($sms), $result == 1 ? "success" : "failed");
    exit(0);
  }
}

sub sms_us_sprint {
  my ($nick, $gsm, $sms) = (shift, shift, shift);

  if (length($gsm) != 10) {
    &ms_notice($nick, "$FXmsg{'bad_phone'}", $gsm);
    return(0);
  }

  if (length($sms) > 160) {
    &ms_notice($nick, "%s", $FXmsg{'sms_long'});
    $sms = substr($sms, 0, 160);
  }

  my $pid = undef;

  if ($pid = fork()) {
  } elsif (defined($pid)) {
    my $socket = &open_socket("tcp", "messaging.sprintpcs.com", 80, $FXbot{'MobiServ'}{'search_timeout'});
    my $result = 0;

    if (!$socket) {
      &ms_notice($nick, "%s", $FXmsg{'serv_down'});
      exit(0);
    }

    $socket->autoflush(1);

    my $cookie = "";

    eval {
      local $SIG{'ALRM'} = sub { die(); };
      alarm($FXbot{'MobiServ'}{'sms_timeout'});

      &http_get($socket, "/textmessaging", "messaging.sprintpcs.com", undef, undef);

      while (my $buffer = <$socket>) {
        if ($buffer =~ /^Set-Cookie: JSESSIONID=(\S+); path=\/textmessaging/) {
          $cookie = $1;
          last;
        }
      }

      alarm(0);
    };

    close($socket);

    if (!$cookie) {
      &ms_notice($nick, "%s", $FXmsg{'sms_error'});
      exit(0);
    }

    $socket = &open_socket("tcp", "messaging.sprintpcs.com", 80, $FXbot{'MobiServ'}{'search_timeout'});

    if (!$socket) {
      &ms_notice($nick, "%s", $FXmsg{'serv_down'});
      exit(0);
    }

    $socket->autoflush(1);

    eval {
      local $SIG{'ALRM'} = sub { die(); };
      alarm($FXbot{'MobiServ'}{'sms_timeout'});

      my $post = "";
      my $txt = $sms; $txt =~ tr/ /+/;

      $post .= "phoneNumber=" . $gsm;
        $post .= "\&message=" . $txt;
      $post .= "\&characters=" . (160 - length($sms));
      $post .= "\&callBackNumber=";
      $post .= "\&x=0&y=0";

      &http_post($socket, "/textmessaging/composeconfirm", "messaging.sprintpcs.com", "http://messaging.sprintpcs.com/textmessaging/compose;jsessionid=" . $cookie, $post, undef, "JSESSIONID=" . $cookie);

      while (my $buffer = <$socket>) {
        if ($buffer =~ /^<li>This recipient cannot receive SMS$/) {
          &ms_notice($nick, "%s", $FXmsg{'sms_error'});
          $result = -1;
          last;
        } elsif ($buffer =~ /<p>This message is being transmitted now, but has not yet reached its destination. Use the tracking number to confirm receipt of the message.<\/p>$/) {
          &ms_notice($nick, "$FXmsg{'sms_sent'}", "+1", $gsm);
          $result = 1;
          last;
        }
      }

      alarm(0);
    };

    close($socket);

    if (!$result) { &ms_notice($nick, "%s", $FXmsg{'serv_down'}); }
    $gsm = substr($gsm, 0, length($gsm) - 2) . "XX";
    &servicelog($nick, $FXbot{'MobiServ'}{'nick'}, "SMS +1 %s (%d chars, %s)", $gsm, length($sms), $result == 1 ? "success" : "failed");
    exit(0);
  }
}

sub sms_us_tmobile {
  my ($nick, $gsm, $sms) = (shift, shift, shift);

  if (length($gsm) != 10) {
    &ms_notice($nick, "$FXmsg{'bad_phone'}", $gsm);
    return(0);
  }

  if (length($sms) > 140) {
    &ms_notice($nick, "%s", $FXmsg{'sms_long'});
    $sms = substr($sms, 0, 140);
  }

  my $pid = undef;

  if ($pid = fork()) {
  } elsif (defined($pid)) {
    my $socket = &open_socket("tcp", "www.t-mobile.com", 80, $FXbot{'MobiServ'}{'search_timeout'});
    my $result = 0;

    if (!$socket) {
      &ms_notice($nick, "%s", $FXmsg{'serv_down'});
      exit(0);
    }

    $socket->autoflush(1);

    eval {
      local $SIG{'ALRM'} = sub { die(); };
      alarm($FXbot{'MobiServ'}{'sms_timeout'});

      my $post = "";
      my $txt = $sms; $txt =~ tr/ /+/;

      $post .= "txtNum=" . $gsm;
      $post .= "\&txtFrom=" . $nick;
      $post .= "\&txtMessage=" . $txt;
      $post .= "\&hdnpublic=1&msgTermsUse=on";
      $post .= "\&x=0&y=0";

      &http_post($socket, "/messaging/default.asp?To=" . $gsm, "www.t-mobile.com", "http://www.t-mobile.com/messaging/default.asp?nav=hm", $post, undef, undef);

      while (my $buffer = <$socket>) {
        $buffer =~ s/\s+$//; $buffer =~ s/^\s+//; $buffer =~ tr/ //s;

        if ($buffer =~ /<strong>Your message could not be delivered to $gsm<\/strong><p>$/) {
          &ms_notice($nick, "%s", $FXmsg{'sms_error'});
          $result = -1;
          last;
        } elsif ($buffer =~ /<strong>Your message was sent to $gsm<\/strong><p>$/) {
          &ms_notice($nick, "$FXmsg{'sms_sent'}", "+1", $gsm);
          $result = 1;
          last;
        }
      }

      alarm(0);
    };

    close($socket);

    if (!$result) { &ms_notice($nick, "%s", $FXmsg{'serv_down'}); }
    $gsm = substr($gsm, 0, length($gsm) - 2) . "XX";
    &servicelog($nick, $FXbot{'MobiServ'}{'nick'}, "SMS +1 %s (%d chars, %s)", $gsm, length($sms), $result == 1 ? "success" : "failed");
    exit(0);
  }
}

1;