use strict;
use Digest::MD5 qw(md5_hex);

our (%FXim, %FXbot, %FXirc, %FXserv, %FXboard, %FXcfg, %FXflags, %FXsock, %FXmsg, $ts);

my %msn_error = (
  200 => 'Syntax error',
  201 => 'Invalid parameter',
  205 => 'Invalid user',
  206 => 'Domain name missing',
  207 => 'Already logged in',
  208 => 'Invalid user name',
  209 => 'Nickname change illegal',
  210 => 'User list full',
  215 => 'User already on list',
  216 => 'User not on list',
  217 => 'User not online',
  218 => 'Already in mode',
  219 => 'User is in the opposit list',
  223 => 'Too many groups',
  224 => 'Invalid group',
  225 => 'User not in group',
  229 => 'Group name too long',
  230 => 'Cannot remove group zero',
  280 => 'Switchboard failed',
  281 => 'Transfer to switchboard failed',

  300 => 'Required field missing',
  302 => 'Not logged in',

  500 => 'Internal server error',
  501 => 'Database server error',
  510 => 'File operation failed',
  520 => 'Memory allocation failed',
  540 => 'Challenge response failed',

  600 => 'Server is busy',
  601 => 'Server is unavaliable',
  602 => 'Peer nameserver is down',
  603 => 'Database connection failed',
  604 => 'Server is going down',
  605 => 'Server is unavaliable',

  707 => 'Could not create connection',
  711 => 'Write is blocking',
  712 => 'Session is overloaded',
  713 => 'Calling too rapidly',
  714 => 'Too many sessions',
  715 => 'Not expected',
  717 => 'Bad friend file',
  731 => 'Not expected',

  800 => 'Changing too rapidly',

  910 => 'Server too busy',
  911 => 'Authentication failed',
  912 => 'Server too busy',
  913 => 'Not allowed when hidding',
  914 => 'Server is unavaliable',
  915 => 'Server is unavaliable',
  916 => 'Server is unavaliable',
  920 => 'Not accepting new users',
  921 => 'Server too busy',
  928 => 'Bad ticket'
);

my @msn_close = (540, 600 .. 605, 731, 910 .. 916, 920, 921, 928);

my %msn_status = (
  'NLN' => 'Online',
  'FLN' => 'Offline',
  'HDN' => 'Hidden',
  'BSY'	=> 'Busy',
  'IDL'	=> 'Idle',
  'BRB'	=> 'Be Right Back',
  'AWY'	=> 'Away',
  'PHN' => 'On the Phone',
  'LUN' => 'Out to Lunch'
);

sub msn_connect {
  my ($server, $port, $ns) = (shift, shift, shift);

  $FXim{'msn'}{'attempt'} = $ts;
  $FXim{'msn'}{'ping'} = $ts;
  $FXim{'msn'}{'board'} = ();
  $FXim{'msn'}{'ns'} = 0;
  $FXsock{'msn'} = "";

  if (!$ns) { $FXim{'msn'}{'trid'} = 1; }

  my $socket = &open_socket("tcp", $server, $port, 3);

  if (!$socket) {
    &putinfo("y", "Unable to connect to MSN server %s:%d", $server, $port);
    return(0);
  }

  &add_socket($socket, "MSN");

  if ($ns) {
    $FXim{'msn'}{'ns'} = 1;
    &putinfo("y", "Connected to MSN server %s:%d", $server, $port);
  }

  &putsock($socket, "VER %d MSNP9 MSNP8 CVR0", $FXim{'msn'}{'trid'}++);

  return(0);
}

sub msnx_connect {
  my ($server, $port, $session, $trid) = (shift, shift, shift, shift);

  if (!$FXboard{'msn'}{$trid}) { return(0); }

  my $id = $FXboard{'msn'}{$trid}{'buddy'};

  my $socket = &open_socket("tcp", $server, $port, 3);

  if (!$socket) {
    &putinfo("y", "Unable to connect to MSN Switchboard server %s:%d", $server, $port);
    foreach my $nick (@{$FXboard{'msn'}{$trid}{'nick'}}) { &putsock($FXsock{'irc'}, ":%s %s %s :Unexpected error", $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'}, $FXcfg{'general'}{'communicate'}, $nick); }
    delete($FXboard{'msn'}{$trid});
    return(0);
  }

  &add_socket($socket, "MSN-X");
  $FXsock{'list'}{$socket}{'session'} = $session;
  $FXboard{'msn'}{$trid}{'session'} = $session;

  &putsock($socket, "USR %d %s %s", $FXim{'msn'}{'trid'}++, $FXbot{'PageServ'}{'msn_id'}, $session);
  return(0);
}

sub msn_board {
  my ($session, $socket, $buffer) = (shift, shift, shift);

  my $cmd = (split(/ /, $buffer, 0))[0];
  my $key = (split(/ /, $buffer, 0))[2];

  if (($cmd =~ /^\d{3}$/) && $msn_error{$cmd}) {
    &putinfo("y", "MSN error: $msn_error{$cmd} ($cmd)");

    if (grep($_ eq $cmd, @msn_close)) {
      foreach my $trid (keys(%{$FXboard{'msn'}})) {
        if ($FXboard{'msn'}{$trid}{'session'} eq $session) {
          my $id = $FXboard{'msn'}{$trid}{'buddy'};
          foreach my $nick (@{$FXboard{'msn'}{$trid}{'nick'}}) { &putsock($FXsock{'irc'}, ":%s %s %s :Unexpected error", $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'}, $FXcfg{'general'}{'communicate'}, $nick); }
          delete($FXboard{'msn'}{$trid});
          last;
        }
      }

      &del_socket($FXim{'msn'}{'board'}{$session}{'id'});
    }

    return(0);
  }

  if ($cmd eq "USR") {
    if ($key eq "OK") {
      foreach my $trid (keys(%{$FXboard{'msn'}})) {
        if ($FXboard{'msn'}{$trid}{'session'} eq $session) {
          &putsock($socket, "CAL %d %s", $FXim{'msn'}{'trid'}++, $FXboard{'msn'}{$trid}{'buddy'});
          last;
        }
      }

      return(0);

    } else {
      foreach my $trid (keys(%{$FXboard{'msn'}})) {
        if ($FXboard{'msn'}{$trid}{'session'} eq $session) {
          my $id = $FXboard{'msn'}{$trid}{'buddy'};
          foreach my $nick (@{$FXboard{'msn'}{$trid}{'nick'}}) { &putsock($FXsock{'irc'}, ":%s %s %s :Unexpected error", $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'}, $FXcfg{'general'}{'communicate'}, $nick); }
          delete($FXboard{'msn'}{$trid});
          last;
        }
      }

      return(0);
    }

  } elsif ($cmd eq "BYE") {
    &msn_remsession($session, "*");
    return(0);

  } elsif ($cmd eq "CAL") {
    if ($key ne "RINGING") {
      foreach my $trid (keys(%{$FXboard{'msn'}})) {
        if ($FXboard{'msn'}{$trid}{'session'} eq $session) {
          my $id = $FXboard{'msn'}{$trid}{'buddy'};
          foreach my $nick (@{$FXboard{'msn'}{$trid}{'nick'}}) { &putsock($FXsock{'irc'}, ":%s %s %s :Unexpected error", $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'}, $FXcfg{'general'}{'communicate'}, $nick); }
          delete($FXboard{'msn'}{$trid});
          last;
        }
      }

      return(0);
    }

  } elsif ($cmd eq "JOI") {
    foreach my $trid (keys(%{$FXboard{'msn'}})) {
      if ($FXboard{'msn'}{$trid}{'session'} eq $session) {
        my $id = $FXboard{'msn'}{$trid}{'buddy'};

        $FXim{'msn'}{'board'}{$session}{'session'} = $session;
        $FXim{'msn'}{'board'}{$session}{'id'} = $socket;
        $FXim{'msn'}{'board'}{$session}{'buddy'} = $id;
        @{$FXim{'msn'}{'board'}{$session}{'nick'}} = ();
        $FXim{'msn'}{'board'}{$session}{'channel'} = "#msn" . "." . $ts;
        $FXim{'msn'}{'board'}{$session}{'start'} = $ts;
        $FXim{'msn'}{'board'}{$session}{'last'} = $ts;

        &msn_sjoin($id, $FXim{'msn'}{'board'}{$session}{'channel'}, 1);

        foreach my $nick (@{$FXboard{'msn'}{$trid}{'nick'}}) {
          &putsock($FXsock{'irc'}, ":%s %s %s :$FXmsg{'buddy_request'}", $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'}, $FXcfg{'general'}{'communicate'}, $nick, $FXim{'msn'}{'board'}{$session}{'channel'});
          &putsock($FXsock{'irc'}, ":%s INVITE %s %s", $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'}, $nick, $FXim{'msn'}{'board'}{$session}{'channel'});
        }

        delete($FXboard{'msn'}{$trid});
        last;
      }

      return(0);
    }

  } elsif ($cmd eq "MSG") {
    $FXim{'msn'}{'board'}{$session}{'last'} = $ts;

    my @data = ();
    foreach (split(/\r\n/, $buffer)) { push(@data, $_); }

    my $idx = 0;
    my $cnt = 0;
    my $text = 0;

    foreach my $line (@data) {
      if (substr($line, 0, 24) eq "Content-Type: text/plain") { $text = 1; }
      if (($line eq "") && $text) {
        $idx = $cnt + 1;
        last;
      }

      $cnt++;
    }

    if ($idx) {
      my $id = $FXim{'msn'}{'board'}{$session}{'buddy'};
      for (my $cnt = $idx; $cnt < $#data + 1; $cnt++) {
        utf8::decode($data[$cnt]);
        &putsock($FXsock{'irc'}, ":%s PRIVMSG %s :%s", $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'}, $FXim{'msn'}{'board'}{$session}{'channel'}, $data[$cnt]);
      }
    }

    return(0);
  }

  return(0);
}

sub msn_id {
  my ($host, $auth) = (shift, shift);
  my $info = "";

  my $socket = &open_socket("tcp", $host, 80, 3);
  if (!$socket) { return(0); }

  &putsock($socket, "GET /login2.srf HTTP/1.1\r");
  &putsock($socket, "Authorization: %s\r", $auth);
  &putsock($socket, "Host: %s\r", $host);
  &putsock($socket, "\r");

  while (my $line = <$socket>) {
    if ($line =~ /^Authentication-Info: (.*)/) {
      $info = $1;
      last;
    }
  }

  close($socket);

  if ($info =~ /Passport1.4 da-status=redir/) { $info = ""; }

  return($info);
}

sub msn_parse {
  my $buffer = shift;
  chomp($buffer);
  $buffer =~ s/\r$//;

  $FXim{'msn'}{'ping'} = $ts;

  my $cmd = (split(/ /, $buffer, 0))[0];
  my $key = (split(/ /, $buffer, 0))[2];

  if (($cmd =~ /^\d{3}$/) && $msn_error{$cmd}) {
    &putinfo("y", "MSN error: $msn_error{$cmd} ($cmd)");

    if (grep($_ eq $cmd, @msn_close)) { &del_socket($FXsock{'msn'}); }
    return(0);
  }

  if ($cmd eq "OUT") {
    my $error = (split(/ /, $buffer, 0))[1];
    if ($error eq "OTH") { &putinfo("y", "MSN: You have logged in from another computer"); }
    if ($error eq "SSD") { &putinfo("y", "MSN: Server going down for an upgrade"); }
    &del_socket($FXsock{'msn'});
    return(0);

  } elsif ($cmd eq "VER") {
    &putsock($FXsock{'msn'}, "CVR %d 0x0409 winnt 5.1 i386 MSNMSGR 6.0.0602 MSMSGS %s", $FXim{'msn'}{'trid'}++, $FXbot{'PageServ'}{'msn_id'});
    return(0);

  } elsif ($cmd eq "CVR") {
    &putsock($FXsock{'msn'}, "USR %d TWN I %s", $FXim{'msn'}{'trid'}++, $FXbot{'PageServ'}{'msn_id'});
    return(0);

  } elsif ($cmd eq "XFR") {
    if ($key eq "NS") {
      &del_socket($FXsock{'msn'});
      my $ns = (split(/ /, $buffer, 0))[3];
      my ($server, $port) = ((split(/:/, $ns, 0))[0], (split(/:/, $ns, 0))[1]);
      &msn_connect($server, $port, 1);
      return(0);
    } elsif ($key eq "SB") {
      my $trid = (split(/ /, $buffer, 0))[1];

      if ($FXboard{'msn'}{$trid}) {
        my $sb = (split(/ /, $buffer, 0))[3];
        my ($server, $port) = ((split(/:/, $sb, 0))[0], (split(/:/, $sb, 0))[1]);
        my $session = (split(/ /, $buffer, 0))[5];
        &msnx_connect($server, $port, $session, $trid);
        return(0);
      }
    }

  } elsif ($cmd eq "USR") {
    if ($key eq "TWN") {
      $FXim{'msn'}{'string'} = (split(/ /, $buffer, 0))[4];

      my $auth = "";
      my $info = "";

      my %arg = map { split(/=/, $_) } split(/,/, $FXim{'msn'}{'string'});

      $auth .= "Passport1.4 ";
      $auth .= "OrgVerb=GET,";
      $auth .= "OrgURL=" . $arg{'ru'} . ",";
      $auth .= "sign-in=" . $FXbot{'PageServ'}{'msn_id'} . ",";
      $auth .= "pwd=" . $FXbot{'PageServ'}{'msn_password'} . ",";
      $auth .= $FXim{'msn'}{'string'};

      my @passport = ("loginnet.passport.com", "login.passport.com");

      foreach my $host (@passport) {
        $info = &msn_id($host, $auth);
        if ($info) { last; }
      }

      if (!$info) {
        &putlog("y", "Unable to obtain ticket from .NET Passport");
        &del_socket($FXsock{'msn'});
        return(0);
      }

      my ($string) = $info =~ /from-PP='(t=.*)'/;

      &putsock($FXsock{'msn'}, "USR %d TWN S %s\r", $FXim{'msn'}{'trid'}++, $string);
      return(0);

    } elsif ($key eq "OK") {
      &putsock($FXsock{'msn'}, "SYN %d 0", $FXim{'msn'}{'trid'}++);
      return(0);
    }

  } elsif ($cmd eq "SYN") {
    my $display = $FXbot{'PageServ'}{'msn_display'}; $display =~ s/ /\%20/g;
    &putsock($FXsock{'msn'}, "CHG %d NLN 268435492", $FXim{'msn'}{'trid'}++);
    &putsock($FXsock{'msn'}, "REA %d %s %s", $FXim{'msn'}{'trid'}++, $FXbot{'PageServ'}{'msn_id'}, $display);
    return(0);

  } elsif ($cmd eq "CHG") {
    &putsock($FXsock{'msn'}, "PNG");
    return(0);

  } elsif ($cmd eq "CHL") {
    &putsock($FXsock{'msn'}, "QRY %d %s 32", $FXim{'msn'}{'trid'}++, 'msmsgs@msnmsgr.com');

    my $digest = md5_hex($key, 'Q1P7W2E4J9R8U3S5');
    print {$FXsock{'msn'}} $digest;
    &putsock($FXsock{'msn'}, "PNG");
    return(0);

  } elsif ($cmd eq "QRY") {
    #CHL ok
    return(0);

  } elsif ($cmd eq "QNG") {
    #pong!
    return(0);

  } elsif ($cmd eq "LST") {
    my $id  = (split(/ /, $buffer, 0))[1];
    my $info = (split(/ /, $buffer, 0))[2]; $info =~ s/\%20/ /g; utf8::decode($info);

    foreach (keys(%{$FXim{'msn'}{'buddies'}})) {
      if (lc($id) eq lc($FXim{'msn'}{'buddies'}{$_}{'id'})) {
        if (!$FXim{'msn'}{'buddies'}{$_}{'buddy'}) {
          $FXim{'msn'}{'buddies'}{$_}{'id'} = $id;
          $FXim{'msn'}{'buddies'}{$_}{'status'} = "";
          $FXim{'msn'}{'buddies'}{$_}{'online'} = 0;
          $FXim{'msn'}{'buddies'}{$_}{'seen'} = 0;
          $FXim{'msn'}{'buddies'}{$_}{'ircnick'} = "";
        }

        $FXim{'msn'}{'buddies'}{$_}{'info'} = $info;
        last;
      }
    }

    if (!$FXim{'msn'}{'buddies'}{lc($id)}) {
      $FXim{'msn'}{'buddies'}{lc($id)}{'id'} = $id;
      $FXim{'msn'}{'buddies'}{lc($id)}{'buddy'} = 0;
      $FXim{'msn'}{'buddies'}{lc($id)}{'info'} = $info;
      $FXim{'msn'}{'buddies'}{lc($id)}{'status'} = "";
      $FXim{'msn'}{'buddies'}{lc($id)}{'online'} = 0;
      $FXim{'msn'}{'buddies'}{lc($id)}{'seen'} = 0;
      $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'} = "";
    }

    return(0);

  } elsif ($cmd eq "ILN") {
    my $id  = (split(/ /, $buffer, 0))[3];
    my $status = (split(/ /, $buffer, 0))[2];
    my $info = (split(/ /, $buffer, 0))[4]; $info =~ s/\%20/ /g; utf8::decode($info);

    foreach (keys(%{$FXim{'msn'}{'buddies'}})) {
      if (lc($id) eq lc($FXim{'msn'}{'buddies'}{$_}{'id'})) {
        $FXim{'msn'}{'buddies'}{$_}{'online'} = 1;
        $FXim{'msn'}{'buddies'}{$_}{'info'} = $info;
        $FXim{'msn'}{'buddies'}{$_}{'status'} = $status;
        $FXim{'msn'}{'buddies'}{$_}{'seen'} = $ts;

        if ($FXserv{'connected'} && $FXim{'msn'}{'buddies'}{$_}{'buddy'}) { &msn_introduce($id); }
        last;
      }
    }

    return(0);

  } elsif ($cmd eq "NLN") {
    my $id  = (split(/ /, $buffer, 0))[2];
    my $status = (split(/ /, $buffer, 0))[1];
    my $info = (split(/ /, $buffer, 0))[3]; $info =~ s/\%20/ /g; utf8::decode($info);

    foreach (keys(%{$FXim{'msn'}{'buddies'}})) {
      if (lc($id) eq lc($FXim{'msn'}{'buddies'}{$_}{'id'})) {
        $FXim{'msn'}{'buddies'}{$_}{'online'} = 1;
        $FXim{'msn'}{'buddies'}{$_}{'info'} = $info;
        $FXim{'msn'}{'buddies'}{$_}{'status'} = $status;
        $FXim{'msn'}{'buddies'}{$_}{'seen'} = $ts;

        if ($FXserv{'connected'} && $FXim{'msn'}{'buddies'}{$_}{'buddy'} && !$FXim{'msn'}{'buddies'}{$_}{'ircnick'}) { &msn_introduce($id); }
        last;
      }
    }

    return(0);

  } elsif ($cmd eq "FLN") {
    my $id = (split(/ /, $buffer, 0))[1];

    foreach (keys(%{$FXim{'msn'}{'buddies'}})) {
      if ((lc($id) eq lc($FXim{'msn'}{'buddies'}{$_}{'id'})) && $FXim{'msn'}{'buddies'}{$_}{'buddy'}) {
        &msn_disconnect($id, "Signed off");
        last;
      }
    }

    return(0);
  }

  return(0);
}

sub msn_sjoin {
  my ($id, $channel, $session) = (shift, shift, shift);

  my $channel_ts = $ts;
  my $channel_mode = "";
  my $mode = '';
  my $prefix = "";

  my $nick = $FXim{'msn'}{'buddies'}{lc($id)}{'nick'};
  my $ircnick = $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'};

  if ($FXirc{'channel'}{lc($channel)}) {
    $channel_ts = $FXirc{'channel'}{lc($channel)}{'time'};
  } else {
    $mode |= $FXflags{'channel'}{'n'};
    $mode |= $FXflags{'channel'}{'t'};
    $channel_mode = "nt";

    if ($session) {
      $mode |= $FXflags{'channel'}{'i'};
      $mode |= $FXflags{'channel'}{'s'};
      $channel_mode .= "is";
    }

    $channel_ts = $ts;
    &add_channel($channel, $ts, $mode);
  }

  $FXirc{'channel'}{lc($channel)}{'names'}{lc($ircnick)}{'nick'} = $ircnick;
  $FXirc{'channel'}{lc($channel)}{'names'}{lc($ircnick)}{'mode'} = '';

  if ($session) {
    $prefix = "\@";
    $FXirc{'channel'}{lc($channel)}{'names'}{lc($ircnick)}{'mode'} |= $FXflags{'channel'}{'o'};
  }

  if (++$FXirc{'channel'}{lc($channel)}{'current'} > $FXirc{'channel'}{lc($channel)}{'max'}) {
    $FXirc{'channel'}{lc($channel)}{'max'} = $FXirc{'channel'}{lc($channel)}{'current'};
    $FXirc{'channel'}{lc($channel)}{'max_ts'} = $ts;
  }

  &add_seen($ircnick, "JOIN", $channel);
  &chanlog("j", $channel, $ircnick, undef, undef); 
  &putsock($FXsock{'irc'}, ":%s SJOIN %d %s +%s :%s%s", $FXserv{'name'}, $channel_ts, $channel, $channel_mode, $prefix, $ircnick);

  if ($session) {
    &chanlog("m", $channel, $FXserv{'name'}, "+o", $ircnick);

    if (!($FXirc{'channel'}{lc($channel)}{'mode'} & $FXflags{'channel'}{'i'})) {
      &channel_mode($ircnick, $channel, "+i", undef);
      &chanlog("m", $channel, $ircnick, "+i", undef);
    }

    if (!($FXirc{'channel'}{lc($channel)}{'mode'} & $FXflags{'channel'}{'s'})) {
      &channel_mode($ircnick, $channel, "+s", undef);
      &chanlog("m", $channel, $ircnick, "+s", undef);
    }
  }

  return(0);
}

sub msn_introduce {
  my $id = shift;

  my $nick = $FXim{'msn'}{'buddies'}{lc($id)}{'nick'};
  my $ircnick = $FXbot{'PageServ'}{'msn_prefix'}; $ircnick =~ s/\%/$nick/;

  if ($FXirc{'client'}{lc($ircnick)}) {
    for (my $idx = 1; 1; $idx++) {
      $ircnick = $ircnick . $idx;
      if (!$FXirc{'client'}{lc($ircnick)}) { last; }
    }
  }

  my $umode = ''; $umode |= $FXflags{'client'}{'i'};
  my $rand = "";

  $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'} = $ircnick;
  $FXim{'msn'}{'buddies'}{lc($id)}{'online'} = 1;

  for (my $idx = 0; $idx < 6; $idx++) { $rand .= chr(97 + int(rand(26))); }
  my $host = $rand . ".messenger.msn.com";
  my $info = $FXim{'msn'}{'buddies'}{lc($id)}{'info'};

  &introduce_nick($ircnick, $nick, $host, undef, $umode, $info);
  &add_client($ircnick, $nick, $host, $info, undef, $FXserv{'name'}, $ts, $umode);

  &msn_sjoin($id, $FXbot{'PageServ'}{'msn_channel'}, 0);
  return(0);
}

sub msn_disconnect {
  my ($id, $reason) = (shift, shift);

  my $ircnick = $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'};

  if ($ircnick && $FXserv{'connected'}) { 
    foreach my $session (keys(%{$FXim{'msn'}{'board'}})) {
      if (lc($FXim{'msn'}{'board'}{$session}{'buddy'}) eq lc($id)) {
        &msn_remsession($session, "*");
        last;
      }
    }

    &add_seen($ircnick, "QUIT", $reason);
    &chanlog("q", undef, $ircnick, $reason, undef); 
    &putsock($FXsock{'irc'}, ":%s QUIT :%s", $ircnick, $reason);
    &del_client($ircnick);
  }

  $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'} = "";
  $FXim{'msn'}{'buddies'}{lc($id)}{'seen'} = $ts;
  $FXim{'msn'}{'buddies'}{lc($id)}{'status'} = "";
  $FXim{'msn'}{'buddies'}{lc($id)}{'online'} = 0;

  return(0);
}

sub msn_addbuddy {
  my ($id, $nick) = (shift, shift);
  my $done = 0;

  foreach (keys(%{$FXim{'msn'}{'buddies'}})) {
    if ((lc($id) eq lc($FXim{'msn'}{'buddies'}{$_}{'id'})) && !$FXim{'msn'}{'buddies'}{$_}{'buddy'}) {
      $FXim{'msn'}{'buddies'}{lc($id)}{'buddy'} = 1;
      $FXim{'msn'}{'buddies'}{lc($id)}{'nick'} = $nick;

      if ($FXserv{'connected'} && $FXim{'msn'}{'buddies'}{lc($id)}{'online'}) { &msn_introduce($id); }

      $done = 1;
      last;
    }
  }

  if (!$done) {
    $FXim{'msn'}{'buddies'}{lc($id)}{'id'} = $id;
    $FXim{'msn'}{'buddies'}{lc($id)}{'nick'} = $nick;
    $FXim{'msn'}{'buddies'}{lc($id)}{'info'} = "";
    $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'} = "";
    $FXim{'msn'}{'buddies'}{lc($id)}{'online'} = 0;
    $FXim{'msn'}{'buddies'}{lc($id)}{'status'} = "";
    $FXim{'msn'}{'buddies'}{lc($id)}{'seen'} = 0;
    $FXim{'msn'}{'buddies'}{lc($id)}{'buddy'} = 1;
  }

  &putsock($FXsock{'msn'}, "ADD %d FL %s %s 0", $FXim{'msn'}{'trid'}++, $id, $id);
  &putsock($FXsock{'msn'}, "ADD %d AL %s %s", $FXim{'msn'}{'trid'}++, $id, $id);

  return(0);
}

sub msn_rembuddy {
  my $id = shift;

  &putsock($FXsock{'msn'}, "REM %d FL %s %s", $FXim{'msn'}{'trid'}++, $id, $id);

  &msn_disconnect($id, "Removed");
  $FXim{'msn'}{'buddies'}{lc($id)}{'buddy'} = 0;

  return(0);
}

sub msn_renamebuddy {
  my ($id, $nick) = (shift, shift);

  if ($FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'}) {
    my $newnick = $FXbot{'PageServ'}{'msn_prefix'}; $newnick =~ s/\%/$nick/;
    my $oldnick = $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'};

    &putsock($FXsock{'irc'}, ":%s NICK %s", $oldnick, $newnick);
    &add_seen($oldnick, "NICK", $newnick);
    &chanlog("n", undef, $oldnick, $newnick, undef); 
    &rename_client($oldnick, $newnick);

    $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'} = $newnick;
  }

  $FXim{'msn'}{'buddies'}{lc($id)}{'nick'} = $nick;

  return(0);
}

sub msn_text {
  my ($socket, $format, @arg) = (shift, shift, @_);
  my $buffer = sprintf($format, @arg);

  utf8::encode($buffer);

  my $payload = 19 + 41 + 50 + 2 + length($buffer);
  &putsock($socket, "MSG %d N %d\r", $FXim{'msn'}{'trid'}++, $payload);
  &putsock($socket, "MIME-Version: 1.0\r");
  &putsock($socket, "Content-Type: text/plain; charset=UTF-8\r");
  &putsock($socket, "X-MMS-IM-Format: FN=Arial; EF=; CO=; CS=0; PF=22\r");
  &putsock($socket, "\r");
  print $socket $buffer;

  return(0);
}

sub msn_privmsg {
  my ($nick, $id, $query) = (shift, shift, shift);

  if (&check_flood($nick)) { return(0); }

  if ($query =~ /\cA(.*)\cA/) {
    my $ctcp = uc((split(/ /, $1, 0))[0]);

    if ($ctcp eq "VERSION") {
      &buddy_ctcp($nick, $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'}, "VERSION %s", "FXserv IRC Services $FXserv{'version'}");
    } elsif ($ctcp eq "PING") {
      &buddy_ctcp($nick, $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'}, "PING %s", (split(/ /, $1, 0))[1]);
    } elsif ($ctcp eq "TIME") {
      &buddy_ctcp($nick, $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'}, "TIME %s", scalar(localtime()));
    }

    return(0);
  }

  my $done = 0;

  foreach my $session (keys(%{$FXim{'msn'}{'board'}})) {
    if (lc($FXim{'msn'}{'board'}{$session}{'buddy'}) eq lc($id)) {
      if (grep(lc($_) eq lc($nick), @{$FXim{'msn'}{'board'}{$session}{'nick'}})) {
        &putsock($FXsock{'irc'}, ":%s %s %s :$FXmsg{'buddy_channel'}", $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'}, $FXcfg{'general'}{'communicate'}, $nick, $FXim{'msn'}{'board'}{$session}{'channel'});
        $done = 1;
        last;
      } else {
        &putsock($FXsock{'irc'}, ":%s %s %s :$FXmsg{'buddy_request'}", $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'}, $FXcfg{'general'}{'communicate'}, $nick, $FXim{'msn'}{'board'}{$session}{'channel'});
        &putsock($FXsock{'irc'}, ":%s INVITE %s %s", $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'}, $nick, $FXim{'msn'}{'board'}{$session}{'channel'});
        $done = 1;
        last;
      }
    }
  }

  if (!$done) {
    foreach my $trid (keys(%{$FXboard{'msn'}})) {
      if (grep(lc($_) eq lc($nick), @{$FXboard{'msn'}{$trid}{'nick'}})) { return(0); }
    }

    foreach my $trid (keys(%{$FXboard{'msn'}})) {
      if (lc($FXboard{'msn'}{$trid}{'buddy'}) eq lc($id)) {
        push(@{$FXboard{'msn'}{$trid}{'nick'}}, $nick);
        &putsock($FXsock{'irc'}, ":%s %s %s :$FXmsg{'buddy_invite'}", $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'}, $FXcfg{'general'}{'communicate'}, $nick);
        return(0);
      }
    }

    my $trid = $FXim{'msn'}{'trid'}++;
    $FXboard{'msn'}{$trid}{'buddy'} = $id;
    @{$FXboard{'msn'}{$trid}{'nick'}} = ();
    $FXboard{'msn'}{$trid}{'trid'} = $trid;
    $FXboard{'msn'}{$trid}{'session'} = "";
    push(@{$FXboard{'msn'}{$trid}{'nick'}}, $nick);

    &putsock($FXsock{'msn'}, "XFR %d SB", $trid);
    &putsock($FXsock{'irc'}, ":%s %s %s :%s", $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'}, $FXcfg{'general'}{'communicate'}, $nick, $FXmsg{'buddy_invite'});

    return(0);
  }

  return(0);
}

sub msn_remsession {
  my ($session, $nick) = (shift, shift);
  my $id = $FXim{'msn'}{'board'}{$session}{'buddy'};
  my $socket = $FXim{'msn'}{'board'}{$session}{'id'};

  if ($nick && ($nick ne "*")) {
    my $cnt = 0;

    foreach my $user (@{$FXim{'msn'}{'board'}{$session}{'nick'}}) {
      if (lc($nick) eq lc($user)) {
        splice(@{$FXim{'msn'}{'board'}{$session}{'nick'}}, $cnt, 1);
        &msn_text($FXim{'msn'}{'board'}{$session}{'id'}, $FXmsg{'buddy_part'}, $nick);
        last;
      }

      $cnt++;
    }
  }

  if (($nick eq "*") || !scalar(@{$FXim{'msn'}{'board'}{$session}{'nick'}})) {
    my $ircnick = $FXim{'msn'}{'buddies'}{lc($id)}{'ircnick'};
    my $channel = $FXim{'msn'}{'board'}{$session}{'channel'};
    &putsock($FXsock{'irc'}, ":%s PART %s", $ircnick, $channel);

    &add_seen($ircnick, "PART", $channel);
    &chanlog("p", $channel, $ircnick, undef, undef);

    if (scalar(@{$FXim{'msn'}{'board'}{$session}{'nick'}})) {
      foreach my $nick (@{$FXim{'msn'}{'board'}{$session}{'nick'}}) { &msn_text($socket, $FXmsg{'buddy_part'}, $nick); }
    }

    delete($FXirc{'channel'}{lc($channel)}{'names'}{lc($ircnick)});
    &del_channel($channel);

    delete($FXim{'msn'}{'board'}{$session});
    &putsock($socket, "OUT");
    &del_socket($socket);
  }

  return(0);
}

1;