use strict;

our (%FXbot, %FXserv, %FXim, %FXirc, %FXsock, %FXoper, %FXflags, $bit, $ConfigFile, $PidFile, $ts);

sub substitute {
  my $string = shift;

  $string =~ s/\%D/$FXbot{'DataServ'}{'nick'}/g;
  if ($FXbot{'FileServ'}{'enabled'}) { $string =~ s/\%X/$FXbot{'FileServ'}{'nick'}/g; }
  if ($FXbot{'FindServ'}{'enabled'}) { $string =~ s/\%F/$FXbot{'FindServ'}{'nick'}/g; }
  if ($FXbot{'NewsServ'}{'enabled'}) { $string =~ s/\%N/$FXbot{'NewsServ'}{'nick'}/g; }
  if ($FXbot{'MobiServ'}{'enabled'}) { $string =~ s/\%M/$FXbot{'MobiServ'}{'nick'}/g; }
  if ($FXbot{'PageServ'}{'enabled'}) { $string =~ s/\%P/$FXbot{'PageServ'}{'nick'}/g; }
  if ($FXbot{'GameServ'}{'enabled'}) { $string =~ s/\%G/$FXbot{'GameServ'}{'nick'}/g; }
  if ($FXbot{'ScanServ'}{'enabled'}) { $string =~ s/\%S/$FXbot{'ScanServ'}{'nick'}/g; }

  $string =~ s/\%a/$FXserv{'admin'}/g;
  $string =~ s/\%v/$FXserv{'version'}/g;
  $string =~ s/\%b/\002/g;

  my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime($ts);
  $year += 1900;

  $mday = sprintf("%02d", $mday);
  $mon = sprintf("%02d", ++$mon);
  $hour = sprintf("%02d", $hour);
  $min = sprintf("%02d", $min);
  $sec = sprintf("%02d", $sec);

  $string =~ s/\%d/$mday/g;
  $string =~ s/\%m/$mon/g;
  $string =~ s/\%y/$year/g;
  $string =~ s/\%h/$hour/g;
  $string =~ s/\%i/$min/g;
  $string =~ s/\%s/$sec/g;

  return($string);
}

sub parse_range {
  my ($list) = @_;
  my @range = ();
  my @result = ();

  $list =~ s/\s+//g;

  foreach my $port (split(/,/ , $list)) {
    @range = split(/\.\./, $port);
    if (scalar(@range) > 1) {
      for (my $idx = $range[0]; $idx < $range[1] + 1; $idx++) { push(@result, $idx); }
    } else {
      push(@result, $port);
    }
  }
    
  return(@result);
}

sub valid_port {
  my $port = shift;

  if (($port > 1) && ($port < 65536)) { return(1); }
  return(0);
}

sub valid_oper {
  my ($nick, $flag) = (shift, shift);

  if ($FXsock{'list'}{$nick} && $FXsock{'list'}{$nick}{'oper'}) {
    if (!$flag) { return(1); }

    my $oper = $FXsock{'list'}{$nick}{'oper'};
    if ($FXoper{lc($oper)}{'flags'} & $FXflags{'oper'}{'o'}) { return(1); }
  }

  if ($FXirc{'client'}{lc($nick)} && $FXirc{'client'}{lc($nick)}{'oper'}) {
    if (!$flag) { return(1); }

    my $oper = $FXirc{'client'}{lc($nick)}{'oper'};
    if ($FXoper{lc($oper)}{'flags'} & $FXflags{'oper'}{$flag}) { return(1); }
  }

  return(0);
}

sub send_motd {
  my ($socket, $file) = (shift, shift);

  if (!open(FILE, "<" . $file)) {
    &putlog("y", "Unable to open MOTD file: %s", $file);
    &putsock($socket, "MOTD is missing");
    return(1);
  }

  my @lines = <FILE>;
  close(FILE);

  foreach my $buffer (@lines) {
    chomp($buffer);
    $buffer = &substitute($buffer);
    $buffer =~ s/\s+$//;

    &putsock($socket, $buffer);
  }

  return(0);
}

sub disconnect {
  my $reason = shift;

  if ($FXbot{'PageServ'}{'enabled'}) {
    my $im_disconnect = {
      'msn'    => \&msn_disconnect,
      'yahoo'  => \&yahoo_disconnect,
      'icq'    => \&icq_disconnect,
      'aim'    => \&aim_disconnect,
      'jabber' => \&jabber_disconnect
    };

    foreach my $im ("msn", "yahoo", "icq", "aim", "jabber") {
      if (!$FXbot{'PageServ'}{$im . "_id"}) { next; }

      if ($FXsock{$im}) {
        if ($im eq "msn") {
          foreach (keys(%{$FXsock{'list'}})) {
            if ($FXsock{'list'}{$_}{'type'} eq "MSN-X") { &msn_remsession($FXsock{'list'}{$_}{'session'}, "*"); }
          }
        }

        foreach (keys(%{$FXim{$im}{'buddies'}})) {
          if (!$FXim{$im}{'buddies'}{$_}{'buddy'} || !$FXim{$im}{'buddies'}{$_}{'online'}) { next; }
          &{$im_disconnect->{$im}}($FXim{$im}{'buddies'}{$_}{'id'}, "Disconnected");
        }
      }
    }
  }

  foreach (keys(%FXbot)) {
    if ($FXbot{$_}{'enabled'}) {
      &putsock($FXsock{'irc'}, ":%s QUIT :%s", $FXbot{$_}{'nick'}, $reason);
      &chanlog("q", undef, $FXbot{$_}{'nick'}, $reason, undef);
      &del_client($FXbot{$_}{'nick'});
    }
  }

  &putsock($FXsock{'irc'}, ":%s ERROR :%s", $FXserv{'name'}, $reason);
  &putsock($FXsock{'irc'}, ":%s QUIT", $FXserv{'name'});
  &del_socket($FXsock{'irc'});

  return(0);
}

sub restart {
  &save_conf();
  &putinfo("y", "Restarting services");
  &disconnect("Restarting services");
  if (!unlink($PidFile)) { &putlog("y", "Unable to delete pid file: %s", $PidFile); }
  exec("perl $0");
  exit(0);
}

sub shutdown {
  my ($save, $reason) = (shift, shift);

  &disconnect($reason);

  foreach (keys(%{$FXsock{'list'}})) {
    vec($bit, $FXsock{'list'}{$_}{'fileno'}, 1) = 0;
    $FXsock{'list'}{$_}{'sock'}->close();
    delete($FXsock{'list'}{$_});
  }

  if ($save) { &save_conf(); }
  &putinfo("y", "Shutting down services %s", $reason ? "\(" . $reason . "\)" : "...");
  if (!unlink($PidFile)) { &putlog("y", "Unable to delete pid file: %s", $PidFile); }
  exit(0);
}

sub format_timestr {
  my $str = shift;
  my $sec = 0;
  my $tmp = 0;
  my $output;

  foreach (split(//, $str)) {
    if ($_ =~ /\d/) {
      $tmp .= $_;
    } else {
      if (lc($_) eq "w") {
        $sec += $tmp * 604800;
      } elsif (lc($_) eq "d") {
        $sec += $tmp * 86400;
      } elsif (lc($_) eq "h") {
        $sec += $tmp * 3600;
      } elsif (lc($_) eq "m") {
        $sec += $tmp * 60;
      } elsif (lc($_) eq "s") {
        $sec += $tmp;
      } else {
        $sec = 0;
      }

      $tmp = "";
    }
  }

  $output = $sec;

  return($output);
}

sub format_date {
  my $date = shift;
  my $output;

  my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime($date);
  $year = sprintf("%02d", $year % 100);

  $output = sprintf("%02d/%02d/%02d %02d:%02d:%02d", $mday, ++$mon, $year, $hour, $min, $sec);

  return($output);
}

sub format_size {
  my $bytes = shift;
  my $output;

  if ($bytes > 1073741824) {
    $output = sprintf("%.2f", $bytes / 1073741824);
    $output .= "Gb";
  } elsif ($bytes > 1048576) {
    $output = sprintf("%.2f", $bytes / 1048576);
    $output .= "Mb";
  } elsif ($bytes > 1024) {
    $output = sprintf("%.2f", $bytes / 1024);
    $output .= "Kb";
  } else {
    $output = sprintf("%.2f", $bytes);
    $output .= "b";
  }

  return($output);
}

sub format_time {
  my $time = shift;
  my $output;

  my $diff = abs($ts - $time);

  my $wks = int($diff / 604800); $diff %= 604800;
  my $day = int($diff / 86400);  $diff %= 86400;
  my $hrs = int($diff / 3600);   $diff %= 3600;
  my $min = int($diff / 60);     $diff %= 60;

  if ($wks) { $output .= $wks . "w "; }
  if ($day) { $output .= $day . "d "; }
  if ($hrs) { $output .= $hrs . "h "; }
  if ($min) { $output .= $min . "m "; }
  $output .= $diff . "s";
  $output =~ s/\s+$//;

  return($output);
}

sub plain2regexp {
  my $plain = shift;

  $plain = quotemeta($plain);
  $plain =~ s/\\\*/.*/g;
  $plain =~ s/\\\?/./g;

  return('^' . $plain . '$');
}

sub int2bytes {
  my ($bytes, $value) = (shift, shift);
  my @result;

  if (!$value) { $value = 0; }

  for (my $idx = 0; $idx < $bytes; $idx++) { unshift @result, ($value >> ($idx * 8) & 0xFF); }

  return(@result);
}

sub int2endian {
  my ($bytes, $value) = (shift, shift);
  my @result;

  for (my $idx = 0; $idx < $bytes; $idx++) { push(@result, ($value >> ($idx * 8) & 0xFF)); }

  return(@result);
}

sub str2bytes {
  my $string = shift;
  my @result;

  foreach (split(//, $string)) { push(@result, ord($_)); }

  return(@result);
}

sub bytes2int {
  my ($array, $start, $bytes) = (shift, shift, shift);
  my $result = 0;

  for (my $idx = $start; $idx < $start + $bytes; $idx++) {
    $result <<= 8;
    $result |= ($array->[$idx] || 0);
  }

  return($result);
}

sub endian2int {
  my ($array, $start, $bytes) = (shift, shift, shift);
  my $result = 0;

  for (my $idx = $start + $bytes - 1; $idx >= $start; $idx--) {
    $result <<= 8;
    $result |= ($array->[$idx] || 0);
  }

  return($result);
}

sub bytes2str {
  my ($array, $start, $bytes) = (shift, shift, shift);
  my $result;

  for (my $idx = $start; $idx < $start + $bytes; $idx++) { $result .= $array->[$idx] ? chr($array->[$idx]) : ''; }

  return($result);
}

sub base64 {
  my ($v, $n) = (shift, shift);
  my $result = '';
  my $salt = "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

  while (--$n >= 0) {
    $result .= substr($salt, $v & 0x3f, 1);
    $v >>= 6;
  }

  return($result);
}

1;