use strict;

our (%FXbot, %FXstatus, %FXserv, %FXoper, %FXirc, %FXcfg, %FXmsg, %FXsock, %FXstats, %FXscan, $scan, $ts, $LangPack);
$scan = 0;

$FXbot{'ScanServ'}{'sub'} = {
  'HELP'    => \&ss_help,
  'SCAN'    => \&ss_scan,
  'VERSION' => \&ss_version
};

$FXbot{'ScanServ'}{'public_sub'} = {
  'SCAN' => \&ss_scan
};

$FXbot{'ScanServ'}{'other_sub'} = {
  'process_ss'  => \&process_ss,
  'scan_add'    => \&scan_add,
  'scan_del'    => \&scan_del,
  'scan_reaper' => \&scan_reaper,
  'ss_check'    => \&ss_check,
  'ss_reply'    => \&ss_reply,
  'ss_notice'   => \&ss_notice,
  'ss_ctcp'     => \&ss_ctcp
};

sub scan_reaper {
  my $pid;

  while (($pid = waitpid(-1, &WNOHANG)) > 0) {
    my $exit = $? >> 8;
    $FXstatus{$pid} = $exit;
  }
}

sub ss_notice {
  my ($nick, $format, @arg) = (shift, shift, @_);
  my $buffer = sprintf($format, @arg);

  &putsock($FXsock{'irc'}, ":%s %s %s :%s", $FXbot{'ScanServ'}{'nick'}, $FXcfg{'general'}{'communicate'}, $nick, $buffer);
  return(0);
}

sub ss_ctcp {
  my ($nick, $notice, $format, @arg) = (shift, shift, @_);
  my $buffer = sprintf($format, @arg);

  &putsock($FXsock{'irc'}, ":%s %s %s :\cA%s\cA", $FXbot{'ScanServ'}{'nick'}, $notice ? "NOTICE" : "PRIVMSG", $nick, $buffer);
  return(0);
}

sub process_ss {
  my ($nick, $query) = (shift, shift);

  if (&check_flood($nick)) { return(0); }

  if ($query =~ /\cA(.*)\cA/) {
    my $ctcp = uc((split(/ /, $1, 0))[0]);

    if ($ctcp eq "VERSION") {
      &ss_ctcp($nick, 1, "VERSION %s", "FXserv IRC Services $FXserv{'version'}");
      &ss_ctcp($nick, 1, "VERSION %s", "Administrator: $FXserv{'admin'}");
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $FXbot{'ScanServ'}{'nick'}, $ctcp, $FXirc{'client'}{lc($nick)}{'nick'}, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    } elsif ($ctcp eq "PING") {
      &ss_ctcp($nick, 1, "PING %s", (split(/ /, $1, 0))[1]);
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $FXbot{'ScanServ'}{'nick'}, $ctcp, $FXirc{'client'}{lc($nick)}{'nick'}, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    } elsif ($ctcp eq "TIME") {
      &ss_ctcp($nick, 1, "TIME %s", scalar localtime());
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $FXbot{'ScanServ'}{'nick'}, $ctcp, $FXirc{'client'}{lc($nick)}{'nick'}, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    } else {
      &putinfo("y", "%s: CTCP [%s] received from %s!%s\@%s", $FXbot{'ScanServ'}{'nick'}, $1, $FXirc{'client'}{lc($nick)}{'nick'}, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'});
    }

    return(0);
  }

  my $cmd = (split(/ /, $query, 0))[0];
  my $arg = (split(/ /, $query, 2))[1];

  if ($FXbot{'ScanServ'}{'sub'}->{uc($cmd)}) {
    &{$FXbot{'ScanServ'}{'sub'}->{uc($cmd)}}($nick, $arg);
  } else {
    &ss_notice($nick, $FXmsg{'unkn_cmd'}, $cmd);
  }

  return(0);
}

sub ss_help {
  my ($nick, $arg) = (shift, shift);

  if (!$arg) { $arg = "index"; }

  my $file = lc($arg);
  $file =~ s/ /\//g;
  $file = "help/" . $LangPack . "/scanserv/" . $file;

  if (!open(FILE, "<" . $file)) {
    &ss_notice($nick, $FXmsg{'no_help'}, $arg);
    return(0);
  }

  my @lines = <FILE>;
  close(FILE);

  foreach my $buffer (@lines) {
    chomp($buffer);
    $buffer = &substitute($buffer);

    if ($buffer =~ /^\%\+a/) {
      if (&valid_oper($nick, "a")) {
        $buffer =~ s/^\%\+a//;
        &ss_notice($nick, $buffer);
      }
    } elsif ($buffer =~ /^\%\+o/) {
      if (&valid_oper($nick, "o")) {
        $buffer =~ s/^\%\+o//;
        &ss_notice($nick, $buffer);
      }
    } else {
      &ss_notice($nick, $buffer);
    }
  }

  &servicelog($nick, $FXbot{'ScanServ'}{'nick'}, "HELP %s", $arg);
  return(0);
}

sub ss_version {
  my ($nick, $arg) = (shift, shift);

  if (!&valid_oper($nick, "o")) {
    &ss_notice($nick, $FXmsg{'no_priv'});
    return(0);
  }

  if (!$arg) {
    &ss_notice($nick, $FXmsg{'syntax'}, "VERSION <ADD|DEL|LIST|STATS> [reply] [action] [reason]");
    return(0);
  }

  my $what = (split(/ /, $arg, 0))[0];
  my $reply = (split(/ /, $arg, 0))[1]; $reply =~ s/\+/ /g;
  my $action = (split(/ /, $arg, 0))[2];
  my $reason = (split(/ /, $arg, 4))[3];

  if (uc($what) eq "ADD") {
    if (!$reply || ((uc($action) ne "KILL") && (uc($action) ne "WARN")) || !$reason) {
      &ss_notice($nick, $FXmsg{'syntax'}, "VERSION ADD <reply> <action> <reason>");
      return(0);
    }

    if ($FXirc{'version'}{lc($reply)}) {
      &ss_notice($nick, $FXmsg{'version_in'}, $reply, $reason);
      return(0);
    }

    $FXirc{'version'}{lc($reply)}{'reply'} = $reply;
    $FXirc{'version'}{lc($reply)}{'action'} = uc($action);
    $FXirc{'version'}{lc($reply)}{'reason'} = $reason;
    &ss_notice($nick, $FXmsg{'version_add'}, $reply, $reason);
    &servicelog($nick, $FXbot{'ScanServ'}{'nick'}, "VERSION ADD %s (%s) [%s]", $reply, uc($action), $reason);
    return(0);

  } elsif (uc($what) eq "DEL") {
    if (!$reply) {
      &ss_notice($nick, $FXmsg{'syntax'}, "VERSION DEL <reply>");
      return(0);
    }

    if (!$FXirc{'version'}{lc($reply)}) {
      &ss_notice($nick, $FXmsg{'no_version'}, $reply);
      return(0);
    }

    delete($FXirc{'version'}{lc($reply)});
    &ss_notice($nick, $FXmsg{'version_del'}, $reply);
    &servicelog($nick, $FXbot{'ScanServ'}{'nick'}, "VERSION DEL %s", $reply);
    return(0);

  } elsif (uc($what) eq "LIST") {
    &ss_notice($nick, "-- Listing bad clients version --");
    &ss_notice($nick, "[Client reply                  ] [Action] [Reason                   ]");

    foreach (keys(%{$FXirc{'version'}})) {
      my $reason = $FXirc{'version'}{$_}{'reason'};
      if (length($reason) > 22) { $reason = substr($reason, 0, 22) . "..."; }
      &ss_notice($nick, "[%-30s] [%-6s] [%-25s]", $FXirc{'version'}{$_}{'reply'}, uc($FXirc{'version'}{$_}{'action'}), $reason);
    }

    my $idx = scalar(keys(%{$FXirc{'version'}}));
    &ss_notice($nick, "-- End of list [%d version%s found] --", $idx, $idx > 1 ? "s" : "");
    &servicelog($nick, $FXbot{'ScanServ'}{'nick'}, "VERSION LIST");
    return(0);

  } elsif (uc($what) eq "STATS") {
    &ss_notice($nick, "-- Client version statistics --");
    &ss_notice($nick, "[Replies] [Client                                  ]");

    my @version = sort { $FXstats{'version'}{$b}{'current'} <=> $FXstats{'version'}{$a}{'current'} } keys(%{$FXstats{'version'}});
    my $idx = 0;
    my $cnt = 0;

    foreach (@version) {
      ++$idx;
      $cnt += $FXstats{'version'}{$_}{'current'};

      if (!$FXstats{'version'}{$_}{'current'}) { next; }
      &ss_notice($nick, "[%-7d] [%-40s]", $FXstats{'version'}{$_}{'current'}, $FXstats{'version'}{$_}{'reply'});
    }

    &ss_notice($nick, "-- End of list [%d repl%s, %d online] --", $idx, $idx > 1 ? "ies" : "y", $cnt);
    &servicelog($nick, $FXbot{'ScanServ'}{'nick'}, "VERSION STATS");
    return(0);

  } else {
    &ss_notice($nick, $FXmsg{'syntax'}, "VERSION <ADD|DEL|LIST|STATS> [reply] [action] [reason]");
    return(0);
  }
}

sub ss_reply {
  my ($nick, $reply) = (shift, shift);

  if (!$FXirc{'client'}{lc($nick)}{'reply'}) {
    $FXirc{'client'}{lc($nick)}{'reply'} = $reply;

    if (!$FXstats{'version'}{$reply}) {
      $FXstats{'version'}{$reply}{'reply'} = $reply;
      $FXstats{'version'}{$reply}{'current'} = 0;
      $FXstats{'version'}{$reply}{'max'} = 0;
      $FXstats{'version'}{$reply}{'max_ts'} = 0;
    }

    if (++$FXstats{'version'}{$reply}{'current'} > $FXstats{'version'}{$reply}{'max'}) {
      $FXstats{'version'}{$reply}{'max'} = $FXstats{'version'}{$reply}{'current'};
      $FXstats{'version'}{$reply}{'max_ts'} = $ts;
    }
  }

  foreach (keys(%{$FXirc{'version'}})) {
    my $regex = &plain2regexp($FXirc{'version'}{$_}{'reply'});

    if ($reply =~ /$regex/i) {
      &putlog("y", "Bad VERSION client: %s (%s\@%s) [%s]", $nick, $FXirc{'client'}{lc($nick)}{'user'}, $FXirc{'client'}{lc($nick)}{'host'}, $FXirc{'version'}{$_}{'reply'});

      if (uc($FXirc{'version'}{$_}{'action'}) eq "KILL") {
        &kill_client($nick, $FXirc{'version'}{$_}{'reason'}, $FXbot{'ScanServ'}{'nick'}, $FXbot{'ScanServ'}{'service'});
        &add_seen($nick, "QUIT", $FXirc{'version'}{$_}{'reason'});
        &chanlog("q", undef, $nick, $FXirc{'version'}{$_}{'reason'}, undef);
        &del_client($nick);
        ++$FXstats{'kills'}{'oper'};
      } elsif (uc($FXirc{'version'}{$_}{'action'}) eq "WARN") {
        &ss_notice($nick, $FXirc{'version'}{$_}{'reason'});
      }

      return(0);
    }
  }

  return(0);
}

sub ss_scan {
  my ($nick, $target) = (shift, shift);

  if (!&valid_oper($nick, "o")) {
    &ss_notice($nick, $FXmsg{'no_priv'});
    return(0);
  }

  if ($FXirc{'client'}{lc($target)}) {
    $target = $FXirc{'client'}{lc($target)}{'nick'};
  } elsif ($target !~ /\./) {
    &ss_notice($nick, $FXmsg{'syntax'}, "SCAN <nick|host>");
    return(0);
  }

  &ss_check($nick, "query", $target);

  &servicelog($nick, $FXbot{'ScanServ'}{'nick'}, "SCAN %s", $target);
  return(0);
}

sub ss_check {
  my ($nick, $what, $target) = (shift, shift, shift);

  if ($what eq "scan") {
    if (!open(FILE, "<" . $FXbot{'ScanServ'}{'scan_notice'})) { }

    my @lines = <FILE>;
    close(FILE);

    foreach my $buffer (@lines) {
      chomp($buffer);
      $buffer = &substitute($buffer);

      &ss_notice($nick, $buffer);
    }
  }

  if ($FXbot{'ScanServ'}{'id'} >= 65535) { $FXbot{'ScanServ'}{'id'} = 0; }
  my $id = ++$FXbot{'ScanServ'}{'id'};

  $FXscan{$what}{$id}{'nick'} = $FXirc{'client'}{lc($nick)}{'nick'};
  $FXscan{$what}{$id}{'total'} = 0;
  $FXscan{$what}{$id}{'completed'} = 0;
  $FXscan{$what}{$id}{'list'} = ();
  $FXscan{$what}{$id}{'unsecure'} = 0;
  if ($what eq "scan") { $FXscan{$what}{$id}{'host'} = $FXirc{'client'}{lc($nick)}{'host'}; }
  if ($what eq "query") { $FXscan{$what}{$id}{'host'} = $target; }
  $FXscan{$what}{$id}{'port'} = 0;
  $FXscan{$what}{$id}{'proto'} = "";
  $FXscan{$what}{$id}{'start'} = $ts;

  my $idx = 0;

  foreach (&parse_range($FXbot{'ScanServ'}{'scan_wingate'})) {
    ++$idx;

    $FXscan{$what}{$id}{'list'}{$idx}{'proto'} = "Wingate";
    $FXscan{$what}{$id}{'list'}{$idx}{'port'} = $_;
    $FXscan{$what}{$id}{'list'}{$idx}{'start'} = 0;

    if ($scan >= $FXbot{'ScanServ'}{'scan_max'}) {
      $FXscan{$what}{$id}{'list'}{$idx}{'pid'} = 0;
      $FXscan{$what}{$id}{'list'}{$idx}{'queued'} = 1;
    } else {
      my $pid = &scan_add($id, $what, "Wingate", $_);
      $FXscan{$what}{$id}{'list'}{$idx}{'pid'} = $pid;
      $FXscan{$what}{$id}{'list'}{$idx}{'start'} = $ts;
    }
  }

  foreach (&parse_range($FXbot{'ScanServ'}{'scan_http'})) {
    ++$idx;

    $FXscan{$what}{$id}{'list'}{$idx}{'proto'} = "HTTP";
    $FXscan{$what}{$id}{'list'}{$idx}{'port'} = $_;
    $FXscan{$what}{$id}{'list'}{$idx}{'start'} = 0;

    if ($scan >= $FXbot{'ScanServ'}{'scan_max'}) {
      $FXscan{$what}{$id}{'list'}{$idx}{'pid'} = 0;
      $FXscan{$what}{$id}{'list'}{$idx}{'queued'} = 1;
    } else {
      my $pid = &scan_add($id, $what, "HTTP", $_);
      $FXscan{$what}{$id}{'list'}{$idx}{'pid'} = $pid;
      $FXscan{$what}{$id}{'list'}{$idx}{'start'} = $ts;
    }
  }

  foreach (&parse_range($FXbot{'ScanServ'}{'scan_post'})) {
    ++$idx;

    $FXscan{$what}{$id}{'list'}{$idx}{'proto'} = "POST";
    $FXscan{$what}{$id}{'list'}{$idx}{'port'} = $_;
    $FXscan{$what}{$id}{'list'}{$idx}{'start'} = 0;

    if ($scan >= $FXbot{'ScanServ'}{'scan_max'}) {
      $FXscan{$what}{$id}{'list'}{$idx}{'pid'} = 0;
      $FXscan{$what}{$id}{'list'}{$idx}{'queued'} = 1;
    } else {
      my $pid = &scan_add($id, $what, "POST", $_);
      $FXscan{$what}{$id}{'list'}{$idx}{'pid'} = $pid;
      $FXscan{$what}{$id}{'list'}{$idx}{'start'} = $ts;
    }
  }

  foreach (&parse_range($FXbot{'ScanServ'}{'scan_socks4'})) {
    ++$idx;

    $FXscan{$what}{$id}{'list'}{$idx}{'proto'} = "SOCKS4";
    $FXscan{$what}{$id}{'list'}{$idx}{'port'} = $_;
    $FXscan{$what}{$id}{'list'}{$idx}{'start'} = 0;

    if ($scan >= $FXbot{'ScanServ'}{'scan_max'}) {
      $FXscan{$what}{$id}{'list'}{$idx}{'pid'} = 0;
      $FXscan{$what}{$id}{'list'}{$idx}{'queued'} = 1;
    } else {
      my $pid = &scan_add($id, $what, "SOCKS4", $_);
      $FXscan{$what}{$id}{'list'}{$idx}{'pid'} = $pid;
      $FXscan{$what}{$id}{'list'}{$idx}{'start'} = $ts;
    }
  }

  foreach (&parse_range($FXbot{'ScanServ'}{'scan_socks5'})) {
    ++$idx;
    $FXscan{$what}{$id}{'list'}{$idx}{'proto'} = "SOCKS5";
    $FXscan{$what}{$id}{'list'}{$idx}{'port'} = $_;
    $FXscan{$what}{$id}{'list'}{$idx}{'start'} = 0;

    if ($scan >= $FXbot{'ScanServ'}{'scan_max'}) {
      $FXscan{$what}{$id}{'list'}{$idx}{'pid'} = 0;
      $FXscan{$what}{$id}{'list'}{$idx}{'queued'} = 1;
    } else {
      my $pid = &scan_add($id, $what, "SOCKS5", $_);
      $FXscan{$what}{$id}{'list'}{$idx}{'pid'} = $pid;
      $FXscan{$what}{$id}{'list'}{$idx}{'start'} = $ts;
    }
  }

  $FXscan{$what}{$id}{'total'} = $idx;

  return(0);
}

sub scan_add {
  my ($id, $what, $proto, $port) = (shift, shift, shift, shift);
  my $host = "";
  my $result = 0;
  my $buffer = "";

  if ($what eq "scan") { $host = $FXscan{$what}{$id}{'host'}; }
  if ($what eq "query") {
    my $target = $FXscan{$what}{$id}{'host'};
    if ($FXirc{'client'}{lc($target)}) {
      $host = $FXirc{'client'}{lc($target)}{'host'};
    } else {
      $host = $target;
    }
  }

  my $pid = undef;

  if ($pid = fork()) {
    ++$scan;
    return($pid);
  } elsif (defined($pid)) {
    eval {
      local $SIG{'ALRM'} = sub { die("alarm!\n"); };
      alarm($FXbot{'ScanServ'}{'scan_timeout'});

      my $socket = &open_socket("tcp", $host, $port, $FXbot{'ScanServ'}{'scan_timeout'});

      if (!$socket) { exit(0); }

      $socket->autoflush(1);

      if ($proto eq "Wingate") {
        &putsock($socket, "%s\r", $FXbot{'ScanServ'}{'scan_test'});
        recv($socket, $buffer, 512, 0);
        my $string = quotemeta($FXbot{'ScanServ'}{'scan_string'});
        if ($buffer =~ /$string/) { $result = 1; }

      } elsif ($proto eq "HTTP") {
        &putsock($socket, "CONNECT %s HTTP/1.0\r\n\r", $FXbot{'ScanServ'}{'scan_test'});
        recv($socket, $buffer, 512, 0);
        if ($buffer =~ /^HTTP\/1.[01] 200 [Cc]onnection/) { $result = 1; }

      } elsif ($proto eq "POST") {
        &putsock($socket, "POST http://%s/ HTTP/1.0\r\n\r", $FXbot{'ScanServ'}{'scan_test'});
        recv($socket, $buffer, 512, 0);
        if ($buffer =~ /^HTTP\/1.[01] 200 [Cc]onnection/) { $result = 1; }

      } elsif ($proto eq "SOCKS4") {
        my @x  = split(/:/, $FXbot{'ScanServ'}{'scan_test'});
        my @ip = split(/\./, $x[0]);
        print $socket sprintf("%c%c%c%c%c%c%c%c%c", 4, 1, int($x[1] / 256), $x[1] % 256, $ip[0], $ip[1], $ip[2], $ip[3], 0);
        recv($socket, $buffer, 512, 0);
        if ((ord(substr($buffer, 0, 1)) == 0) && (ord(substr($buffer, 1, 1)) == 90)) { $result = 1; }

      } elsif ($proto eq "SOCKS5") {
        my @x = split(/:/, $FXbot{'ScanServ'}{'scan_test'});
        my @ip = split(/\./, $x[0]);
        print $socket sprintf("%c%c%c", 5, 1, 0);
        print $socket sprintf("%c%c%c%c%c%c%c%c%c%c", 5, 1, 0, 1, $ip[0], $ip[1], $ip[2], $ip[3], int($x[1] / 256), $x[1] % 256);
        recv($socket, $buffer, 512, 0);
        if ((ord(substr($buffer, 0, 1)) == 5) && (ord(substr($buffer, 1, 1)) == 0)) { $result = 1; }
      }

      close($socket);
      alarm(0);
    };

    exit($result);
  }
}

sub scan_del {
  my ($nick, $what) = (shift, shift);

  foreach my $id (keys(%{$FXscan{$what}})) {
    if (lc($nick) eq lc($FXscan{$what}{$id}{'nick'})) {
      foreach my $proc (keys(%{$FXscan{$what}{$id}{'list'}})) {
        if ($FXscan{$what}{$id}{'list'}{$proc}{'pid'}) {
          my $pid = $FXscan{$what}{$id}{'list'}{$proc}{'pid'};

          if ($pid) {
            if ($FXstatus{$pid}) { delete($FXstatus{$pid}); }
            ++$FXscan{$what}{$id}{'completed'};
            --$scan;
          }

          $FXscan{$what}{$id}{'list'}{$proc}{'pid'} = 0;
          $FXscan{$what}{$id}{'list'}{$proc}{'queued'} = 0;
        }
      }

      last;
    }
  }

  return(0);
}

1;